/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef RG3DROM_VERTEX_ARRAY_H
#define RG3DROM_VERTEX_ARRAY_H

#include "rg3DRom.h"

class sk3DRomVertexArray_RenderGroup
{
public:
        int m_start;
        int m_count;
        int m_mdl_texid;
	rgDWORD m_min_vertex_index;
	rgDWORD m_max_vertex_index;
        GLuint m_texid;

        sk3DRomVertexArray_RenderGroup() {
                m_start = 0;
                m_count = 0;
                m_mdl_texid = -1;
                m_texid = 0;
		m_min_vertex_index = 0;
		m_max_vertex_index = 0x7FFFFFFF;
        }

        sk3DRomVertexArray_RenderGroup( int p_start) {
                m_start = p_start;
                m_count = 0;
                m_mdl_texid = -1;
                m_texid = 0;
		m_min_vertex_index = 0;
		m_max_vertex_index = 0x7FFFFFFF;
        }
};


// Vertex array rendering facility.
// This is provided for convinience.
class sk3DRomVertexArray
{
public:
	sk3DRomVertexArray() {
		m_index = 0;
		m_force_use_all_faces = false;
		m_force_share_vertex = false;
		m_texcoord_size = 2;
		m_aux_vertex_count = 0;
		m_aux_vertex = 0;
		m_vertex_fid_count = 0;
		m_vertex_fid = 0;
		m_vn_accel = 0;
	}
	virtual ~sk3DRomVertexArray() {
		if( m_index) delete [] m_index;
		if( m_aux_vertex) delete [] m_aux_vertex;
		if( m_vertex_fid) delete [] m_vertex_fid;
		if( m_vn_accel) delete [] m_vn_accel;
	}

	virtual int create_arrays( const sk3DRomModel &mdl) = 0;
	virtual void update_arrays( const sk3DRomModel &mdl) = 0;
	virtual void render_arrays() = 0;
	virtual void delete_arrays() = 0;

	// Getting vertex array attributes.
	virtual float *vertex_array_start() = 0;
	virtual float *normal_array_start() = 0;
	virtual float *texcoord_array_start() = 0;

	virtual const int vertex_stride() const = 0;
	virtual const int normal_stride() const = 0;
	virtual const int texcoord_stride() const = 0;

        int non_alpha_face_count( const sk3DRomModel &mdl);
        void non_alpha_faces( const sk3DRomModel &mdl, std::vector <int> &naf);

	int create_face_index_array( const sk3DRomModel &mdl) {
		if( m_index) { delete [] m_index; m_index = 0;}
		m_index = create_face_index_array( mdl, m_face_count,
			m_render_group);
		if( !m_index) return -1;
		return 0;
	}

	std::vector <sk3DRomVertexArray_RenderGroup> &
                face_render_group() { return m_render_group;}
        int face_count() const { return m_face_count;}
        const rgWORD *index_array() const { return m_index;}

	virtual void set_use_all_faces( const bool b) {
		m_force_use_all_faces = b;}
	virtual void set_share_vertex( const bool b) {
		m_force_share_vertex = b;}
                                                                                
        virtual bool supports_direct_transform() const { return false;}

	int create_auxiliary_vertices(
		const sk3DRomModel &mdl,
		const int idx_size,
		rgWORD *idx,
		const int *fid);

	virtual void dump_vertex( std::ostream &o, const int i){}

	int aux_vertex_count() const { return m_aux_vertex_count;}

	const int *vertex_face_id_array() const { return m_vertex_fid;}

	const int vertex_face_fid( const int i) const {
		return ((m_vertex_fid[i] >> 8) & 0x00FFFFFF);}

	const int vertex_face_vid( const int i) const {
		return (m_vertex_fid[i] & 0x03);}

	void set_vertex_face_fid( const int i, const int fid) {
		m_vertex_fid[i] = (m_vertex_fid[i] & 0x3) |
			((fid & 0x00FFFFFF) << 8);
	}

	void set_vertex_face_vid( const int i, const int vid) {
		m_vertex_fid[i] = (m_vertex_fid[i] & 0xFFFFFFFC) | (vid & 0x3);
	}

	virtual bool export_to_OBJ( std::ostream &o, const sk3DRomModel &mdl) {
		o << "# Exporting to OBJ not supported by abstract class "
			"\"sk3DRomVertexArray\"\n";
		return false;
	}

	int calc_accel_array_size( const sk3DRomModel &mdl) const;
	bool create_accelerated_transformation_array( const sk3DRomModel &mdl);

	rgDWORD *vn_accelerated_array() { return m_vn_accel;}

protected:
	bool m_force_use_all_faces;
        bool m_force_share_vertex;
	int m_texcoord_size;

	// Vertex-Normal accelerator array. This array allows for a more
	// cache coherent transformation. As well, it reduces the number of
	// pointer indirections required per transformed vertex.
	//
	// Format:
	//	bone 0 [jcs][mid][count][vn-0][vn-1][vn-2]...[vn-n]
	//	bone 1 [jcs][mid][count][vn-0][vn-1][vn-2]...[vn-n]
	//	.
	//	.
	//	.
	//	bone n [jcs][mid][count][vn-0][vn-1][vn-2]...[vn-n]
	rgDWORD *m_vn_accel;

	int m_aux_vertex_count;
	int *m_aux_vertex;

private:
	// Contains the index numer of vertices used by faces. The size of
	// this array is always a multiple of 3 (since each face/trangle) has
	// 3 vertices.
        rgWORD *m_index;

	// This array contains the index number of the face and the face-vertex
	// (vid) used to describe the vertices "normal" and "texture" data.
	int m_vertex_fid_count;
	int *m_vertex_fid;

	int m_face_count;
	std::vector <sk3DRomVertexArray_RenderGroup> m_render_group;
                                                                                
	// Description: Creates and returns an array containing face-vertex
	//      index values. The faces are sorted by texture usage. Also,
	//      "rgrp" is populated with a list of "render-groups". These
	//      indiciate which faces should be rendered together for optimal
	//      speed.
	rgWORD *create_face_index_array( const sk3DRomModel &mdl,
		int &count,
		std::vector <sk3DRomVertexArray_RenderGroup> &rgrp);

	void prepare_render_range( const rgWORD *idx,
		std::vector <sk3DRomVertexArray_RenderGroup> &rgrp) const;

	sk3DRomVertexArray( const sk3DRomVertexArray &);
	sk3DRomVertexArray &operator =( const sk3DRomVertexArray &);
};

class sk3DRomVNTArray : public sk3DRomVertexArray
{
public:
	sk3DRomVNTArray();
	virtual ~sk3DRomVNTArray();

	int create_arrays( const sk3DRomModel &mdl);
	void update_arrays( const sk3DRomModel &mdl);
	void update_texture_ids( const sk3DRomModel &mdl);
	void render_arrays();
	void delete_arrays();

        float *vertex_array_start() { return m_data;}
        float *normal_array_start() { return normal_offset( m_data);}
	float *texcoord_array_start() { return texcoord_offset( m_data);}
                                                                                
        const int vertex_stride() const { return 16;}
        const int normal_stride() const { return 16;}
	const int texcoord_stride() const { return 16;}

	void set_use_all_faces( const bool b);
	void set_share_vertex( const bool b);
	bool supports_direct_transform() const { return m_force_share_vertex;}

	void dump_vertex( std::ostream &o, const int i); 

	bool export_to_OBJ( std::ostream &o, const sk3DRomModel &mdl);
                                                                                
protected:

	void update_auxiliary_vertices();
                                                                                
private:
        int m_vertex_count;
        std::vector <sk3DRomVertexArray_RenderGroup> m_vertex_render_group;

	float *m_data;
	float *normal_offset( float *p_data) const { return p_data + 4;}
	float *fid_offset( float *p_data) const { return p_data + 7;}
	float *texcoord_offset( float *p_data) const { return p_data + 8;}
	rgWORD &fid_value( float *p_data) const {
		return *((rgWORD *)(fid_offset( p_data)));}

	void update_vertex( float *p, const sk3DRomModel &mdl,
		const sk3DRomFace &f, const int vid) const
	{
                mdl.fast_get_tv_vertex( f.vid[vid], p[0], p[1], p[2]);
                float *n = normal_offset( p);
                n[0] = f.n[vid].x;
                n[1] = f.n[vid].y;
                n[2] = f.n[vid].z;
        }
                                                                                
        int create_shared_vertex_array( const sk3DRomModel &mdl);
        int create_non_shared_vertex_array( const sk3DRomModel &mdl);

        sk3DRomVNTArray( const sk3DRomVNTArray &);
        sk3DRomVNTArray &operator =( const sk3DRomVNTArray &);
};

class sk3DRomVCTArray : public sk3DRomVertexArray
{
public:
	sk3DRomVCTArray();
	virtual ~sk3DRomVCTArray();

protected:

	float *v_array;	// Array of packed vertices in the format (x, y, z, w)
	float *ct_array;	// Array of packed color-texture.
			// format is (r, g, b, a, u1, v1, u2, v2)

private:
	sk3DRomVCTArray( const sk3DRomVCTArray &);
	sk3DRomVCTArray &operator =( const sk3DRomVCTArray &);
};

#endif
