/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "rgBSP.h"
#include <rgOpenGL.h>

rgBSP_NTTC_ExtData::rgBSP_NTTC_ExtData()
{
	edata = 0;
}

bool rgBSP_point_is_same( const rgVector &a, const rgVector &b)
{
	if(     (fabs( a.x - b.x) <= RGBSP_POINT_SAME_EPSILON) &&
		(fabs( a.y - b.y) <= RGBSP_POINT_SAME_EPSILON) &&
		(fabs( a.z - b.z) <= RGBSP_POINT_SAME_EPSILON)) return true;
	return false;
}

bool rgBSP_edge_can_be_split( rgVector pp, rgVector pn, rgVector a,
	rgVector b, const float epsilon, float &t, float &k)
{
	t = RGBSP_POINT_PLANE_DISTANCE( pp, pn, a);
	k = RGBSP_POINT_PLANE_DISTANCE( pp, pn, b);

	// Both points from the edge MUST not be "too close" to the splitting
	// plane.
	if( fabs( t) <= epsilon) return false;
	if( fabs( k) <= epsilon) return false;

	// Both split points MUST be on opposite sides.
	if( t > 0.0f) {
		// Point "a" is in front, so "b" MUST be behind.
		if( k > 0.0f) return false;
		else return true;

	}

	// Point "a" is behind, so "b" MUST be in front.
	if( k < 0.0f) return false;
	return true;
}

void rgBSP_NTTC_interp_fn(
	rgBSPVertex <rgBSP_NTTC_ExtData> &sp,
	const rgBSPVertex <rgBSP_NTTC_ExtData> &srcA,
	const rgBSPVertex <rgBSP_NTTC_ExtData> &srcB,
        const float t)
{
	const float k = 1.0f - t;
	// Interpolating first texture unit coordinate.
	sp.data.u1 = (srcA.data.u1 * k) + (srcB.data.u1 * t);
	sp.data.v1 = (srcA.data.v1 * k) + (srcB.data.v1 * t);

	// Interpolating second texture unit coordinate.
	sp.data.u2 = (srcA.data.u2 * k) + (srcB.data.u2 * t);
	sp.data.v2 = (srcA.data.v2 * k) + (srcB.data.v2 * t);

	// Interpolating colour component.
	//std::cerr << "a = " << srcA.data.a << ", " << srcB.data.a << "\n";
	sp.data.r = (srcA.data.r * k) + (srcB.data.r * t);
	sp.data.g = (srcA.data.g * k) + (srcB.data.g * t);
	sp.data.b = (srcA.data.b * k) + (srcB.data.b * t);
	sp.data.a = (srcA.data.a * k) + (srcB.data.a * t);

	// Interpolate the normals and then normalize the result.
	// FIXME: It is possible for the calculated normal to be a null
	//	vector. This can happen if the two source normals are
	//	parallel and point in the opposite direction.
	//	A fix for this should be found, or at least made known to
	//	the modeler so they do not create this condition for the
	//	BSP generater.
	sp.data.n = ((srcA.data.n * k) + (srcB.data.n * t)).normalize();
}

void rgBSP_NTTC_convex_hull_renderer(
	std::vector <rgBSPFace <rgBSP_NTTC_ExtData> > &p_f,
	rgMatrix *p_m, void *edata)
{
	const rgBSP_NTTC_ExtData *ptr;
        int j;
	if( !p_m) {
		// FIXME: Is it not-efficient to redundantly bind textures?
		//	That is, if texture "1" is bound, is binding it again
		//	inefficient?
		// Render the convex hull without performing a matrix pre-
		// transformation on the vertices and normals.
	        for( int i = 0; i < int( p_f.size()); i++) {
// FIXME: Replace rgBSP_NTTC_ExtData with polymorphism.
			glBindTexture( GL_TEXTURE_2D, (p_f[i]).m_tid);
	                glBegin( GL_POLYGON);
	                for( j = 0; j < int( p_f[i].v.size()); j++) {
				ptr = &(p_f[i].v[j].data);
				glColor4f( ptr->r, ptr->g, ptr->b, ptr->a);
				glTexCoord2f( ptr->u1, ptr->v1);
				glNormal( ptr->n);
	                        glVertex( p_f[i].v[j].v);
			}
	                glEnd();
	        }

	} else {
		// Render the convex hull while performing a matrix pre-
		// transformation on the vertices and normals.

		// FIXME: Does creating a local matrix variable improve
		//	caching???
		const rgMatrix m( *p_m);

		for( int i = 0; i < int( p_f.size()); i++) {
			glBindTexture( GL_TEXTURE_2D, (p_f[i]).m_tid);
	                glBegin( GL_POLYGON);
	                for( j = 0; j < int( p_f[i].v.size()); j++) {
				ptr = &(p_f[i].v[j].data);
				glColor4f( ptr->r, ptr->g, ptr->b, ptr->a);
				glTexCoord2f( ptr->u1, ptr->v1);

				// NOTE: Normals are vectors.
				glNormal( m * (ptr->n));

				// NOTE: Vertices are points.
				glVertex( m.pointMul( p_f[i].v[j].v));
			}
	                glEnd();
	        }
	}
        return;
}
