/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef RGBSP_H
#define RGBSP_H

#include <rgStandard.h>
#include <vector>

#define RGBSP_FACE_SPLIT_EPSILON	0.00001f
#define RGBSP_POINT_SAME_EPSILON	0.0001f

#define RGBSP_FACE_IS_PLANAR	0
#define RGBSP_FACE_ON_FRONT	1
#define RGBSP_FACE_ON_BACK	2

#define RGBSP_BACK_TO_FRONT_RENDER	0
#define RGBSP_FRONT_TO_BACK_RENDER	1

#define RGBSP_POINT_PLANE_DISTANCE( mv_pp, mv_pn, mv_p) \
        (mv_pn).dot( (mv_p) - (mv_pp))

// Description: Class that provides extended vertex data to rgBSPVertex.
//	The extended data contains information that needs to be interpolated
//	whenever a split vertex is created. The extended data provided in this
//	class is 2D texture coordinate's for two texture units, a four
//	colour component (RGBA) and a vertex normal. The cooresponding
//	interpolating function is "rgBSP_NTTC_interp_fn" though you may use
//	your own.
//
class rgBSP_NTTC_ExtData
{
public:
	rgVector n;		// Vertex normal
//	int tid1;		// Texture ID (unit 1)
//	int tid2;		// Texture ID (unit 2)
	float u1, v1;		// Vertex texture coordinate (unit 1)
	float u2, v2;		// Vertex texture coordinate (unit 2)
	float r, g, b, a;	// Vertex colour

	// NOTE: A "shallow" copy of "edata" is done. That is, only the
	//	pointer value (value of "edata") is copied, not what it is
	//	pointing to. "edata" is added for future exapnsion.
	void *edata;		// Hook for extra data.

	rgBSP_NTTC_ExtData();
};

template <class T>
class rgBSPVertex
{
public:
	rgVector v;
	T data;		// Additional vertex data (i.e. textures, colours)

	/*
	// The identification number of this vertex. When BSP generation
	// starts, each vertex is given a unique ID number. This allows for
	// colour, texture and other data to be interpolated whenever split
	// verticies are generated.
	int id[4];


	int srcA;	// ID number of one of the source verticies
	int srcB;	// ID number of the other source vertex

	// Value for interpolating both source verticies. "interp" is only
	// meaningfull if both "srcA" and "srcB" exist (i.e. are >= 0).
	rgFloat interp;
	*/

	rgBSPVertex() {}
	rgBSPVertex( rgVector p_v) { v = p_v;}
	rgBSPVertex( rgVector p_v, T &p_data) { v = p_v; data = p_data;}
};

template <class T>
class rgBSPFace
{
public:
	int m_tid;

	std::vector <rgBSPVertex <T> > v;
	rgVector normal;

	rgBSPFace() { m_tid = 0;}

	void gen_normal();

	void transform_vertices( void (*fn)( std::vector <rgBSPVertex<T> *> &,
		void *), void *p_data);

};

template <class T>
class rgBSPNode
{
public:
	rgVector plane_point;
	rgVector normal;

	// List of faces beloning in this node.
	// NOTE: Faces are only stored in "leaf" nodes. That is, if "f" is
	//	not empty (contains faces), then both "front" and "back" will
	//	be NULL.
	std::vector <rgBSPFace <T> > f;

	rgBSPNode <T> *front;
	rgBSPNode <T> *back;

	rgBSPNode();

	bool is_leaf() const {
		if( front || back) return false;
		return true;}

	void transform_face_verts(
		void (*fn)( std::vector <rgBSPVertex <T> *> &, void *),
		void *p_data);
};

template <class T>
class rgBSPTree
{
public:

	rgBSPNode <T> *root;

	bool use_bsp_matrix;
	rgMatrix bsp_matrix;
	void *extra_render_data;
	void (*convex_hull_renderer)( std::vector <rgBSPFace <T> > &,
		rgMatrix *m, void *edata);

	void (*interp_fn)( rgBSPVertex <T> &sp, const rgBSPVertex <T> &srcA,
		const rgBSPVertex <T> &srcB, const float k);

	rgBSPTree();
	~rgBSPTree();

	void clear();

	// Description: Runs through the BSP tree starting at node "node" or
	//	"root" (if no "node" parameter given). By default the renderer
	//	tries to render the geometry in the back (relative to the
	//	position "p"), followed by geometry in the front. This is the
	//	painter's algorithm. Supplying RGBSP_BACK_TO_FRONT_RENDER will
	//	also cause the rendering to be done back to front. The order
	//	can be reversed by supplying RGBSP_FRONT_TO_BACK_RENDER.
	//
	//	When a leaf node is reached, the geometry contained in the node
	//	is rendered (this geometry is convex so can be rendered in any
	//	order). "convex_hull_renderer" is called to render the geometry
	//	if it is not NULL, otherwise, "default_convex_hull_renderer" is
	//	called to render the geometry.
	//
	void render( const rgVector p) { render( p, root);}

	void render( const rgVector p, rgBSPNode <T> *node) {
		render( p, node, RGBSP_BACK_TO_FRONT_RENDER);}

	void render( const rgVector p, rgBSPNode <T> *node,
		const int order) {
		if( RGBSP_FRONT_TO_BACK_RENDER == order)
	                render_front_to_back( p, node);
	        else render_back_to_front( p, node);
	}

	void transform_node_verts(
		void (*fn)( std::vector <rgBSPVertex<T> *> &, void *),
		void *p_data){
		transform_node_verts( fn, p_data, root);
	}
	void transform_node_verts(
		void (*fn)( std::vector <rgBSPVertex<T> *> &, void *),
		void *p_data,
		rgBSPNode <T> *node);

	void print_tree() const { print_tree( 0, root);}
	void print_tree( const int tab, const rgBSPNode<T> *n) const;

	int gen_tree( std::vector <rgBSPFace <T> > &f,
		void (*p_interp)( rgBSPVertex <T> &, const rgBSPVertex <T> &,
		const rgBSPVertex <T> &, const float)) {
		clear();			// Delete any existing tree
		if( f.empty()) return 0;	// Check parameters
		return hlp_build_bsp_tree( f, root, (interp_fn = p_interp));
	}

private:

	void delete_tree( rgBSPNode <T> *&r) {
		if( r->front) delete_tree( r->front);
		if( r->back) delete_tree( r->back);
		delete r;
		r = 0;
	}

	void default_convex_hull_renderer( std::vector <rgBSPFace <T> > &p_f,
		rgMatrix *m, void *edata);

	void render_node( rgBSPNode <T> *node);
	void render_back_to_front( const rgVector p, rgBSPNode <T> *node);
	void render_front_to_back( const rgVector p, rgBSPNode <T> *node);
};

template <class T>
int rgBSP_choose_splitting_face( const std::vector <rgBSPFace <T> > &f);

template <class T>
bool rgBSP_face_splits_face( const rgBSPFace <T> &src,
	const rgBSPFace <T> &tar, const float epsilon, int &side);

template <class T>
bool rgBSP_plane_splits_face( rgVector pp, rgVector pn,
	const rgBSPFace <T> &tar, const float epsilon, int &side);

template <class T>
bool rgBSP_axis_aligned_face( const rgBSPFace <T> &f);

template <class T>
bool rgBSP_world_is_convex( const std::vector <rgBSPFace <T> > &f);

template <class T>
bool rgBSP_face_partitions_world( const int id,
	const std::vector <rgBSPFace <T> > &f);

template <class T>
void rgBSP_split_convex_face( rgVector pp, rgVector pn, rgBSPFace <T> &f,
	rgBSPFace <T> &front, rgBSPFace <T> &back,
	void (*interp_fn)( rgBSPVertex <T> &, const rgBSPVertex <T> &,
	const rgBSPVertex <T> &, const float));

template <class T>
int rgBSP_partition_face_world(
	std::vector <rgBSPFace <T> > &f,
	std::vector <rgBSPFace <T> > &front,
	std::vector <rgBSPFace <T> > &back,
	std::vector <rgBSPFace <T> > &center,
	const rgVector pp,
	const rgVector pn,
	void (*interp_fn)( rgBSPVertex <T> &, const rgBSPVertex <T> &,
		const rgBSPVertex <T> &, const float));

bool rgBSP_edge_can_be_split( rgVector pp, rgVector pn, rgVector a,
	rgVector b, const float epsilon, float &t, float &k);

bool rgBSP_point_is_same( const rgVector &a, const rgVector &b);


void rgBSP_NTTC_interp_fn(
	rgBSPVertex <rgBSP_NTTC_ExtData> &sp,
	const rgBSPVertex <rgBSP_NTTC_ExtData> &srcA,
	const rgBSPVertex <rgBSP_NTTC_ExtData> &srcB,
	const float t);

void rgBSP_NTTC_convex_hull_renderer(
	std::vector <rgBSPFace <rgBSP_NTTC_ExtData> > &p_f,
	rgMatrix *p_m,
	void *edata);

#endif
