/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <rgBSP.h>
#include <rgOpenGL.h>

template <class T>
int hlp_build_bsp_tree( std::vector <rgBSPFace <T> > &f, rgBSPNode <T> *&n,
	void (*interp_fn)( rgBSPVertex <T> &, const rgBSPVertex <T> &,
		const rgBSPVertex <T> &, const float));

template <class T>
void rgBSPFace<T>::gen_normal()
{
	// FIXME FIXME FIXME: This is NOT the robust way to generate normals
	// 	from an arbitrary face. This MUST be corrected to a more
	//	robust method.
	if( v.size() >= 3);
	normal = (rgCalcTriangleNormal( v[0].v, v[1].v, v[2].v)).normalize();
}

template <class T>
void rgBSPFace<T>::transform_vertices(
	void (*fn)( std::vector <rgBSPVertex <T> *> &, void *), void *p_data)
{
	if( !fn) return;
	if( v.empty()) return;
	std::vector <rgBSPVertex <T> *> tmp;
	for( int i = 0; i < int( v.size()); i++) tmp.push_back( &(v[i]));
	fn( tmp, p_data);
}

template <class T>
rgBSPNode<T>::rgBSPNode()
{
	front = 0;
	back = 0;
}

template <class T>
void rgBSPNode<T>::transform_face_verts(
	void (*fn)( std::vector <rgBSPVertex <T> *> &, void *), void *p_data)
{
	if( !fn) return;
	for( int i = 0; i < int( f.size()); i++)
		(f[i]).transform_vertices( fn, p_data);
}

template <class T>
rgBSPTree<T>::rgBSPTree()
{
	root = 0;
	use_bsp_matrix = false;
	extra_render_data = 0;
	convex_hull_renderer = 0;
	interp_fn = 0;
}

template <class T>
rgBSPTree<T>::~rgBSPTree()
{
	clear();
}

template <class T>
void rgBSPTree<T>::clear()
{
	if( root) { delete_tree( root); root = 0;}
	use_bsp_matrix = false;
	bsp_matrix.loadIdentity();
	extra_render_data = 0;
	convex_hull_renderer = 0;
	interp_fn = 0;
}

template <class T>
void rgBSPTree<T>::print_tree( const int tab, const rgBSPNode <T> *n) const
{
	if( !n) return;
	for( int i = 0; i < tab; i++) std::cerr << "\t";
	std::cerr << "[" << n->f.size() << ", pp = " << n->plane_point
		<< ", pn = " << n->normal;
	if( !(n->front) && !(n->back)) std::cerr << ", leaf";
	std::cerr << "]\n";
	print_tree( tab + 1, n->front);
	print_tree( tab + 1, n->back);
}

template <class T>
void rgBSPTree<T>::transform_node_verts(
	void (*fn)( std::vector <rgBSPVertex <T> *> &, void *),
	void *p_data,
	rgBSPNode <T> *node)
{
	if( !fn) return;
	if( !node) return;
	transform_node_verts( fn, p_data, node->front);
	node->transform_face_verts( fn, p_data);
	transform_node_verts( fn, p_data, node->back);
}

template <class T>
void rgBSPTree<T>::default_convex_hull_renderer(
	std::vector <rgBSPFace <T> > &p_f,
	rgMatrix *m, void *edata)
{
	for( int i = 0; i < int( p_f.size()); i++) {
		glBegin( GL_POLYGON);
		for( int j = 0; j < int( p_f[i].v.size()); j++)
			glVertex3fv( (float *)&(p_f[i].v[j].v));
		glEnd();
	}
}

template <class T>
void rgBSPTree<T>::render_node( rgBSPNode <T> *node)
{
	// If this is a leaf node, then render the faces and return.
	// NOTE: Leaf nodes store convex hulls, so the faces can be
	// rendered in any order.
	//
	// If this is an interior node, then it stores planar (possibly
	// double sided) faces. These can be rendered in any order.
	// FIXME: Incidently, the "convex hull renderer" can render both
	//	types, perhaps the rendering function should be renamed then.
	if( !(node->f.empty())) {
		// Use the given convex hull renderer (if given).
		if( convex_hull_renderer) {
			if( use_bsp_matrix) {
				convex_hull_renderer( node->f, &bsp_matrix,
					extra_render_data);

			} else {
				convex_hull_renderer( node->f, 0,
					extra_render_data);
			}

		// Use the default convex hull renderer if a custom renderer
		// was not given.
		} else {
			if( use_bsp_matrix) {
				default_convex_hull_renderer( node->f,
					&bsp_matrix, extra_render_data);

			} else {
				default_convex_hull_renderer( node->f, 0,
					extra_render_data);
			}
		}
	}
}

template <class T>
void rgBSPTree<T>::render_back_to_front( const rgVector p, rgBSPNode <T> *node)
{
	// NOTE: The rendering order changes based upon which side the
	//      point "p" is with respect to the node's plane.
	if( !node) return;
	if( node->normal.dot( p - node->plane_point) >= 0.0f) {
		render_back_to_front( p, node->back);
		render_node( node);
		render_back_to_front( p, node->front);

	} else {
		render_back_to_front( p, node->front);
		render_node( node);
		render_back_to_front( p, node->back);
	}
}

template <class T>
void rgBSPTree<T>::render_front_to_back( const rgVector p, rgBSPNode <T> *node)
{
	// NOTE: The rendering order changes based upon which side the
	//      point "p" is with respect to the node's plane.
	if( !node) return;
	if( node->normal.dot( p - node->plane_point) >= 0.0f) {
		render_front_to_back( p, node->front);
		render_node( node);
		render_front_to_back( p, node->back);

	} else {
		render_front_to_back( p, node->back);
		render_node( node);
		render_front_to_back( p, node->front);
	}
}

template <class T>
bool rgBSP_face_splits_face( const rgBSPFace <T> &src,
	const rgBSPFace <T> &tar, const float epsilon, int &side)
{
	if( src.v.empty()) return false;
	return rgBSP_plane_splits_face( src.v[0].v, src.normal, tar,
		epsilon, side);
}

template <class T>
bool rgBSP_plane_splits_face( rgVector pp, rgVector pn,
	const rgBSPFace <T> &tar, const float epsilon, int &side)
{
	// Inorder for a split to occur, at least one vertex from "tar"
	// MUST exist on BOTH sides of the plane defined by "src".
	//
	// NOTE: Vertices within the epsilon range (both positive and
	//	negative) are considered to be planner and thus are neither
	//	in front nor behind the plane defined by "src".

	int front_count = 0;
	int back_count = 0;
	float t;
	for( int i = 0; i < int( tar.v.size()); i++) {
		t = RGBSP_POINT_PLANE_DISTANCE( pp, pn, tar.v[i].v);
		if( fabs( t) > epsilon) {
		//	std::cerr << "t = " << t << ", v = " << tar.v[i].v
		//		<< ", pp = " << pp << ", pn = " << pn << "\n";
			if( t > 0.0f) front_count++;
			else back_count++;
		}
	}

	if( front_count && back_count) return true;

	// The face is not split, determine which side the target face is
	// on.
	if( front_count) side = RGBSP_FACE_ON_FRONT;
	else if( back_count) side = RGBSP_FACE_ON_BACK;
	else side = RGBSP_FACE_IS_PLANAR;
	return false;
}

template <class T>
void hlp_gen_split_point( rgVector pp, rgVector pn, rgBSPVertex <T> &sp, int a,
	int b, const int count, const rgBSPFace <T> &f,
	void (*interp_fn)( rgBSPVertex <T> &, const rgBSPVertex <T> &,
		const rgBSPVertex <T> &, const float))
{
	if( ((a + 1) % count) == b) {
		// Must intersect edge to generate point.
		float t;
		rgLineSegPlaneIntersect( pp, pn, f.v[a].v, f.v[b].v, sp.v, t);
		if( interp_fn) interp_fn( sp, f.v[a], f.v[b], t);

	} else {
		// Choose the center vertex as the splitting edge.
		// Choose the middle index between "a" and "b" (remember that
		// this is a circular array) as the split point.
		int i = 0;
		int tmp = a;
		while( a != b) { a = (a + 1) % count; i++;}

		i >>= 1;	// Divide by 2
		while( i--) tmp = (tmp + 1) % count;

		// "tmp" is now the middle index (center vertex).
		sp = f.v[ tmp];
	}
}

#define TMP_PLANAR	-1
#define TMP_FRONT	0
#define TMP_BACK	1
template <class T>
void hlp_construct_sub_faces( rgVector pp, rgVector pn, char *v,
	const int count, rgBSPFace <T> &f, rgBSPFace <T> &front,
	rgBSPFace <T> &back,
	void (*interp_fn)( rgBSPVertex <T> &, const rgBSPVertex <T> &,
		const rgBSPVertex <T> &, const float))
{
	// Find the first front vertex.
	int ffv = 0;
	while( TMP_FRONT != v[ffv]) ffv++;
	while( TMP_FRONT == v[(ffv + count - 1) % count])
		ffv = (ffv + count - 1) % count;

	// Find the last front vertex.
	int lfv = ffv;
	while( TMP_FRONT == (v[(lfv + 1) % count])) lfv = (lfv + 1) % count;

	// Find the first back vertex.
	int fbv = lfv;
	while( TMP_BACK != v[fbv]) fbv = (fbv + 1) % count;

	// Find the last back vertex.
	int lbv = fbv;
	while( TMP_BACK == (v[(lbv + 1) % count])) lbv = (lbv + 1) % count;


	// Determine the split points now.
	rgBSPVertex <T> spa;
	rgBSPVertex <T> spb;
	hlp_gen_split_point( pp, pn, spa, lbv, ffv, count, f, interp_fn);
	hlp_gen_split_point( pp, pn, spb, lfv, fbv, count, f, interp_fn);

	// Construct the two sub faces.
	// Construct the front face.
	front.normal = f.normal;
	front.m_tid = f.m_tid;
	front.v.clear();
	front.v.push_back( spa);
	while( ffv != lfv) {
		front.v.push_back( f.v[ffv]);
		ffv = (ffv + 1) % count;
	}
	front.v.push_back( f.v[lfv]);
	front.v.push_back( spb);

	// Construct the back face.
	back.normal = f.normal;
	back.m_tid = f.m_tid;
	back.v.clear();
	back.v.push_back( spb);
	while( fbv != lbv) {
		back.v.push_back( f.v[fbv]);
		fbv = (fbv + 1) % count;
	}
	back.v.push_back( f.v[lbv]);
	back.v.push_back( spa);
}

template <class T>
void rgBSP_split_convex_face( rgVector pp, rgVector pn, rgBSPFace <T> &f,
	rgBSPFace <T> &front, rgBSPFace <T> &back,
	void (*interp_fn)( rgBSPVertex <T> &, const rgBSPVertex <T> &,
		const rgBSPVertex <T> &, const float))
{
//	front.m_tid = f.m_tid;
//	back.m_tid = f.m_tid;

	// Inorder for a split to occur, at least one vertex from "f"
	// MUST exist on BOTH sides of the splitting plane.

	// This is an array of vertex locations. It is used to determine
	// where (if any) split points should be generated.
	rgTempArray <char> v( f.v.size());

	// These flags determine if the face was split at all. If BOTH "fvert"
	// and "bvert" are true, then the face was split.
	bool fvert = false;
	bool bvert = false;

	// Determine the location of all verticies (based upon the splitting
	// plane).
	float t;
	const int istop = int( f.v.size());
	for( int i = 0; i < istop; i++) {
		t = RGBSP_POINT_PLANE_DISTANCE( pp, pn, f.v[i].v);
		if( fabs( t) > RGBSP_FACE_SPLIT_EPSILON) {
			if( t > 0.0f) { v.data[i] = TMP_FRONT; fvert = true;}
			else { v.data[i] = TMP_BACK; bvert = true;}
		} else v.data[i] = TMP_PLANAR;
	}

	// If the face was not split, then place it in the appropriate
	// side and return. If the face is planar, then it is placed in the
	// front list. (A face is not split if at least one vertex does not
	// occupy BOTH sides of the splitting plane.
	if( !fvert || !bvert) {
		if( bvert) back = f;
		else front = f;
		return;
	}

	// Construct the two resulting sub faces (front and back).
	hlp_construct_sub_faces( pp, pn, v.data, istop, f, front, back,
		interp_fn);
}
#undef TMP_PLANAR
#undef TMP_FRONT
#undef TMP_BACK

template <class T>
bool hlp_non_splitting( const int id, const std::vector <rgBSPFace <T> > &f)
{
	// The face "id" is non-splitting if it does not split any other face.
	int temp;
	for( int i = 0; i < int( f.size()); i++) {
		if( (i != id) && rgBSP_face_splits_face( f[id], f[i],
			RGBSP_FACE_SPLIT_EPSILON, temp)) return false;
	}

	// The face "id" does not split any other face.
	return true;
}

template <class T>
bool rgBSP_axis_aligned_face( const rgBSPFace <T> &f)
{
	// The face is axis aligned if it's normal is parallel to one of the
	// standard coordinate axies (X, Y, Z).
	if(	fabs( (rgVector( 1, 0, 0)).dot( f.normal) == 0) ||
		fabs( (rgVector( 0, 1, 0)).dot( f.normal) == 0) ||
		fabs( (rgVector( 0, 0, 1)).dot( f.normal)) == 0) return true;
	return false;
}

template <class T>
int rgBSP_choose_splitting_face( const std::vector <rgBSPFace <T> > &f)
{
	// Check parameters.
	if( f.empty()) return -1;

	// Seperate the faces that split, from the faces that do not split
	// other faces.	
	std::vector <int> splitting;
	std::vector <int> non_splitting;
	int i;
	for( i = 0; i < int( f.size()); i++) {
		if( hlp_non_splitting( i, f)) non_splitting.push_back( i);
		else splitting.push_back( i);
	}

//	std::cerr << "split = " << splitting.size() << ", non = " <<
//		non_splitting.size() << "\n"; 

	// XXX: Faces MUST partition the world (as per the function name).
	//	If a non-paritioning face is returned, then BSP generation
	//	will recurse forever.
	// Out of the non splitting faces, choose the face that is axis
	// aligned. If there are more than one axis aligned non splitting
	// faces, then choose the first face encountered.
	for( i = 0; i < int( non_splitting.size()); i++) {
		if(	rgBSP_axis_aligned_face( f[ non_splitting[i]]) &&
			rgBSP_face_partitions_world( non_splitting[i], f)) {
//			std::cerr << "non-split axis part = " <<
//				non_splitting[i] << ", (" << i << ")\n";
			return non_splitting[i];
		}
	}

	// If there are non-splitting faces, and they are all non-axis
	// aligned, then just return the first non-splitting face.
	if(	!(non_splitting.empty()) &&
		rgBSP_face_partitions_world( non_splitting[0], f)) {
//		std::cerr << "non-split non-axis part = " <<
//			non_splitting[0] << "\n";
		return non_splitting[0];
	}

	// There are no non-splitting faces, therefore return the axis
	// aligned face out of the splitting faces. If there are more than
	// one, then just return the first face encountered.
	for( i = 0; i < int( splitting.size()); i++) {
		if( rgBSP_axis_aligned_face( f[ splitting[i]])) {
//			std::cerr << "split part = " << splitting[i] << "\n";
			return splitting[i];
		}
	}

	// FIXME: Throw an exception if "splitting" is empty. This is because
	//	the function name implies that there MUST be a splitting face
	//	to choose. If not, then some other function earlier in the
	//	BSP process messed-up (ie. rgBSP_world_is_convex). If a
	//	splitting face cannot be returned, then an infinite recursion
	//	will ensue.
	if( splitting.empty()) return -1;

	// There are no axis aligned faces at all, therefore just return
	// the first face in the split list.
//	std::cerr << "split default = " << splitting[0] << "\n";
	return splitting[0];
}

// Description: Partitions the given world "f" into a "front" and "back" world.
//	"f" will be emptied of all faces, and "front" and "back" together will
//	contain all faces from the world "f".
template <class T>
int rgBSP_partition_face_world(
	std::vector <rgBSPFace <T> > &f,
	std::vector <rgBSPFace <T> > &front,
	std::vector <rgBSPFace <T> > &back,
	std::vector <rgBSPFace <T> > &center,
	const rgVector pp,
	const rgVector pn,
	void (*interp_fn)( rgBSPVertex <T> &, const rgBSPVertex <T> &,
		const rgBSPVertex <T> &, const float))
{
	int splits = 0;
	int side;
	while( !(f.empty())) {
		// If the face would be split, then place both parts on the
		// appropriate side.
		if(	rgBSP_plane_splits_face( pp, pn, f[0],
				RGBSP_FACE_SPLIT_EPSILON, side)) {

			rgBSPFace <T> ftmp;
			rgBSPFace <T> btmp;
			rgBSP_split_convex_face( pp, pn, f[0], ftmp, btmp,
				interp_fn);
			if( !ftmp.v.empty()) front.push_back( ftmp);
			if( !btmp.v.empty()) back.push_back( btmp);
			if( !(ftmp.v.empty() || btmp.v.empty())) splits++;

		// The face has not been split. Therefore, determine which
		// side of the splitting plane it should be on.
		} else {

			switch( side) {
			// NOTE: Planar faces are considered to be infront.
			case RGBSP_FACE_IS_PLANAR:
				center.push_back( f[0]);
				break;

			case RGBSP_FACE_ON_FRONT:
				front.push_back( f[0]);
				break;

			case RGBSP_FACE_ON_BACK:
				back.push_back( f[0]);
				break;
			}
		}

		// The face has been partitioned, so remove it.
		f.erase( f.begin());
	}
	return splits;
}

template <class T>
int hlp_build_bsp_tree( std::vector <rgBSPFace <T> > &f, rgBSPNode <T> *&n,
	void (*interp_fn)( rgBSPVertex <T> &, const rgBSPVertex <T> &,
		const rgBSPVertex <T> &, const float))
{
	if( f.empty()) return 0;
	n = new rgBSPNode <T> ();

	// If all of the faces in "f" form a convex region, then make this
	// node a "leaf" node. There is no need to split faces since the
	// entire region is convex.
	if( rgBSP_world_is_convex( f)) {
		n->f = f;
		n->plane_point = f[0].v[0].v;
		n->normal = f[0].normal;
		f.clear();
		return 0;
	}

	// Select a splitting face that will be used as the splitting plane.
	const int sfid = rgBSP_choose_splitting_face( f);
	// FIXME: "sfid" should NEVER be less than 0. If so, then a function
	//	earlier in the BSP process messed-up.
	if( sfid < 0) {
		n->f = f;
		n->plane_point = f[0].v[0].v;
		n->normal = f[0].normal;
		f.clear();
		return 0;
	}
	const rgVector pp( f[sfid].v[0].v);
	const rgVector pn( f[sfid].normal);


	// The face used to create the splitting plane will always be planar
	// to the splitting plane. Therefore, add it to the "center" list.
	// Also, remove it from the face list "f" since it has already been
	// added to "front".
	n->f.push_back( f[sfid]);
	f.erase( f.begin() + sfid);

	// Parition the remaining faces using the newly obtained splitting
	// plane.
	std::vector <rgBSPFace <T> > front;
	std::vector <rgBSPFace <T> > back;
	int splits = rgBSP_partition_face_world( f, front, back, n->f, pp, pn,
		interp_fn);
	
	// Recursively build the "front" and "back" worlds.
	n->plane_point = pp;
	n->normal = pn;
	return	(splits +
		hlp_build_bsp_tree( front, n->front, interp_fn) +
		hlp_build_bsp_tree( back, n->back, interp_fn));
}

template <class T>
bool rgBSP_world_is_convex( const std::vector <rgBSPFace <T> > &f)
{
	// For EVERY face, the ENTIRE world must be either inside or outside
	// (but not both). So, if a face must split another face, then the
	// world is not convex.
	for( int i = 0; i < int( f.size()); i++)
		if( rgBSP_face_partitions_world( i, f)) return false;
	return true;
}

template <class T>
bool rgBSP_face_partitions_world( const int id,
	const std::vector <rgBSPFace <T> > &f)
{
	// For EVERY face, the ENTIRE world must be either inside or outside
	// (but not both). So, if a face must split another face, then the
	// world is not convex.
	int prev_side = RGBSP_FACE_IS_PLANAR;
	for( int j = 0; j < int( f.size()); j++) {
		int side = RGBSP_FACE_IS_PLANAR;
		if( (id != j) && rgBSP_face_splits_face( f[id], f[j],
			RGBSP_FACE_SPLIT_EPSILON, side))
			return true;

		if(	(RGBSP_FACE_ON_FRONT == side) &&
			(f[id].normal.dot( f[j].normal) > 0.0f))
			return true;
	
		// If the previous face was planar, then this face can
		// be on any side and the world is still convex.
		// However, once a side has been selected (other than
		// planar), ALL other faces MUST be on the same side,
		// otherwise the world is not convex.
		if( side != RGBSP_FACE_IS_PLANAR) {
			if( prev_side != RGBSP_FACE_IS_PLANAR) {
				// Previous face was not planar, so
				// this face MUST be on the same side.
				if( prev_side != side) return true;

			// Pick a side and stick with it.
			} else prev_side = side;
		}
	
	}
	return false;
}






















