/*
Copyright (C) 2003-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef RGBEZIER_H
#define RGBEZIER_H

//#include "rgStandard.h"

class rgVector;

/*
 * Description: Takes a list/array of verticies and produces a "smooth" Bezier
 *	curve. The Bezier curve produced approximates the "smooth" curve
 *	formed by the given list/array of verticies. The number of points in
 *	the Bezier curve is returned or a integer less then 0 is returned if
 *	an error occured. If 0 is returned, then bPoints will be set to NULL
 * 	which means that it does not point to allocated memory (since there
 *	were no Bezier points generated).
 *
 * Parameters:
 *
 * 	count - Number of verticies (control points) in cPoints. The valid
 *		range of count is >= 0.
 *
 *	cPoints - A pointer to the list/array of control points.
 *
 *	bPoints - A pointer which will contain the Bezier produced (if no error
 *		occurs). The memory to store all of the points will be allocated
 *		by rgGenBezier. So when the caller is finished with the Bezier
 *		points, they should call "delete [] bPoints" (unless bPoints
 *		was returned equal to NULL). The caller should not allocate
 *		memory with bPoints since rgGenBezier will itself use it to
 *		allocate memory.
 *
 *	level - The "smoothness" of the Bezier curve produced.. The valid range
 *		for level is >= 0. A level of 0 produces a Bezier curve that is
 *		exactly identical to the curve represented by cPoints. 
 *
 *	degree - The degree of the Bezier curve produced. Valid values are 1, 2
 *		and 3.
 *
 * Returns: The number of Bezier points generated is returned. A negative value
 *	is returned if an error occured.
 *
 */
int rgGenBezier( int count, rgVector *cPoints, const int cStep,
	rgVector *&bPoints, const int bStep, int level, int degree,
	bool autoAlloc_bPoints);

/*
 * Description: Returns the total number of points produced by the Bezier
 *	algorithm on a given number of points for a specified level.
 */
int rgBezierPointCount( const int pCount, const int level);

#endif
