/*
Copyright (C) 2003-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "rgCSG.h"
#include <rgOpenGL.h>

std::ostream &operator <<( std::ostream &o, const rgCSGBrush *b)
{
	if( !b) return o;
	return (b->print( o));
}

int rgCSGBrush::save_data( std::ostream &des) const
{
	write_dword( des, rgCSGBrush::version());
	m_pos.write_binary( des);
	write_dword( des, m_brush_flags);
	write_dword( des, m_operation);
	write_dword( des, m_tex_id);
	write_dword( des, m_group_id);
	return 0;
}

int rgCSGBrush::load_data( std::istream &src)
{
	// XXX: Read in the version number and load appropriate data.
	switch( read_dword( src)) {
	case 0: return load_version_0_data( src);
	default: return -1;
	}
}

int rgCSGBrush::load_version_0_data( std::istream &src)
{
	m_pos.read_binary( src);
	m_brush_flags = read_dword( src);
	m_operation = read_dword( src);
	m_tex_id = read_dword( src);
	m_group_id = read_dword( src);
	return 0;
}

rgCSGCube::rgCSGCube()
{
	m_display_list = 0;
	m_v[0] = rgVector( 1, 1, 1);
	m_v[1] = rgVector( 1, -1, 1);
}

void rgCSGCube::clear()
{
	if( m_display_list) {
		glDeleteLists( m_display_list, 1);
		m_display_list = 0;
	}
}

void rgCSGCube::refresh()
{
	if( !m_display_list) return;
	glNewList( m_display_list, GL_COMPILE);
	glBegin( GL_QUADS);
	real_render();
	glEnd();
	glEndList();
}

void rgCSGCube::draw()
{
	if( m_display_list) { glCallList( m_display_list); return;}

	m_display_list = glGenLists( 1);
	if( !m_display_list) {
		glBegin( GL_QUADS);
		real_render();
		glEnd();
		return;
	}
	refresh();
}

void rgCSGCube::real_draw_cursor( const rgVector vc, const rgVector ec)
{
	glPolygonMode( GL_FRONT_AND_BACK, GL_LINE);
	glColor3f( ec.x, ec.y, ec.z);
	rgCSGCube::draw();

	glBegin( GL_POINTS);
	glColor3f( vc.x, vc.y, vc.z);
	real_render();
	glVertex( rgCSGCube::position());
	glEnd();
}

void rgCSGCube::real_render() const
{
	rgVector v[8];
	gen_verts( v);
	#define DRAW_QUAD(a,b,c,d) \
		glNormal( (rgCalcTriangleNormal( v[a], v[b], v[c])). \
			normalize()); \
		glVertex( v[a]); glVertex( v[b]); glVertex( v[c]); \
		glVertex( v[d])
	DRAW_QUAD( 0, 1, 2, 3);	// Top
	DRAW_QUAD( 4, 5, 6, 7);	// Bottom
	DRAW_QUAD( 0, 3, 5, 4);	// Left
	DRAW_QUAD( 2, 1, 7, 6);	// Right
	DRAW_QUAD( 3, 2, 6, 5);	// Front
	DRAW_QUAD( 0, 4, 7, 1);	// Back
	#undef DRAW_QUAD
}

void rgCSGCube::gen_verts( rgVector *v) const
{
	std::vector <rgVector> tmp;
	gen_verts( tmp);
	v[0] = tmp[0];
	v[1] = tmp[1];
	v[2] = tmp[2];
	v[3] = tmp[3];
	v[4] = tmp[4];
	v[5] = tmp[5];
	v[6] = tmp[6];
	v[7] = tmp[7];
}

void rgCSGCube::gen_verts( std::vector <rgVector> &v) const
{
	map_verts( v, m_pos);
}

void rgCSGCube::map_verts( std::vector <rgVector> &v, const rgCoordSystem &cs)
	const
{
	v.clear();
	const rgVector &p = cs.p;
	const rgVector &a = m_v[0];//( m_v[0] - p);
	const rgVector &b = m_v[1];//( m_v[1] - p);
	// 0--1   |
	// |\ |\  |
	// | 3--2 |
	// 4-|7 | |
	//  \| \| |
 	//   5--6 |
	const rgVector &s = cs.side;
	const rgVector &u = cs.up;
	const rgVector &e = cs.eye;
	v.push_back( p + (s * -a.x) + (u * a.y) + (e * -a.z));
	v.push_back( p + (s * a.x) + (u * a.y) + (e * -a.z));
	v.push_back( p + (s * a.x) + (u * a.y) + (e * a.z));
	v.push_back( p + (s * -a.x) + (u * a.y) + (e * a.z));
	v.push_back( p + (s * -b.x) + (u * b.y) + (e * -b.z));
	v.push_back( p + (s * -b.x) + (u * b.y) + (e * b.z));
	v.push_back( p + (s * b.x) + (u * b.y) + (e * b.z));
	v.push_back( p + (s * b.x) + (u * b.y) + (e * -b.z));
}

int rgCSGCube::save_data( std::ostream &des) const
{
	write_dword( des, rgCSGCube::version());
	(m_v[0]).write_binary( des);
	(m_v[1]).write_binary( des);
	rgCSGBrush::save_data( des);
	return 0;
}

int rgCSGCube::load_data( std::istream &src)
{
	// XXX: Read in the version number and load appropriate data.
	switch( read_dword( src)) {
	case 0: return load_version_0_data( src);
	default: return -1;
	}
}

int rgCSGCube::load_version_0_data( std::istream &src)
{
	(m_v[0]).read_binary( src);
	(m_v[1]).read_binary( src);
	rgCSGBrush::load_data( src);
	return 0;
}



rgCSGWorld::rgCSGWorld()
{
}

rgCSGWorld::~rgCSGWorld()
{
	clear();
}

void rgCSGWorld::clear()
{
	while( !(m_brush.empty())) {
		const int last_id = int( m_brush.size() - 1);
		if( (m_brush[ last_id])) delete (m_brush[ last_id]);
		m_brush.erase( m_brush.begin() + last_id);
	}
}

int rgCSGWorld::add_brush( rgCSGBrush *b)
{
	if( !b) return -1;
	m_brush.push_back( b);
	return (brush_count() - 1);
}

void rgCSGWorld::delete_brush( const int bid)
{
	if( (bid < 0) || (bid >= brush_count())) return;
	m_brush.erase( m_brush.begin() + bid);
}

void rgCSGWorld::delete_brush( const rgCSGBrush *b)
{
	if( !b) return;
	for( int i = 0; i < brush_count(); i++) {
		if( b == (m_brush[i])) {
			delete_brush( i);
			return;
		}
	}
}

int rgCSGWorld::select_brush( rgVector p, const rgVector n,
	const float min_dis) const
{
	p = rgMakePlanar( p, n);
	int sel = -1;
	float dis = 16777816.0f;
	for( int i = 0; i < brush_count(); i++) {
		if( (m_brush[i])->selectable()) {
			const float d = (p - rgMakePlanar(
				(m_brush[i])->position(), n)).length();
			if( (d < min_dis) && (d < dis)) {
				sel = i;
				dis = d;
			}
		}
	}
	return sel;
}
