/*
Copyright (C) 2003-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

/*
 *
 * Search this file for the string "EXTENDING THE OBJ LOADER" to see how to
 * easily add new commands to the obj parser/loader.
 *
 */

#include <stdio.h>
#include <rgError.h>
#include <iostream>
#include "rgOBJ.h"
#include "rgOBJ_parser.h"

/*****************************************************************************
 * OBJ command handler functions and the corresponding vector table.
 *
 * EXTENDING THE OBJ LOADER (adding new command handlers):
 *
 *	1. Add a function prototype for the new command. The handler function
 *	   prototype section is located in "rgOBJ_parser.h". For example, to
 *	   add the "example" handler function, you would write:
 *
 *	   "void hf_example( Obj_parse_data *p);"
 *
 *	2. Add the command's name and handler function to the command handler
 *	   vector table below. The last entry of the table MUST ALWAYS be
 *	   { 0,	0 }. Adding your command to the vector table is the only
 *	   modification you "need" to make to this file.
 *
 *	3. Add the implementation for the handler function you prototyped from
 *	   step 1. The primary handler function implementations are placed in
 *	   the file "rgOBJ_hf.cpp". You can place your implementations in
 *	   "rgOBJ_hf.cpp" or in another source file (you are encouraged to
 *	   place your implementations in a seperate file).
 *
 *	TIPS:	- Always check for errors where appropriate in your handler.
 *		  If you detect an error, report it to the main parser/loader
 *		  by setting the "err" member of the "obj_parse_data *p"
 *		  parameter. If the error is serious/fatal, then also set the
 *		  "done" member of "obj_parser_data *p" to true (this will
 *		  cause the main parser/loader to terminate immediatley after
 *		  your handler returns.
 *
 *		- Use the "hf_unknown" handler to ignore any command.
 *
 *		- Look at how some of the other function handlers were
 *		  implemented. This will give you an idea on how to write your
 *		  own implementation.
 *
 *****************************************************************************/

struct obj_command_handler
{
        char *cname;
        void (* fn)(Obj_parse_data *p);
};

obj_command_handler obj_commands[] =
{
	/* NOTES: Commands must be in alphabetical (ascii) order. It is ok to
	 *	place them out of order since they will be sorted any way at
	 *	runtime. Having a sorted list of commands allows for faster
	 *	command parsing when loading ".obj" files.
	 *
	 * IMPORTANT: If a command has a NULL handler function, then the obj
	 *	parser will act as if the command was not read (i.e. it will
	 *	discard the command's text but not the entire line from the
	 *	file).
	 *
	 */
	/* TODO: Write sorting routine for this function handler table. */

	/********** COMMAND HANDLER VECTOR TABLE **********/
	/* Cmd name	Handler function */
	{ "g",		hf_group},	/* Group */
	{ "v",		hf_vertex},	/* Vertex */
	{ "vn",		hf_normal},	/* Vertex normal */
	{ "vt",		hf_texcoord},	/* Vertex texture coord */
	{ "f",		hf_face},
	{ 0,		0}		/* END OF LIST */
};

Obj_parse_data::Obj_parse_data()
{
	inf = 0;
	obj = 0;
	cmd = new char[ RGOBJ_MAX_COMMAND_LENGTH + 1];
	cmd_len = 0;

	m_negative_vertex_offset = 0;
	m_negative_normal_offset = 0;
	m_negative_texcoord_offset = 0;
}

Obj_parse_data::~Obj_parse_data()
{
	delete [] cmd;
}

rgOBJ::rgOBJ()
{
	fail_on_errors = true;
}

rgOBJ::~rgOBJ()
{
	clear();
}

void rgOBJ::clear()
{
	m_vert.clear();
	m_normal.clear();
	m_texcoord.clear();	
}

bool rgOBJ::face_is_good( rgOBJFace &f) const
{
	if( f.m_indices.size() < 3) return false;
	for( int i = 0; i < int( f.m_indices.size()); i++) {
		if( f.m_indices[i].m_vid < 0) return false;
		if( f.m_indices[i].m_vid >= int( m_vert.size()))
			return false;

		if( f.m_indices[i].m_nid >= int( m_normal.size()))
			return false;

		if( f.m_indices[i].m_tcid >= int( m_texcoord.size()))
			return false;
	}
	return true;
}

/*************************************************************************
 * The main OBJ parser/loader begins.
 *************************************************************************/
int rgOBJ::load( const char *fname)
{
	if( !fname) return RGERR_BAD_PARAM;
        std::ifstream inf( fname, std::ios::binary);
	if( !(inf.is_open())) return RGERR_COULD_NOT_OPEN_FILE; 
	return load( inf);
}

int rgOBJ::load( std::istream &inf)
{
	/* Remove the previous OBJ model (if any). */
	clear();

	/* Initialize parsing data. */
	Obj_parse_data d;
	d.inf = &inf;
	d.obj = this;
	d.line = 1;
	d.done = false;

	int cid;	/* Command identifier */
	while( !d.done) {
	
		switch( (cid = get_command( inf, d.cmd, d.cmd_len))) {
		case CM_FILE_ERROR:
			m_emsg = "Line " + rg_to_string( d.line) +
				": get_command() got a file read error.";
			d.err = RGERR_FAIL;
			d.done = true;
			break;

		case CM_TOO_LONG:
			m_emsg = "Line " + rg_to_string( d.line) +
				": Command is too long.";
			d.err = RGERR_FAIL;
			d.done = true;
			break;

		/**/
		case CM_EOF: d.done = true; break;

		/* No command existed on the line. */
		case CM_NONE: break;

		/* OBJ comment read, so skip the entire line. */
		case CM_COMMENT:
			hf_comment( &d);
			break;

		case CM_UNKNOWN:
			hf_unknown( &d);
			break;

		/* Command read is recognized. */
		default:
			/* Execute the command. */
			if( obj_commands[ cid].fn) {
				d.err = 0;
				obj_commands[ cid].fn( &d);
				if( fail_on_errors && d.err) d.done = true;
			}
			break;
		}
	}

	if( d.err) return RGERR_FAIL;
	return RGERR_SUCCESS;
}

void rgOBJ::remove_duplicate_vertices()
{
	// Create list of unique vertices.
	std::vector <rgOBJVertex> unique;
	std::vector <int> vid_map;
	{
		for( int i = 0; i < int( m_vert.size()); i++) {
			bool is_duplicate = false;
			for( int j = 0; j < int( unique.size()); j++) {
				if( m_vert[i] == unique[j]) {
					is_duplicate = true;
					vid_map.push_back( j);
					break;
				}
			}
			if( !is_duplicate) {
				vid_map.push_back( int( unique.size()));
				unique.push_back( m_vert[i]);
			}
		}
	}

	// Remap all faces to use unique vertices
	for( int i = 0; i < int( m_group.size()); i++) {
		for( int j = 0; j < int( m_group[i].m_face.size()); j++) {
			rgOBJFace &f = m_group[i].m_face[j];
			for( int k = 0; k < int( f.m_indices.size()); k++) {
				f.m_indices[k].m_vid =
					vid_map[ f.m_indices[k].m_vid];
			}
		}
	}

	vid_map.clear();
	m_vert.clear();
	m_vert = unique;
}

rgOBJFace *rgOBJ::get_face( const int id)
{
	if( id < 0) return 0;
	int count = 0;
	for( int i = 0; i < group_count(); i++) {
		if( id < (count + m_group[i].face_count()))
			return &(m_group[i].m_face[ id - count]);
		count += m_group[i].face_count();
	}
	return 0;
}

/*************************************************************************
 * OBJ parser utility functions begin.
 *************************************************************************/
int get_command( std::istream &inf, char *cmd, int &i)
{
	if( inf.eof()) return CM_EOF;

	i = 0;
	char c;
	bool command_char_read = false;
	bool done = false;
	while( !done) {
		/* Read a character from the file. */
		c = inf.get();

		/* Stop searching for a command if any I/O errors occur. */
		if( inf.fail()) {
			if( !inf.eof()) return CM_FILE_ERROR;
			done = true;
		}

		/* Check for special characters. */
		if( !done) {
			switch( c) {
			/* Comment characters. */
			case '#':
				if( 0 == i) return CM_COMMENT;
				inf.putback( c);
				done = true;
				break;
	
			/* White space*/	
			case ' ':
			case '\t':
			case '\r':
				if( command_char_read) {
					inf.putback( c);
					done = true;
				}
				break;
	
			case '\n':
				if( 0 == i) return CM_NONE;
				inf.putback( c);
				done = true;
				break;
	
			/* All other characters are valid in a command. */
			default:
				if( i >= RGOBJ_MAX_COMMAND_LENGTH)
					return CM_TOO_LONG;
				cmd[ i++] = c;
				command_char_read = true;
				break;
			}
		} /* if( !done) */
	} /* while( !done) */

	/*
	 * Return the status of the command read. If the command matches a
	 * predefined command, then the index number of the predefined command
	 * is returned. Otherwise, CM_UNKNOWN is returned.
	 */
	return match_command( cmd, i);
}

int match_command( char *cmd, const int len)
{
	int i = 0;
	obj_command_handler *p = obj_commands;
	while( p->cname) {
	
		/* If the current command matches, then return it. */
		if( 0 == strncmp( p->cname, cmd, len)) return i;

		/* Move on to the next command. */
		i++;
		p++;
	}

	return CM_UNKNOWN;
}

void end_line( std::istream &inf)
{
	/* TODO: Implement a faster method of skiping lines. */
	while( true) {
		if( '\n' == inf.get()) return;
		if( inf.fail()) return;
	}
}
