/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef RGPARSE_H
#define RGPARSE_H

#include <fstream>
#include <vector>
#include <rgStandard.h>

#define RGP_DEFAULT_WORD_BUF_LEN	1024

#define RGP_FLAG_ONE_LINE	0x1	/* Stop looking for a word if a newline
					 * character is read. Without this flag
					 * parsing functions will treat leading
					 * newline characters as white space.
					 * This flag is useful when parsing for
					 * words on a single line.
					 */
#define RGP_FLAG_NO_NEWLINE	0x2	/* If a newline character is read,
					 * regardless if it is leading white
					 * space or a word terminator, then
					 * parsing functions will treat this
					 * as a parsing error.
					 */

#define RGP_PNUMERIC_TYPE_INVALID	-1
#define RGP_PNUMERIC_TYPE_INT		0
#define RGP_PNUMERIC_TYPE_LONG		1
#define RGP_PNUMERIC_TYPE_FLOAT		2
#define RGP_PNUMERIC_TYPE_DOUBLE	3


class rgPCommandHandler;


class rgPNumeric
{
public:
	int type;

	int m_int;
	long m_long;
	rgFloat m_float;
	double m_double;

	bool is_integer();
	long get_integer();
	double get_float();
};

class rgParseContext
{
public:
	std::ifstream inf;
	int line;
	int error;

	bool case_sensitive;
	bool append_word_terminators;

	char terminating_char;

	void *data;

	/*
	 * "ch_fn" is a dynamic array of command handlers.
	 * "ch_fn" is made public for convinience. This variable should be
	 * treated as "read-only". To add or remove command handlers, use
	 * "add_command_handler" and "remove_command_handler".
	 */
	std::vector <rgPCommandHandler *> ch_fn;

	rgParseContext();
	rgParseContext( int wblen);
	~rgParseContext();

	int add_command_handler( const char *cname,
		int (*fn)( rgParseContext &p));

	void clear();

	char *get_word_buffer();

	int remove_command_handler( const int id);

	int get_word( char *des, const int len);
	int get_word( char *des, const int len, const int flags);
	int parse( const char *fname);

	int read_expected_char( const char c, bool one_line);

	int skip_white_space();
	int skip_white_space( const int flags);
	int skip_line();

	int set_string( char *des, const int len);
	int set_numeric( rgPNumeric &des);
	int set_numeric( rgPNumeric &des, const int flags);

	bool str_numeric_has_decimal_point( char *num);
	bool str_to_int( int &des, char *src, int base, char *&endptr);
	bool str_to_long( long &des, char *src, int base, char *&endptr);
	bool str_to_float( float &des, char *src, char *&endptr);
	bool str_to_double( double &des, char *src, char *&endptr);

	bool add_word_terminator_char( const char c);
	bool add_white_space_char( const char c);
	bool add_escape_char( const char c);


private:

	bool parse_lock;
	char *word;
	int max_word_length;

	/* All of the special character arrays are stored in
	 * "special_char_buf". This is more efficient then having seperate
	 * memory allocations for each array.
	 */
	char *special_char_buf;
	char *ws_chars;	/* White space character array */
	char *wt_chars;	/* Word terminator character array */
	char *ec_chars;	/* Escape character array */

	int ws_len;	/* White space character array length */
	int wt_len;	/* Word terminator character array length */
	int ec_len;	/* Escape character array length */

	void init( const int wblen);
	int process_command( const char *cname);
	bool add_special_char( char *&s, int &len, const char c);
};

class rgPCommandHandler
{
public:
	char *cname;
	int (*fn)( rgParseContext &);

	rgPCommandHandler( const char *p_cname,
		int (*p_fn)( rgParseContext &));

	rgPCommandHandler( const rgPCommandHandler &c);

	~rgPCommandHandler();

	rgPCommandHandler &operator =( rgPCommandHandler &c);

private:

	void copy_data( const rgPCommandHandler &c);
};

#endif
