/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef RGPORT_H
#define RGPORT_H

#define HOME_LINUX

#include <iostream>
#include <fstream>
#include <sstream>
#include <string>

/**************************************************
 *
 * IMPORTANT: The definitions RGPORT_LITTLE_ENDIAN, and RGPORTH_BIG_ENDIAN
 *	should be defined in a makefile or at compile time. This will prevent
 *	a user from moving this file to a machine of the wrong endian type and
 *	compiling the code for it. This way, the user is forced to select the
 *	machine type they are compiling for.
 *
 * Select the byte ordering of the target machine.
 *
 * Only ONE of the endian types can be picked at
 * any one time.
 *
 **************************************************/
/* For Intel type machines */
//#define RGPORT_LITTLE_ENDIAN

/* For motorla type machines */
//#define RGPORT_BIG_ENDIAN

// XXX: Thanks to James Urquhart (jamesu@gmail.com) for the information on why
//	typedefs should be used.
typedef unsigned char rgBYTE;
typedef unsigned short rgWORD;
typedef unsigned int rgDWORD;
typedef float rgFLOAT;

typedef long rgLONG;
typedef long long rgLONG64;

typedef rgFLOAT rgFloat;	/* 32 bit floating point */
typedef double rgDouble;	/* 64-bit floating point*/



class rgPortFileHandle
{
public:
	rgPortFileHandle(){ m_c_file = 0;}
	virtual ~rgPortFileHandle() { clear();}

	void clear() {
		m_str.str( "");
		if( m_inf.is_open()) {
			m_inf.close();
			m_inf.clear();
		}

		set_c_file( (FILE *)0);

		if( m_swap_file_name.length() > 0) {
			remove( m_swap_file_name.c_str());
			m_swap_file_name = std::string();
		}
	}

	bool is_real_file() {
		if( m_inf.is_open()) return true;
		return false;
	}

	// Description: Returns the input stream which contains the file data.
	std::istream &get_data() {
		if( m_inf.is_open()) return m_inf;
		return m_str;
	}

	virtual bool open_file( const std::string &p_fname) { return false;}

	virtual FILE *open_c_file( const std::string &p_fname) = 0;
	virtual void close_c_file( FILE *f) = 0;
	FILE *get_c_file() { return m_c_file;}

	void set_swap_file_name( const std::string &fname) {
		m_swap_file_name = fname;}
	const std::string &swap_file_name() const { return m_swap_file_name;}

//protected:
	// Used if a direct connection to file can be aquired or if a very
	// large archive file must be dumped to a temporary location.
	std::ifstream m_inf;
	std::string m_swap_file_name;

	// If file must be extracted from an archive, then it will be stored
	// in here. There is a size limit to files within an archive (this is
	// because the ENTIRE file must be extracted into a buffer). If an
	// archive file exceeds this, then it will be dumped to a temporary
	// file on the "backing store"/harddrive where there "should" be
	// enough space for it.
	std::stringstream m_str;

protected:
	void set_c_file( FILE *f) {
		if( m_c_file) {
			close_c_file( m_c_file);
		}
		m_c_file = f;
	}

private:
	FILE *m_c_file;

	rgPortFileHandle( const rgPortFileHandle &);
	rgPortFileHandle &operator =( const rgPortFileHandle &);
};


bool rgPortEndianCheck();

/*****************************************************************************
 * Description: Determines the endian type of the 32-bit machine the function
 *	is executed on. For 64-bit machines, use "rgPortWhichEndian_64".
 *
 * Returns: 0 is returned for Big-Endian, 1 is returned for Little-Endian,
 *	2 is returned for Big-Little-Endian and -1 is returned for any other
 *	endian type.
 *****************************************************************************/
int rgPortWhichEndian_32();

/*****************************************************************************
 * FIXME: rgPortWhichEndian_64 does not currently exist.
 *****************************************************************************/
//int rgPortWhichEndian_64();

/*****************************************************************************
 * Description: Takes a little endian WORD, DWORD or FLOAT and converts it
 *	into the machines native endian format.
 *****************************************************************************/
rgWORD rgPortLittleToNativeWORD( rgWORD n);
rgDWORD rgPortLittleToNativeDWORD( rgDWORD n);
rgFLOAT rgPortLittleToNativeFLOAT( rgFLOAT n);

/*****************************************************************************
 * Description: Takes a big endian WORD, DWORD or FLOAT and converts it into
 *	the machines native endian format.
 *****************************************************************************/
rgWORD rgPortBigToNative( rgWORD n);
rgDWORD rgPortBigToNative( rgDWORD n);
rgFLOAT rgPortBigToNative( rgFLOAT n);


rgWORD rgPortLittleToBigWORD( rgWORD n);
rgDWORD rgPortLittleToBigDWORD( rgDWORD n);
rgFLOAT rgPortLittleToBigFLOAT( rgFLOAT n);

rgWORD rgPortBigToLittleWORD( rgWORD n);
rgDWORD rgPortBigToLittleDWORD( rgDWORD n);
rgFLOAT rgPortBigToLittleFLOAT( rgFLOAT n);

/*****************************************************************************
 * Description: These functions return a WORD or DWORD in little endian format,
 *	regardless of the endian format of the parameter (the parameter will
 *	ALWAYS be in the native endian format of the machine as determined at
 *	compile time).
 *****************************************************************************/
rgWORD rgPortLittleEndianWORD( rgWORD n);
rgDWORD rgPortLittleEndianDWORD( rgDWORD n);
rgFLOAT rgPortLittleEndianFLOAT( rgFLOAT n);

/*****************************************************************************
 * Description: These functions return a WORD or DWORD in big endian format,
 *	regardless of the endian format of the parameter (the parameter will
 *	ALWAYS be in the native endian format of the machine as determined at
 *	compile time).
 *****************************************************************************/
rgWORD rgPortBigEndianWORD( rgWORD n);
rgDWORD rgPortBigEndianDWORD( rgDWORD n);
rgFLOAT rgPortBigEndianFLOAT( rgFLOAT n);

rgWORD rgPortBigLittleSwapWORD( rgWORD n);
rgDWORD rgPortBigLittleSwapDWORD( rgDWORD n);
rgFLOAT rgPortBigLittleSwapFLOAT( rgFLOAT n);



// Description: Takes a full path name (or just a file name) and writes out
//	only the file name portion to "des". The full path is interpreted as
//	a Unix path name. "max" indicates the maximum length of "des". If
//	"des" is not NULL and "max" is at least one, then "des" will ALWAYS
//	be NULL terminated.
//
//	DANGER: "des" and "src" must NOT be overlapping memory regions. This
//		function does not check if "des" and "src" overlap. Results
//		are undefined if "des" and "src" overlap.
void rg_get_unix_file_name_only( char *des, const char *src, int max);

// Description: Takes a full path name with a file name possibly included and
//	writes out only the path portion to "des". The full path is interpreted
//	as a Unix path name. "max" indicates the maximum length of "des". If
//	"des" is not NULL and "max" is at least one, then "des" will ALWAYS
//	be NULL terminated.
//
//	DANGER: "des" and "src" must NOT be overlapping memory regions. This
//		function does not check if "des" and "src" overlap. Results
//		are undefined if "des" and "src" overlap.
void rg_get_unix_path_name_only( char *des, const char *src, int max);

bool rg_slash_terminates_unix_path( const char *path);

std::string rg_get_extension( const std::string &fname);

inline void write_word( std::ostream &des, rgWORD w) {
	w = rgPortLittleEndianWORD( w);
	des.write( (char *)&w, sizeof( w));
}

inline void write_dword( std::ostream &des, rgDWORD d) {
	d = rgPortLittleEndianDWORD( d);
	des.write( (char *)&d, sizeof( d));
}

inline void write_float( std::ostream &des, rgFLOAT f) {
	f = rgPortLittleEndianFLOAT( f);
	des.write( (char *)&f, sizeof( f));
}

inline rgWORD read_word( std::istream &inf) {
	rgWORD w;
	inf.read( (char *)&w, sizeof( w));
	return rgPortLittleToNativeWORD( w);
}
                                                                                
inline rgDWORD read_dword( std::istream &inf) {
	rgDWORD d;
	inf.read( (char *)&d, sizeof( d));
	return rgPortLittleToNativeDWORD( d);
}
                                                                                
inline rgFLOAT read_float( std::istream &inf) {
	rgFLOAT f;
	inf.read( (char *)&f, sizeof( f));
	return rgPortLittleToNativeFLOAT( f);
}

char *rgport_getcwd( char *buf, const int len);


int rgport_mkstemp( char *s);
void rgport_close( int);
std::string rgport_file_copy( std::istream &src, FILE *des);


#endif
