/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "arena.h"
#include "engine/engine_game.h"
#include "engine/engine_graphics.h"
#include "team.h"
#include "puck.h"
#include "player.h"
#include "camera.h"
#include "referee.h"
#include <rg3DRom.h>

std::map <const std::string, THING_COMMAND_FN> phArena::m_ccmd;

extern "C" gmThing *spw_phArena()
{
        return new phArena();
}

extern "C" int del_phArena( gmThing *t)
{
	delete t;
	return 0;
}

RPH_THING_MODULE_MAIN_FUNCTION( phArena)
{
	p_ge->m_thing_manager->add_to_spawn_list( "phArena", spw_phArena);
	p_ge->m_thing_manager->add_to_delete_list( "phArena", del_phArena);
	return 0;
}

phArena::phArena()
{
	dsplst_ice_surface = 0;
	dsplst_boards = 0;

	m_ice_color = rgVector( 1, 1, 1);
	m_light_level = rgVector( 0.9, 0.9, 0.9);
	m_ice_alpha = 0.75;
	m_reflections = true;
	m_stencil_shadows = true;
	m_shadow_alpha = 0.5f;

	m_game_puck = 0;
	m_primary_referee = 0;

	m_home_team = 0;
	m_away_team = 0;

	if( !(m_ccmd.empty())) return;
	add_console_command( "set_official_game_thing",
		phArena_set_official_game_thing, m_ccmd);
	add_console_command( "ice_color", phArena_ice_color, m_ccmd);
	add_console_command( "light_level", phArena_light_level, m_ccmd);
	add_console_command( "shadow_color", phArena_shadow_color, m_ccmd);
	add_console_command( "ice_alpha", phArena_ice_alpha, m_ccmd);
	add_console_command( "shadow_alpha", phArena_shadow_alpha, m_ccmd);
	add_console_command( "reflections", phArena_reflections, m_ccmd);
	add_console_command( "stencil_shadows", phArena_stencil_shadows,
		m_ccmd);
	add_console_command( "prepare_for_rendering",
		phArena_prepare_for_rendering, m_ccmd);
	add_console_command( "add_to_render_list",
		phArena_add_to_render_list, m_ccmd);
	add_console_command( "remove_from_render_list",
		phArena_remove_from_render_list, m_ccmd);
}

gmThingError phArena::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.find( key);
	if( iter == m_ccmd.end()) return gmModelThing::call_ccmd( key, args);
	return (iter->second)( args);
}

std::string phArena::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); ++iter) {
		tmp += p_prefix + phArena::type() + "::" + (iter->first) +
			"\n";
	}
	tmp += gmModelThing::list_ccmd( p_prefix);
	return tmp;
}

gmTE phArena_set_official_game_thing( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() < 1) {
		return "usage: set_official_game_thing <key> "
				"[option 1] [option 2] ... [option n]\n"
			"NOTE: \"key\" can be things like a puck or referee.\n"
			"If a puck is added, it will become the official game"
				" puck.\n"
			"If a referee is added, they will become the official "
				"referee.\n"
			"\"Options\" can be specified when setting particular "
				"things.\n"
			"For example, when setting a team, a single option "
				"\"home\" can be given indicating\n"
			"that the team should be made the home team.\n"
			"In the future other things can be added, and anything"
				" that has no \"official\"\n"
			"place in the arena, will not be added.\n";
	}

	ThingControlBlock *t = p.ge->m_thing_manager->get( args[0]);
	if( !t)
		return "No such thing \"" + args[0] + "\"\n";
	if( t->is_free())
		return "Thing \"" + args[0] + "\" is free (unusable)\n";

	//////////////////////////////////////////////////////////////////////
	// XXX: Sub-classes MUST be placed FIRST in the "if-else" ladder
	//	(Super-classes towards the end). If this requirement is not
	//	followed then a single Super-class can prevent other
	//	Sub-classes from ever being accessed (since every Sub-class is
	//	a kind of Super-class, and the Super-class "if" test will
	//	always pass).
	//////////////////////////////////////////////////////////////////////
	phArena *a = (phArena *)(p.thing);
	if( t->m_thing->is_a( "phPuck")) a->set_game_puck( t);

	else if( t->m_thing->is_a( "phReferee")) {
		a->set_primary_referee( t);
		((phReferee *)(t->m_thing))->set_arena( a->tcb());

	} else if( t->m_thing->is_a( "phPerson")) {
		// The given person may not have an "official" position in the
		// arena, however they will be notified that this arena is now
		// the arena they are in.
		((phPerson *)(t->m_thing))->set_arena( a->tcb());

	} else if( t->m_thing->is_a( "phTeam")) {
		const std::string emsg( "Single option \"home\" or \"away\""
			" expected when setting team \"" + args[0] + "\"\n");
		if( args.size() != 2) return emsg;

		if( "home" == args[1]) a->set_home_team( t);
		else if( "away" == args[1]) a->set_away_team( t);
		else return emsg;
		((phTeam *)(t->m_thing))->set_arena( a->tcb());

	} else {
		return "Thing \"" + args[0] + "\" has no official "
			"place in the arena\n";
	}
	return std::string();
}

gmTE phArena_ice_color( const gmCCArg &p)
{
        const std::vector <std::string> &args = *(p.args);
        if( args.size() != 3) {
		return "usage: ice_color <red> <green> <blue>\n"
			"NOTE: All three components are given as floating "
			"point numbers\nwith a range of [0, 1]\n";
	}

	((phArena *)(p.thing))->m_ice_color = gm_str_to_vector( args[0],
		args[1], args[2]);
        return std::string();
}

gmTE phArena_ice_alpha( const gmCCArg &p)
{
        const std::vector <std::string> &args = *(p.args);
        if( args.size() != 1) {
		return "usage: ice_alpha <value>\n"
			"NOTE: \"value\" is a number in the range [0, 1]\n";
	}
	((phArena *)(p.thing))->m_ice_alpha = strtod( (args[0]).c_str(), 0);
        return std::string();
}

gmTE phArena_shadow_alpha( const gmCCArg &p)
{
        const std::vector <std::string> &args = *(p.args);
        if( args.size() != 1) {
		return "usage: shadow_alpha <value>\n"
			"NOTE: \"value\" is a number in the range [0, 1]\n";
	}
	((phArena *)(p.thing))->m_shadow_alpha = strtod( (args[0]).c_str(), 0);
        return std::string();
}

gmTE phArena_light_level( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 3) {
		p.ge->rcout( "usage: light_level <red> <green> <blue>\n"
			"NOTE: All three components are given as floating "
			"point numbers\nwith a range of [0, 1]\n");
		return GM_THING_ERR_PARAM;
	}
	((phArena *)(p.thing))->m_light_level =
		gm_str_to_vector( args[0], args[1], args[2]);
	return 0;
}

gmTE phArena_shadow_color( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 3) {
		p.ge->rcout( "usage: shadow_color <red> <green> <blue>\n"
			"NOTE: All three components are given as floating "
			"point numbers\nwith a range of [0, 1]\n");
		return GM_THING_ERR_PARAM;
	}
	((phArena *)(p.thing))->m_shadow_color =
		gm_str_to_vector( args[0], args[1], args[2]);
	return 0;
}

gmTE phArena_reflections( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) return GM_THING_ERR_PARAM;
	gm_set_state( ((phArena *)(p.thing))->m_reflections, args[0]);
	return 0;
}

gmTE phArena_stencil_shadows( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) return GM_THING_ERR_PARAM;
	gm_set_state( ((phArena *)(p.thing))->m_stencil_shadows, args[0]);
	return 0;
}

gmTE phArena_add_to_render_list( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 2) {
		p.ge->rcout( "usage: add_to_render_list <thing> <list name>\n"
			"NOTE: Available lists are \"opaque\" and "
				"\"player\".\n");
		return GM_THING_ERR_PARAM;
	}
	ThingControlBlock *tcb = p.ge->m_thing_manager->get( args[0]);
	if( !tcb) {
		p.ge->rcout( "No such thing \"" + args[0] + "\"\n");
		return GM_THING_ERR_GENERIC;
	}

	phArena *a = (phArena *)(p.thing);
	if( !(a->get_render_list( args[1]))) {
		p.ge->rcout( "No such render list \"" + args[1] + "\"\n");
		return GM_THING_ERR_GENERIC;
	}
	a->add_to_render_list( tcb, args[1]);
	return 0;
}

gmTE phArena_remove_from_render_list( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 2) {
		p.ge->rcout( "usage: remove_from_render_list <thing> "
			"<list name>\n"
			"NOTE: Available lists are \"opaque\" and "
				"\"player\".\n");
		return GM_THING_ERR_PARAM;
	}
	ThingControlBlock *tcb = p.ge->m_thing_manager->get( args[0]);
	if( !tcb) {
		p.ge->rcout( "No such thing \"" + args[0] + "\"\n");
		return GM_THING_ERR_GENERIC;
	}
	phArena *a = (phArena *)(p.thing);
	if( !(a->get_render_list( args[1]))) {
		p.ge->rcout( "No such render list \"" + args[1] + "\"\n");
		return GM_THING_ERR_GENERIC;
	}
	a->remove_from_render_list( tcb, args[1]);
	return 0;
}

gmTE phArena_prepare_for_rendering( const gmCCArg &p)
{
	phArena *a = (phArena *)(p.thing);
	if( !(a->model())) {
		p.ge->rcout( "cannot prepare to render (no model)\n");
		return -1;
	}

	(a->model())->transform_skel_frame( a->m_draw_context);
	
	{	// For speed, render trusses and flags together in their own
		// display lists (keeping them seperate from the rest of the
		// arena).
		glNewList( a->create_display_list( "truss_flags"), GL_COMPILE);
		glBindTexture( GL_TEXTURE_2D, 0);
		a->render_face_group( "lamps", false, false);
		a->render_face_group( "truss", true, true);
		a->render_face_group( "flags", true, true);
		glEndList();
	}

	{	// Seats should be rendered in their own display list. Also,
		// include the arena box along with other parts of the stands.
		glNewList( a->create_display_list( "seats"), GL_COMPILE);
		a->render_face_group( "stands_floor", true, true);
		a->render_face_group( "seats", true, true);
		a->render_face_group( "arena_box", true, true);
		a->render_face_group( "floor", true, true);
		glEndList();
	}

	a->compile_face_list( "Ice Surface", a->dsplst_ice_surface);

	{ // Prepare the boards for rendering
		const char *tmp[] = { "Dasher Board Top", "Boards",
			"Top of boards (red part)"};
		a->compile_face_list( tmp, 3, a->dsplst_boards);
	}
	p.ge->rcout( "arena ready for rendering\n");
	return 0;
}

phArena::~phArena()
{
	clear();
}

void phArena::clear()
{
	clear_display_list( dsplst_ice_surface);
	clear_display_list( dsplst_boards);
}

void phArena::compile_face_list( const char *p_name, GLuint &id)
{
	const char *tmp[] = { p_name};
	compile_face_list( tmp, 1, id);
}

void phArena::compile_face_list( const char *p_name[], const int p_count,
	GLuint &id)
{
	if( id) { glDeleteLists( id, 1); id = 0;}
	id = glGenLists( 1);
	glNewList( id, GL_COMPILE);
	for( int i = 0; i < p_count; i++) {
		int count;
		const int *flist = (model())->get_faces_in_group(
			(model())->get_group_id( p_name[i]), count);
		if( flist) (model())->render_faces( true, true, flist, count);
	}
	glEndList();
}

void phArena::clear_display_list( GLuint &id)
{
	if( 0 == id) return;
	glDeleteLists( id, 1);
	id = 0;
}

void phArena::draw_boards() const
{
	if( dsplst_boards) glCallList( dsplst_boards);
}

void phArena::draw_glass()
{
	if( !(model())) return;
	gmCamera *c = (ge())->gfx_engine->current_camera();
	(model())->render_bsp_faces( m_draw_context, c->cs.p, true, true,
		0, 0);
}

// Description: The arena's "process" function handles rendering for everything
//	in the arena. This "process" function will only be called when
//	graphics need to be displayed. Therefore, features such as a
//	"jumbo-tron" (which requires processing beyond graphic updates)
//	should be created as a separate "thing" (not part of the arena class)
//	so that it can get its processing done as it likes. If a "thing" wants
//	to be rendered, then it must register with the arena, or a "thing"
//	which is already registered with the arena.
void phArena::process( ThingManager *tm)
{
	{ // Must clear colour buffer if wireframe rendering is on.
		GLuint flags = GL_DEPTH_BUFFER_BIT;
		if( m_stencil_shadows) flags |= GL_STENCIL_BUFFER_BIT;
		GLint front_back[2];
		glGetIntegerv( GL_POLYGON_MODE, front_back);
		if(	(GL_LINE == front_back[0]) ||
			(GL_LINE == front_back[1]))
			glClear( flags | GL_COLOR_BUFFER_BIT);
		else glClear( flags);
	}
	

	// Render reflections
	// FIXME: Write an efficient (and correct) reflection algorithm.
	glDisable( GL_BLEND);
	if( m_reflections) {
		glPushMatrix();
		glScalef( 1, -1, 1);
		glCullFace( GL_FRONT);

		// Boards
		glColor3( m_light_level);
		glDisable( GL_LIGHTING);
		draw_boards();
		glCallList( get_display_list( "truss_flags"));
		glCallList( get_display_list( "seats"));

		// "things"
		glEnable( GL_LIGHTING);
		glColor3f( 1, 1, 1);
		gmRenderArgs tmp;
		render_all_in_list( &tmp, "opaque");

		glPopMatrix();
	}

	{
		glEnable( GL_LIGHTING);
		glCullFace( GL_BACK);
		gmRenderArgs ra;
		ra.render_reflections = m_reflections;
		render_all_in_list( &ra, "player");
		glDisable( GL_LIGHTING);
	}

	if( m_reflections) {
		// Glass
		glPushMatrix();
		glScalef( 1, -1, 1);
		glCullFace( GL_FRONT);

		glEnable( GL_BLEND);
		draw_glass();
		glCullFace( GL_BACK);

		glPopMatrix();
	}

	// Render ice-surface
	glEnable( GL_BLEND);
	glColor4f( m_ice_color.x, m_ice_color.y, m_ice_color.z, m_ice_alpha);
	draw_ice_surface();

	// Render ice decals (including player cursors).
	glDisable( GL_LIGHTING);
	{
		gmRenderArgs ra;
		ra.render_decals = true;
		ra.render_opaque = false;
		glDepthMask( false);
		render_all_in_list( &ra, "player");
		glDepthMask( true);
	}

	// Render remainder of arena
	// boards
	//glEnable( GL_LIGHTING);
	glDisable( GL_BLEND);
	glColor3( m_light_level);
	draw_boards();
	glCallList( get_display_list( "truss_flags"));
	glCallList( get_display_list( "seats"));


	// Render all things with shadowing.
	glDisable( GL_LIGHTING);
	glEnable( GL_BLEND);
	{
		gmRenderArgs ra( 0, m_stencil_shadows);
		ra.shadow_alpha = m_shadow_alpha;
		ra.shadow_color = m_shadow_color;
		render_all_in_list( &ra, "opaque");

		// XXX: For players, ONLY render the shadow and nothing else.
		//	Also, all player use the same GL state when rendering
		//	their shadows.
		if( m_stencil_shadows) {
			glPushMatrix();
			glTranslatef( 0, 0.025, 0);     // 2.5 cm above ice
			glScalef( 1, 0, 1);             // Flat on ice surface

			glColor4f(	ra.shadow_color.x,
					ra.shadow_color.y,
					ra.shadow_color.z,
					ra.shadow_alpha);

			glStencilOp( GL_KEEP, GL_REPLACE, GL_REPLACE);
			glStencilFunc( GL_GREATER, 1, 0xFFFFFFFF);
			glEnable( GL_STENCIL_TEST);

			ra.render_opaque = false;
			glDepthMask( false);
			render_all_in_list( &ra, "player");
			glDepthMask( true);

			glDisable( GL_STENCIL_TEST);

			glPopMatrix();
		}
	}

	// Render all things again, only this time, only the transparent or
	// translucent parts are rendered.
	// FIXME: The "things" still have to be sorted to allow the rendering
	//	to wrok properly. For now, the problem is being ignored.
	glEnable( GL_LIGHTING);
	{
		gmRenderArgs ra;
		ra.render_opaque = false;
		ra.render_transparent = true;
		ra.transform = false;
		render_all_in_list( &ra, "opaque");
		render_all_in_list( &ra, "player");
	}

	// Glass
	glDisable( GL_LIGHTING);
	draw_glass();

}

// Description: Renders all "things" in a given list. At the same time, any
//	"things" which are no longer active (dead) are removed from the list
//	so that they will never be accessed again. This is a convinent
//	automatic clean up in the event that the "thing" did not remove itself
//	from the render list before destroying itself.
void phArena::render_all_in_list( gmRenderArgs *ra, const std::string &rlist)
{
	std::map <ThingControlBlock *, void *> *tmp = get_render_list( rlist);
	if( !tmp) return;
	std::map <ThingControlBlock *, void *> :: const_iterator iter =
                tmp->begin();
	std::vector <ThingControlBlock *> rm_list;
	for( ; iter != tmp->end(); ++iter) {
		if( !((iter->first)->is_free()))
			(iter->first)->m_thing->draw( ra);
		else rm_list.push_back( iter->first);
	}
	while( !(rm_list.empty())) {
		remove_from_render_list( rm_list[0], tmp);
		rm_list.erase( rm_list.begin());
	}
}

std::map <ThingControlBlock *, void *> *phArena::get_render_list(
	const std::string &lname)
{
	if( "opaque" == lname) return &m_opaque;
	else if( "player" == lname) return &m_player;
	else return 0;
}

bool phArena::add_to_render_list( ThingControlBlock *tcb,
	const std::string &des)
{
	typedef std::pair <ThingControlBlock *, void *> TPPair;
	std::map <ThingControlBlock *, void *> *tmp = get_render_list( des);
	if( !tmp) return false;
	return ((tmp->insert( TPPair( tcb, 0))).second);
}

void phArena::remove_from_render_list( ThingControlBlock *tcb,
	std::map <ThingControlBlock *, void *> *tar)
{
	if( !tar) return;
	std::map <ThingControlBlock *, void *> :: iterator iter =
		tar->find( tcb);
	if( iter == tar->end()) return;
	tar->erase( iter);
}

void phArena::remove_from_render_list( ThingControlBlock *tcb,
	const std::string &des)
{
	remove_from_render_list( tcb, get_render_list( des));
}

phPerson *phArena::nearest_puck( const std::vector <phPerson *> &plist)
{
	std::vector <phPerson *> :: const_iterator iter = plist.begin();
	float dis = 16777816.0f;
	phPerson *near = 0;
	phPuck *p = game_puck();
	int count = int( plist.size());
	while( count--) {
		phPerson *per = (*iter);
		const float d = (per->direction_to( p)).length();
		if( d < dis) { near = per; dis = d;}
		++iter;
	}
	return near;
}

phCamera *get_camera_controller( gmCamera *c)
{
	if( !c) return 0;
	ThingControlBlock *tcb = c->owner();
	if( !tcb) return 0;
	if( tcb->is_free()) return 0;
	if( !(tcb->m_thing->is_a( "phCamera"))) return 0;
	return ((phCamera *)(tcb->m_thing));
}
