/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <rgOpenGL.h>
#include <rg3DRom.h>
#include <rgError.h>
#include "engine/engine_game.h"
#include "engine/engine_graphics.h"
#include "player.h"
#include "puck.h"
#include "arena.h"
#include "team.h"

std::map <const std::string, THING_COMMAND_FN> phPerson::m_ccmd;
std::map <const std::string, THING_COMMAND_FN> phPlayer::m_ccmd;

extern "C" gmThing *spw_phPerson()
{
	return new phPerson();
}

extern "C" gmThing *spw_phPlayer()
{
	return new phPlayer();
}

extern "C" int del_phPerson( gmThing *t)
{
	delete t;
	return 0;
}

extern "C" int del_phPlayer( gmThing *t)
{
	delete t;
	return 0;
}

RPH_THING_MODULE_MAIN_FUNCTION( phPerson)
{
	p_ge->m_thing_manager->add_to_spawn_list( "phPerson", spw_phPerson);
	p_ge->m_thing_manager->add_to_spawn_list( "phPlayer", spw_phPlayer);

	p_ge->m_thing_manager->add_to_delete_list( "phPerson", del_phPerson);
	p_ge->m_thing_manager->add_to_delete_list( "phPlayer", del_phPlayer);

	return 0;
}

phPerson::phPerson()
{
	m_arena = 0;
	m_team = 0;

	if( !(m_ccmd.empty())) return;
	add_console_command( "set_full_name", phPerson_set_full_name, m_ccmd);
	add_console_command( "cursor_color", phPerson_cursor_color, m_ccmd);
	add_console_command( "cursor_texture", phPerson_cursor_texture,
		m_ccmd);
	add_console_command( "scale_cursor", phPerson_scale_cursor, m_ccmd);
	add_console_command( "set_head_texture", phPerson_set_head_texture,
		m_ccmd);
}

phPerson::~phPerson()
{
}

gmTE phPerson::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.find( key);
        if( iter == m_ccmd.end()) return gmPhyModel::call_ccmd( key, args);
        return (iter->second)( args);
}
                                                                                
std::string phPerson::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.begin();
        std::string tmp;
        for( ; iter != m_ccmd.end(); ++iter) {
		tmp += p_prefix + phPerson::type() + "::" + (iter->first) +
		"\n";
	}
	tmp += gmPhyModel::list_ccmd( p_prefix);
	return tmp;
}

gmTE phPerson_set_full_name( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) {
		p.ge->rcout( "usage: set_full_name <full name>\n"
			"NOTE: Use quotes around the full name since only a "
				"sngle argument is accepted.\n");
		return GM_THING_ERR_PARAM;
	}
	((phPerson *)(p.thing))->m_full_name = args[0];
	return 0;

}

gmTE phPerson_set_head_texture( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) {
		p.ge->rcout( "usage: set_head_texture <head texture>\n"
			"NOTE: \"head texture\" is the name of a pre-loaded "
				"texture\n");
		return GM_THING_ERR_PARAM;
	}

	TextureHandle *th = p.ge->gfx_engine->get_texture( args[0]);
	if( !th) {
		p.ge->rcout( "No such texture \"" + args[0] + "\"\n");
		return GM_THING_ERR_GENERIC;
	}
	if( !(th->m_tex)) {
		p.ge->rcout( "Texture \"" + args[0] + "\" is not ready (m_tex "
			"== NULL)\n");
		return GM_THING_ERR_GENERIC;
	}
	((phPerson *)(p.thing))->m_head_texture = args[0];
	return 0;

}

gmTE phPerson_cursor_color( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 4) {
		p.ge->rcout( "usage: cursor_color <red> <green> <blue> <alpha>"
				"\n"
                        "NOTE: All three components are given as floating "
                        "point numbers\nwith a range of [0, 1]\n");
                return GM_THING_ERR_PARAM;
        }

	((phPerson *)(p.thing))->m_cursor.m_color = gm_str_to_vector( args[0],
		args[1], args[2]);
	((phPerson *)(p.thing))->m_cursor.m_alpha =
		strtod( (args[3]).c_str(), 0);
	return 0;
}

gmTE phPerson_cursor_texture( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) {
		p.ge->rcout( "usage: cursor_texture <name of texture>\n");
		return GM_THING_ERR_PARAM;
	}
	((phPerson *)(p.thing))->m_cursor.m_tex_name = args[0];
	return 0;
}

gmTE phPerson_scale_cursor( const gmCCArg &p)
{
        const std::vector <std::string> &args = *(p.args);
        if( args.size() != 4) {
                p.ge->rcout( "usage: cursor_scale <x> <y> <z>\n");
                return GM_THING_ERR_PARAM;
        }
	rgMatrix m;
	m.scale(	strtod( (args[0]).c_str(), 0),
			strtod( (args[1]).c_str(), 0),
			strtod( (args[2]).c_str(), 0));
	phPerson *tmp = (phPerson *)(p.thing);
	tmp->m_cursor.m_mat = m * tmp->m_cursor.m_mat;
	return 0;
}

phTeam *phPerson::team()
{
	if( (!m_team) || m_team->is_free()) return 0;
	// FIXME: This is a safety check which is probably not neccasary.
	if( m_team->m_thing->is_a( "phTeam"))
		return (phTeam *)(m_team->m_thing);
	else return 0;
}

phArena *phPerson::arena()
{
	if( !m_arena) return 0;
	// FIXME: This is a safety check which is probably not neccasary.
	if( m_arena->m_thing->is_a( "phArena"))
		return (phArena *)(m_arena->m_thing);
	else return 0;
}



phPlayer::phPlayer()
{
	//m_cursor.m_tex_name = "tex_player_1_cursor";

	if( !(m_ccmd.empty())) return;
	add_console_command( "set_attrib", phPlayer_set_attrib, m_ccmd);
	add_console_command( "create_uniform_patch",
		phPlayer_create_uniform_patch, m_ccmd);
	add_console_command( "set_sub_patch", phPlayer_set_sub_patch, m_ccmd);
}

phPlayer::~phPlayer()
{
}

gmTE phPlayer::call_ccmd( const std::string &key, gmCCArg &args)
{
        std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.find( key);
        if( iter == m_ccmd.end()) return phPerson::call_ccmd( key, args);
        return (iter->second)( args);
}
                                                                                
std::string phPlayer::list_ccmd( const std::string &p_prefix)
{
        std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.begin();
        std::string tmp;
        for( ; iter != m_ccmd.end(); ++iter) {
		tmp += p_prefix + phPlayer::type() + "::" + (iter->first) +
		"\n";
	}
	tmp += phPerson::list_ccmd( p_prefix);
	return tmp;
}


bool hlp_move_to_value( int &i, const std::vector <std::string> &args)
{
	i++; // Move to equal sign (or value)
	if( i >= int( args.size())) return false;
	if( "=" == args[i]) i++;

	// Now at value
	if( i >= int( args.size())) return false;
	return true;
}

gmTE phPlayer_set_attrib( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() <= 0) {
		p.ge->rcout( "usage: set_attrib <attrib1 = value1> "
			"[attrib2 = value2] ... [attribN = valueN]\n");
		return GM_THING_ERR_PARAM;
	}
	phPlayer *ply = (phPlayer *)(p.thing);
	for( int i = 0; i < int( args.size()); ) {
		if( "number" == args[i]) {
			if( hlp_move_to_value( i, args))
				ply->m_number = atoi( (args[i]).c_str());

		} else if( "rank" == args[i]) {
			if( hlp_move_to_value( i, args))
				ply->m_rank = atoi( (args[i]).c_str());

		} else {
			p.ge->rcout( "Unrecognized attribute \"" + args[i] +
				"\"\n");
			if( hlp_move_to_value( i, args))
				i = i; // Do nothing
		}

		i++; // Move onto next attrib value pair
	}
	return 0;
}

gmTE phPlayer_create_uniform_patch( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 4) {
		p.ge->rcout( "usage: create_uniform_patch "
			"<base_uniform_page> <patch_page> <width> <height>\n"
			"NOTE: \"base_uniform_page\" is the name of the "
				"texture page\n"
			"from the players model which contains the base "
				"uniform texture uv-coords.\n"
			"\"patch_page\" is the name of the texture page from "
				"the players model which\n"
			"contains the patch uv-texture coords.\n"
			"\"width\" and \"height\" are the dimensions of the "
				"texture to create\n"
			"for the player.\n");
		return GM_THING_ERR_PARAM;
	}
	phPlayer *ply = (phPlayer *)(p.thing);
	// FIXME: Check to make sure given pages actually exist in model.
	ply->m_upc.m_base_uniform_page = args[0];
	ply->m_upc.m_patch_page = args[1];
	ply->m_upc.m_width = atoi( (args[2]).c_str());
	ply->m_upc.m_height = atoi( (args[3]).c_str());
	const std::string emsg( ply->create_uniform_patch());
	if( emsg.length() > 0) {
		p.ge->rcout( emsg + "\n");
		return GM_THING_ERR_GENERIC;
	}
	return 0;
}


gmTE phPlayer_set_sub_patch( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 8) {
		p.ge->rcout( "usage: set_sub_patch\n\t"
			"<sub patch>\n\t"
			"<left>\n\t"
			"<width>\n\t"
			"<bottom>\n\t"
			"<height>\n\t"
			"<char width>\n\t"
			"<char height>\n\t"
			"<char space>\n"
			"NOTE: \"sub patch\" must be \"big_num\", \"name\", "
				"\"right_num\" or \"left_num\"\n"
			"All dimensions and locations are in percentage of "
				"texture dimensions\n"
			"(That is, values should have a range of [0, 1])\n");
		return GM_THING_ERR_PARAM;
	}

	phPlayer *ply = (phPlayer *)(p.thing);	
	phUniformSubPatch *sp;
	if( "big_num" == args[0]) sp = &(ply->m_upc.m_big_num_patch);
	else if( "name" == args[0]) sp = &(ply->m_upc.m_name_patch);
	else if( "right_num" == args[0]) sp = &(ply->m_upc.m_rnum_patch);
	else if( "left_num" == args[0]) sp = &(ply->m_upc.m_lnum_patch);
	else {
		p.ge->rcout( "Invalid sub patch \"" + args[0] + "\"\n");
		return GM_THING_ERR_GENERIC;
	}

	sp->m_left = strtod( (args[1]).c_str(), 0);
	sp->m_bottom = strtod( (args[2]).c_str(), 0);
	sp->m_width = strtod( (args[3]).c_str(), 0);
	sp->m_height = strtod( (args[4]).c_str(), 0);
	sp->m_char_width = strtod( (args[5]).c_str(), 0);
	sp->m_char_height = strtod( (args[6]).c_str(), 0);
	sp->m_char_space = strtod( (args[7]).c_str(), 0);
	return 0;
}


//
// XXX NOTES: When shadows are enabled, ONLY the player's shadow is rendered,
//	nothing else.
void phPlayer::draw( gmRenderArgs *p_ra)
{
	if( !p_ra) return;
	gmRenderArgs ra( *p_ra);

	if( ra.render_decals && (m_cursor.m_tex_name.length() > 0)) {
		TextureHandle *th = (ge())->gfx_engine->get_texture(
			m_cursor.m_tex_name);
		if( th && (th->m_tex)) {
			// XXX: The arena disables lighting when enabling
			//	decal rendering. Also, blending is enabled.
			th->m_tex->bind();
			m_cursor.draw( (cs()).p);
		}
		return;
	}

	// XXX: The arena already disabled transformation for transparency.
	// if( ra.render_transparent) ra.transform = false;
	if( ra.render_shadow) {

		ModelHandle *mh = ((ge())->gfx_engine->model_man())->get(
			"mdl_player_shadow");
		if( mh && (mh->model)) {

			(mh->model)->transform_skel_frame( m_draw_context);
			(mh->model)->render_vertex_array( 0);
		}
		return;
	}

	// Render the full opaque part of the model OR render the
	// transparent parts of the model. Transformation for the model
	// is suppressed if transparent parts will be rendered.

	// Setup textures for this player
	// XXX: For now, transparent parts of the player do not use texturing.
	if( !(ra.render_transparent)) {
		phTeam *tm = team();
		if( tm) {
			tm->bind_uniform_textures( this);

			// Select head texture
			ModelHandle *mh = model_handle();
			mh->set_alternate_texture( "head", m_head_texture,
				(ge())->gfx_engine);
			mh->set_alternate_texture( "patch", m_uniform_patch,
				(ge())->gfx_engine);
		}
	}

	if( ra.render_reflections) ra.lock_vertices = true;
	gmModelThing::draw( &ra);

	if( ra.render_reflections && !(ra.render_transparent)) {

		glPushMatrix();
		glScalef( 1, -1, 1);
		glCullFace( GL_FRONT);

		ra.transform = false;
		gmModelThing::draw( &ra);

		glCullFace( GL_BACK);
		glPopMatrix();
	}
}


void phPersonCursor::draw( const rgVector p) const
{
	rgMatrix m( p);
	m = m * m_mat;
	m.translate( -p);

	const float h = 0.02;	// 2 centimeters above the ice.
	const rgVector a( m.pointMul( rgVector( p.x - 1.0f, h, p.z - 1.0f)));
	const rgVector b( m.pointMul( rgVector( p.x + 1.0f, h, p.z - 1.0f)));
	const rgVector c( m.pointMul( rgVector( p.x + 1.0f, h, p.z + 1.0f)));
	const rgVector d( m.pointMul( rgVector( p.x - 1.0f, h, p.z + 1.0f)));

	glColor4f( m_color.x, m_color.y, m_color.z, m_alpha);
	glBegin( GL_QUADS);
	glTexCoord2f( 1, 0);
	glVertex( a);
	glTexCoord2f( 0, 0);
	glVertex( b);
	glTexCoord2f( 0, 1);
	glVertex( c);
	glTexCoord2f( 1, 1);
	glVertex( d);
	glEnd();
}

phPuck *phPerson::has_game_puck()
{
	phArena *a = arena();
	if( !a) return 0;
	phPuck *p = a->game_puck();
	if( !p) return 0;
	if( get_binding( p->tcb())) return p;
	return 0;
}

bool phPerson::near_game_puck( const float radius)
{
	phArena *a = arena();
	if( !a) return false;
	phPuck *p = a->game_puck();
	if( !p) return false;
	if( (direction_to( p, rgVector( 0, 1, 0))).length() <= radius)
		return true;
	return false;
}

bool phPerson::near_ice_point( rgVector p, const float radius)
{
	if( (direction_to( p, rgVector( 0, 1, 0))).length() <= radius)
		return true;
	return false;
}

void phPlayer::process( ThingManager *tm)
{
	// Think...think...think...
	ThingControlBlock *ai = get_ai();
	if( ai && !(ai->is_free())) ai->m_thing->process( tm);

	// Animation and physics processing...
	phPerson::process( tm);
}

std::string phPlayer::create_uniform_patch(
	const std::string &p_tex,
	phUniformPatchContext *upc)
{
	if( !upc) return "parameter \"upc\" == NULL";
	if( (upc->m_width <= 32) || (upc->m_height <= 32)) {
		return "invalid texture dimensions (" +
			gm_to_string( upc->m_width) + " x " +
			gm_to_string( upc->m_height) + ")";
	}

	(ge())->gfx_engine->delete_texture( m_uniform_patch);
	TextureHandle *th = (ge())->gfx_engine->add_texture( p_tex);
	if( !th) return "cannot create texture \"" + p_tex + "\"";
	m_uniform_patch = p_tex;
	if( !(th->m_tex))
		return "texture \"" + p_tex + "\" not ready (m_tex == NULL)";

	{
		rgBMP bmp;
		if( bmp.generate_bmp( upc->m_width, upc->m_height, 24))
			return "could not generate image data for texture";
		const int err = th->m_tex->load( bmp, RGGL_SMOOTH_TEXTURE);
		if( err != RGERR_SUCCESS) {
			return "error " + gm_to_string( err) + " while "
				"generating OpenGL texture "
				"(rgGLTexture::load())";
		}
	}

	render_uniform_patch( th->m_tex, upc);
	return std::string();
}

void hlp_render_mapped_face(
	const sk3DRomTextureFace *des,
	const sk3DRomTextureFace *src,
	float width,
	float height)
{
	rgVector a( des->v(0));
	rgVector b( des->v(1));
	rgVector c( des->v(2));
	
	glTexCoord2( src->v( 0));
	glVertex2f( a.x * width, a.y * height);

	glTexCoord2( src->v( 1));
	glVertex2f( b.x * width, b.y * height);

	glTexCoord2( src->v( 2));
	glVertex2f( c.x * width, c.y * height);
}

void phPlayer::render_base_uniform(
	rgGLTexture *p_tex,
	phUniformPatchContext *upc)
{
	const sk3DRomTexturePage *src = (model())->page( (model_handle())->
		page_id( upc->m_base_uniform_page));
	if( !src) return;
	const sk3DRomTexturePage *des = (model())->page( (model_handle())->
		page_id( upc->m_patch_page));
	if( !des) return;

	phTeam *t = team();
	if( t) {
		TextureHandle *th;
		if( is_a( "phGoalie")) {
			th = (ge())->gfx_engine->get_texture(
				t->goalie_upper_uniform_texture());
		} else {
			th = (ge())->gfx_engine->get_texture(
				t->player_upper_uniform_texture());
		}
		if( th && th->m_tex) th->m_tex->bind();
	}

	glBegin( GL_TRIANGLES);
	for( int i = 0; i < (des->face_count()); i++) {
		const int j = src->face_in_texture_page(
			(des->face( i))->fid());
		if( j >= 0) {
			hlp_render_mapped_face(
				des->face( i),
				src->face( j),
				upc->m_width,
				upc->m_height);
		}
	}
	glEnd();
}

phUniformPatch *hlp_get_big_patch( phPerson *php, char c)
{
	phTeam *t = php->team();
	if( !t) return 0;
	const int n = RG_MIN( RG_MAX( 0, c - '0'), 15);
	return (t->m_uniform.m_big_number + n);
}

phUniformPatch *hlp_get_ascii_patch( phPerson *php, char c)
{
	phTeam *t = php->team();
	if( !t) return 0;
	const int n = RG_MIN( RG_MAX( 0, c), 127);
	return (t->m_uniform.m_ascii + n);
}

void phPlayer::render_uniform_string(
	phUniformPatchContext *upc,
	phUniformSubPatch *sp,
	const std::string &str,
	bool use_ascii)
{
	phUniformPatch * (*fn)( phPerson *, char);
	if( use_ascii) fn = hlp_get_ascii_patch;
	else fn = hlp_get_big_patch;

	const int len = str.length();
	if( len <= 0) return;
	float spacing = sp->m_char_space * float( len - 1);
	float w = RG_MIN( (sp->m_width - spacing) / float( len),
		sp->m_char_width);

	float h = sp->m_char_height;
	if( h > sp->m_height) h = sp->m_height;
	float y = sp->m_bottom + ((sp->m_height - h) * 0.5f);
	h *= upc->m_height;
	y *= upc->m_height;

	glBegin( GL_QUADS);

	float x = sp->m_left + ((sp->m_width - (w * float( len) + spacing)) *
		0.5f);

	float s = sp->m_char_space * upc->m_width;
	w *= upc->m_width;
	x *= upc->m_width;
	for( int i = 0; i < len; i++) {
		phUniformPatch &p = *(fn( this, str[i]));
		if( (&p)) {
			glTexCoord2f( p.m_left, p.m_bottom + p.m_height);
			glVertex2f( x, y + h);

			glTexCoord2f( p.m_left + p.m_width, p.m_bottom +
				p.m_height);
			glVertex2f( x + w, y + h);

			glTexCoord2f( p.m_left + p.m_width, p.m_bottom);
			glVertex2f( x + w, y);

			glTexCoord2f( p.m_left, p.m_bottom);
			glVertex2f( x, y);
		}

		// FIXME: Should increment by width of font (which is a per
		//	character width).
		x += w + s;
	}
	glEnd();
}

void phPlayer::render_uniform_patch(
	rgGLTexture *p_tex,
	phUniformPatchContext *upc)
{
	GLboolean texture2D;
	glGetBooleanv( GL_TEXTURE_2D, &texture2D);
	glEnable( GL_TEXTURE_2D);

	GLboolean lighting;
	glGetBooleanv( GL_LIGHTING, &lighting);
	glDisable( GL_LIGHTING);

	GLboolean cullface;
	glGetBooleanv( GL_CULL_FACE, &cullface);
	glDisable( GL_CULL_FACE);

	GLboolean depthtest;
	glGetBooleanv( GL_DEPTH_TEST, &depthtest);
	glDisable( GL_DEPTH_TEST);

	GLboolean blend;
	glGetBooleanv( GL_BLEND, &blend);
	glEnable( GL_BLEND);

	GLint mat_mode;
	glGetIntegerv( GL_MATRIX_MODE, &mat_mode);
	glMatrixMode( GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho( 0, (ge())->screen_width(), 0, (ge())->screen_height(), 1, -1);

	glMatrixMode( GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();

	GLint front_back[2];
	glGetIntegerv( GL_POLYGON_MODE, front_back);

	glPolygonMode( GL_FRONT_AND_BACK, GL_LINE);
	glLineWidth( 2);
	render_base_uniform( p_tex, upc);

	glPolygonMode( GL_FRONT_AND_BACK, GL_FILL);
	render_base_uniform( p_tex, upc);

	{
		phTeam *t = team();
		if( !t) return;
		TextureHandle *th = (ge())->gfx_engine->get_texture(
			t->m_uniform.m_patch);
		if( th && th->m_tex) th->m_tex->bind();
	}

	// FIXME: Provide a function to produce a "patch-ready" name. That is,
	//	the function (when given "m_full_name") should return the last
	//	name of the player, or if two players have same last name, then
	//	whatever preceeding initials required to differentiate the
	//	players plus their last name. Currently, the players entire
	//	name will be "patched" to their sweaters.
	render_uniform_string( upc, &(upc->m_name_patch),
		m_full_name, true);

	render_uniform_string( upc, &(upc->m_big_num_patch),
		gm_to_string( m_number), false);
	render_uniform_string( upc, &(upc->m_rnum_patch),
		gm_to_string( m_number), true);
	render_uniform_string( upc, &(upc->m_lnum_patch),
		gm_to_string( m_number), true);


	if( !texture2D) glDisable( GL_TEXTURE_2D);
	if( !blend) glDisable( GL_BLEND);
	if( depthtest) glEnable( GL_DEPTH_TEST);
	if( cullface) glEnable( GL_CULL_FACE);

	glPolygonMode( GL_FRONT, front_back[0]);
	glPolygonMode( GL_BACK, front_back[1]);

	glPopMatrix(); // Restore MODELVIEW

	glMatrixMode( GL_PROJECTION);
	glPopMatrix(); // Restore PROJECTION
	glMatrixMode( mat_mode);

	if( lighting) glEnable( GL_LIGHTING);

	p_tex->copy_sub_image(
		0, // Level
		0, // U location in texture
		0, // V location in texture
		0,
		0,
		p_tex->getWidth(),
		p_tex->getHeight());
}

