/*
Copyright (C) 2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <stdlib.h>
#include <iostream>
#include <string>
#include <stdio.h>
#include <sstream>
#include <fstream>
#include <rgPort.h>
#include "engine_file.h"
#include <rgStandard.h>

FILE *hlp_get_from_tar_gz( const std::string &, const std::string &);
FILE *hlp_get_from_tar_bz( const std::string &, const std::string &);
FILE *hlp_get_from_tar_zip( const std::string &, const std::string &);
int hlp_gm_file_fclose( FILE *f);
int hlp_gm_file_pclose( FILE *f);

std::string hlp_strip_leading_dot_slash( const std::string &s);
int gm_get_from_archive(
	const std::string &p_archive,
	const std::string &p_file,
	std::iostream &s,
	const int len,
	rgPortFileHandle *p_fh,
	FILE **c_file,
	int (**c_file_close_fn)( FILE *),
	FileEngine *fe);

bool gmFileHandle::open_file( const std::string &p_fname)
{
	if( !m_fe) return false;
	m_emsg = m_fe->get_file( p_fname, *this);
	if( m_emsg.length() > 0) return false;
	return true;
}

FILE *gmFileHandle::open_c_file( const std::string &p_fname)
{
	if( !m_fe) return 0;
	m_emsg = m_fe->get_c_file( p_fname, *this);
	if( m_emsg.length() > 0) return 0;
	return get_c_file();
}

FileEngine::FileEngine( GameEngine *p_ge)
{
	m_ge = p_ge;

	m_max_archive_file_size = MAX_ARCHIVE_FILE_SIZE;
}

std::string FileEngine::create_temporary_file( FILE **f, std::string &fname)
	const
{
	if( !f) return "null (FILE *) pointer passed";
	const int fd = create_temporary_file( fname);
	if( -1 == fd) return "could not create temporary file";
	(*f) = fopen( fname.c_str(), "r+");
	rgport_close( fd);
	if( !f) return "could not open created temporary file";
	return std::string();
}

std::string FileEngine::create_temporary_file( std::fstream &tmp,
	std::string &fname) const
{
	const int fd = create_temporary_file( fname);
	if( fd < 0) return "could not create temporary file";
	tmp.open( fname.c_str());
	rgport_close( fd);
	if( !(tmp.is_open()))
		return "could not open created temporary file";
	return std::string();
}

int FileEngine::create_temporary_file( std::string &fname) const
{
	if( fname.length() <= 0) fname = "temp";
	{ // Create unique file in the temporary directory
		if( fname.length() > 0) {
			switch( fname[ fname.length() - 1]) {
			case '/':
			case '\\':
				fname = m_temp_dir + fname;
				break;
			default:
				fname = m_temp_dir + "/" + fname;
				break;
			}
		}
		fname += "XXXXXX";
	}
	rgTempArray <char> tmp( fname.length() + 1);
	if( tmp.empty()) return -2;
	strncpy( tmp.data, fname.c_str(), fname.length() + 1);
	const int fd = rgport_mkstemp( tmp.data);
	fname = tmp.data;
	return fd;
}

std::string FileEngine::get_c_file( const std::string &p_fname,
	gmFileHandle &f)
{
	f.clear();
	if( p_fname.length() <= 0) return "No file specified";

	{ // If a file outside of the archives matches, then return it.
		FILE *inf = fopen( p_fname.c_str(), "r");
		if( inf) {
			f.set_c_file( inf);
			f.m_cfile_close_fn = hlp_gm_file_fclose;
			return std::string();
		}
	}

	// Load the file from one of the available archives.
	FILE *cfile = 0;
	int (*tmp_cfile_close_fn)( FILE *) = 0;
	for( int i = 0; i < int( m_media_files.size()); i++) {

		const int err = gm_get_from_archive(
			m_media_files[i],
			p_fname,
			f.m_str,
			m_max_archive_file_size,
			0,
			&cfile,
			&tmp_cfile_close_fn,
			this);
		if( 0 == err) break;
	}

	// If the file was loaded from an archive then return it.
	if( cfile) {
		f.set_c_file( cfile);
		f.m_cfile_close_fn = tmp_cfile_close_fn;
		return std::string();
	}

	return "Could not get file \"" + p_fname + "\" from current working " +
		"directory (cwd) or archives";
}

std::string FileEngine::get_file( const std::string &p_fname,
	rgPortFileHandle &f)
{
	f.clear();
	if( p_fname.length() <= 0) return "No file specified";

	{ // If a file outside of the archives matches, then return it.
		std::ifstream inf( p_fname.c_str());
		if( inf.is_open()) {
			f.m_inf.open( p_fname.c_str(), std::ios::binary);
			if( !(f.m_inf.is_open())) {
				return "Could not get unarchived file \"" +
					p_fname + "\"";
			}
			return std::string();
		}
	}

	// Load the file from one of the available archives.
	int err = -1;
	for( int i = 0; i < int( m_media_files.size()); i++) {
		err = gm_get_from_archive( m_media_files[i], p_fname, f.m_str,
			m_max_archive_file_size, &f, 0, 0, this);
		// FIXME: Not all versions of some archive utilities return
		//	errors WHEN THEY SHOULD. Thus if "0 == err" don't
		//	trust it and check to see if the file was "really"
		//	extracted.
		if(	(0 == err) &&
			( (f.m_str.tellp() > 0) || f.is_real_file())) break;
	}
	
	if( 0 == err) {
		// If archive file was loaded into buffer, then return it.
		if( f.m_str.tellp() > 0) return std::string();

		// If swap file (or direct file link) was used, then return it.
		if( f.is_real_file()) return std::string();
	}
	
	return "Could not get file \"" + p_fname + "\" from current working " +
		"directory (cwd) or archives";
}

inline bool hlp_is_name_good( const std::string &p_name)
{
	for( int i = 0; i < int( p_name.length()); i++) {
		if( 0 == isspace( p_name[i])) {
			switch( p_name[i]) {
			case '"': break;
			case '\\': break;
			default: return true;
			}
		}
	}
	return false;
}

std::string hlp_strip_leading_space( const std::string &s)
{
	std::string tmp;
	int i;
	for( i = 0; i < int( s.length()); i++)
		if( (' ' != s[i]) || ('\t' != s[i])) break;

	for( ; i < int( s.length()); i++)
		tmp += s[i];

	return tmp;
}

std::string hlp_strip_leading_dot_slash( const std::string &s)
{
	std::string tmp;
	if(	((s.length() > 2) &&
		('.' == s[0]) &&
		('/' == s[1]))) {
		for( int i = 2; i < int( s.length()); i++)
			tmp += s[i];
	} else if(
		((s.length() > 3) &&
		('.' == s[0]) &&
		('.' == s[1]) &&
		('/' == s[2]))) {
		for( int i = 3; i < int( s.length()); i++)
			tmp += s[i];
	} else tmp = s;
	return tmp;
}

FILE *hlp_get_from_tar_gz( const std::string &p_tar, const std::string &p_file)
{
	// XXX: If the file to get starts with "./" then those two characters
	//	must be removed since the tar file does not support those.
	std::string t_fname = hlp_strip_leading_dot_slash( p_file);

	if( !hlp_is_name_good( t_fname)) return 0;
	fflush( NULL);	// FIXME: Is flushing "NULL" really neccasary?
	return popen(
		("tar -Oxzf \"" + p_tar + "\" \"" + t_fname + "\"").c_str(),
		"r");
}

FILE *hlp_get_from_tar_bz( const std::string &p_tar, const std::string &p_file)
{
	// XXX: If the file to get starts with "./" then those two characters
	//	must be removed since the tar file does not support those.
	std::string t_fname = hlp_strip_leading_dot_slash( p_file);

	if( !hlp_is_name_good( t_fname)) return 0;
	fflush( NULL);	// FIXME: Is flushing "NULL" really neccasary?
	return popen(
		("tar -Oxjf \"" + p_tar + "\" \"" + t_fname + "\"").c_str(),
		"r");
}

FILE *hlp_get_from_zip( const std::string &p_zip, const std::string &p_file)
{
	// XXX: If the file to get starts with "./" then those two characters
	//	must be removed since the zip file does not support those.
	std::string t_fname = hlp_strip_leading_dot_slash( p_file);

	if( !hlp_is_name_good( t_fname)) return 0;
	fflush( NULL);	// FIXME: Is flushing "NULL" really neccasary?
	return popen(
		("unzip -p \"" + p_zip + "\" \"" + t_fname + "\"").c_str(),
		"r");
}

bool hlp_has_extension( const std::string &f, const std::string &ext)
{
	const int elen = ext.length();
	const int flen = f.length();
	if( flen <= elen) return false;
	int i = flen - elen;
	int j = 0;
	for( ; j < elen; i++, j++)
		if( ext[j] != f[i]) return false;
	return true;
}

int hlp_gm_file_fclose( FILE *f)
{
	if( !f) return -1;
	fclose( f);
	return 0;
}

int hlp_gm_file_pclose( FILE *f)
{
	if( !f) return -1;
	const int result = pclose( f);
	if( -1 == result) return -5;
	if( 0 == WIFEXITED( result)) return -6;
	return 0;
}

int gm_get_from_archive(
	const std::string &p_archive,
	const std::string &p_file,
	std::iostream &s,
	const int len,
	rgPortFileHandle *p_fh,
	FILE **c_file,
	int (**c_file_close_fn)( FILE *),
	FileEngine *fe)
{
	if( p_file.length() <= 0) return -1;

	// Test if archive exsits:
	// FIXME: Some extractors (unzip for one) do not return an error when
	//	a non-existant zip file is specified. Therefore the existance
	//	of an archive must be tested manually.
	{
		std::ifstream arc( p_archive.c_str());
		if( !arc.is_open()) return -444;
	}

	// Choose archive by extension
	FILE *f = 0;
	if( hlp_has_extension( p_archive, ".tar.gz")) {
		f = hlp_get_from_tar_gz( p_archive,
			hlp_strip_leading_space( p_file));

	} else if(	hlp_has_extension( p_archive, ".tar.bz") ||
			hlp_has_extension( p_archive, ".tar.bz2")) {
		f = hlp_get_from_tar_bz( p_archive,
			hlp_strip_leading_space( p_file));

	} else if( hlp_has_extension( p_archive, ".zip")) {
		f = hlp_get_from_zip( p_archive,
			hlp_strip_leading_space( p_file));
	}

	if( !f) return -3;
	if( c_file) {
		(*c_file) = f;
		(*c_file_close_fn) = hlp_gm_file_pclose;
		return 0;
	}

	// Load the uncompressed file into a high speed buffer.
	bool dump_to_swap = false;
	rgDWORD count = 0;
	const int buf_size = 64;
	rgTempArray <char> buf( buf_size);
	while( true) {
		const int nread = fread( buf.data, 1, buf_size, f);
		s.write( buf.data, nread);
		if( nread != buf_size) break;
		count += buf_size;
		if( count >= rgDWORD( len)) {
			dump_to_swap = true;
			break;
		}
	}

	if( dump_to_swap) {
		std::string tmp_file;
		std::fstream tmp_swap;
		const std::string emsg(
			fe->create_temporary_file( tmp_swap, tmp_file));
		if( emsg.length() > 0) return -777;
		// Write buffered portion to swap file.
		while( count > 0) {
			if( count >= rgDWORD( buf_size)) {
				s.read( buf.data, buf_size);
				tmp_swap.write( buf.data, buf_size);
				count -= buf_size;

			} else {
				s.read( buf.data, count);
				tmp_swap.write( buf.data, count);
				count = 0;
			}
		}
		s.clear();
		s.seekp( 0, std::ios::beg);

		// Write remainder of file to swap.
		// XXX: Must check if "tmp_swap" fails (out of swap space).
		while( true) {
			const int nread = fread( buf.data, 1, buf_size, f);
			tmp_swap.write( buf.data, nread);
			if( nread != buf_size) break;
		}

		{
			p_fh->m_inf.open( tmp_file.c_str());
			if( !(p_fh->m_inf.is_open())) return -888;
			p_fh->set_swap_file_name( tmp_file);
		}
	}

	return hlp_gm_file_pclose( f);
}
