/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <rgTime.h>
#include <rgNetwork.h>
#include <pthread.h>
#include <errno.h>
#include <string.h>
#include <string>
#include <SDL/SDL.h>
#include "main_header.h"
#include "engine_game.h"
#include "engine_file.h"
#include "engine_physics.h"
#include "engine_graphics.h"
#include "engine_sound.h"
#include "things/things.h"


ConsoleMessage gbl_cmesg;

GameEngine *GameEngine::m_root_master = 0;

GameEngine::GameEngine( const std::string &p_boot)
{
	m_shutdown = false;

	// XXX: Somethings are better shared between "root-master" and "slave"
	//	GameEngines. This includes things like MainData and the
	//	FileEngine.
	if( m_root_master) mdata = m_root_master->mdata;
	else mdata = new MainData( this);

	// Init file engine
	#ifdef ENGINE_FILE_SHARED_WITH_ROOT_MASTER
	if( m_root_master) file_engine = m_root_master->file_engine;
	else file_engine = new FileEngine( this);
	#else
	file_engine = new FileEngine( this);
	#endif
	
	// Init network engine
	m_use_console_server = false;
	m_console = new GameConsole( this);

	// Init timing engine
	m_game_timer = 0;
	//m_partial_game_clock = 0.0;
	m_fps_timer = new rgFPSCounter( 1);
	m_game_clock_scalar = 1.0;
	m_game_clock_hz = 100;

	// Init physics
	phy_engine = new PhysicsEngine( this);

	// Init graphics
	gfx_engine = new GraphicsEngine( this);

	// Init audio
	snd_engine = new SoundEngine( this);

	// Init thing manager
	m_thing_manager = 0;

	m_batch_limit = 32;

	// Initialize console commands
	max_definition_recursion = 64;
	// FIXME: Perhaps there are too many commands. Some commands can be
	//	compacted together. Though, this is not a big deal (yet).
	add_console_command( "audio", ccmd_audio);
	add_console_command( "add_archive_file", ccmd_add_archive_file);
	add_console_command( "add_bounding_thing", ccmd_add_bounding_thing);
	add_console_command( "add_camera", ccmd_add_camera);
	add_console_command( "add_to_collision_list",
		ccmd_add_to_collision_list);
	add_console_command( "apply_texture_alpha_mask",
		ccmd_apply_texture_alpha_mask);
	add_console_command( "batch", ccmd_batch);
	add_console_command( "batch_list", ccmd_batch_list);
	add_console_command( "bind", ccmd_bind);
	add_console_command( "bind_front", ccmd_bind_front);
	add_console_command( "bindings", ccmd_bindings);
	add_console_command( "build_sector_map", ccmd_build_sector_map);
	add_console_command( "clear_color", ccmd_clear_color);
	add_console_command( "clear_sector_map", ccmd_clear_sector_map);
	add_console_command( "color_material", ccmd_color_material);
	add_console_command( "compile_model_alpha_faces",
		ccmd_compile_model_alpha_faces);
	add_console_command( "create_collision_list",
		ccmd_create_collision_list);
	add_console_command( "cross", ccmd_cross);
	add_console_command( "cull_face", ccmd_cullface);
	add_console_command( "define", ccmd_define);
	add_console_command( "depth_test", ccmd_depth_test);
	add_console_command( "do", ccmd_do);
	add_console_command( "dot", ccmd_dot);
	add_console_command( "draw_sectors", ccmd_draw_sectors);
	add_console_command( "dump_all_things", ccmd_dump_all_things);
	add_console_command( "dump_camera", ccmd_dump_camera);
	add_console_command( "dump_model", ccmd_dump_model);
	add_console_command( "dump_sector_map", ccmd_dump_sector_map);
	add_console_command( "dump_thing", ccmd_dump_thing);
	add_console_command( "echo", ccmd_echo);
	add_console_command( "endif", ccmd_endif);
	add_console_command( "game_clock", ccmd_game_clock);
	add_console_command( "game_clock_scalar", ccmd_game_clock_scalar);
	add_console_command( "gfx_engine", ccmd_gfx_engine);
	add_console_command( "gravity", ccmd_gravity);
	add_console_command( "ifdef", ccmd_ifdef);
	add_console_command( "ifndef", ccmd_ifndef);
	add_console_command( "kill", ccmd_kill);
	add_console_command( "list_cameras", ccmd_list_cameras);
	add_console_command( "list_commands", ccmd_list_commands);
	add_console_command( "list_models", ccmd_list_models);
	add_console_command( "list_spawnable", ccmd_list_spawnable);
	add_console_command( "load_audio", ccmd_load_audio);
	add_console_command( "load_model", ccmd_load_model);
	add_console_command( "load_model_textures", ccmd_load_model_textures);
	add_console_command( "load_module", ccmd_load_module);
	add_console_command( "load_texture", ccmd_load_texture);
	add_console_command( "normalize", ccmd_normalize);
	add_console_command( "perspective", ccmd_perspective);
	add_console_command( "ping", ccmd_ping);
	add_console_command( "pwd", ccmd_pwd);
	add_console_command( "quit", ccmd_quit);
	add_console_command( "remove_camera", ccmd_remove_camera);
	add_console_command( "remove_model", ccmd_remove_model);
	add_console_command( "remove_thing", ccmd_remove_thing);
	add_console_command( "set_alt_texture", ccmd_set_alt_texture);
	add_console_command( "set_camera_position", ccmd_set_camera_position);
	add_console_command( "set_current_camera", ccmd_set_current_camera);
	add_console_command( "setenv", ccmd_setenv);
	add_console_command( "set_game_clock_hz", ccmd_set_game_clock_hz);
	add_console_command( "set_root_renderer", ccmd_set_root_renderer);
	add_console_command( "set_screen", ccmd_set_screen);
	add_console_command( "set_server_port_range",
		ccmd_set_server_port_range);
	add_console_command( "set_system", ccmd_set_system);
	add_console_command( "set_texture_colorkey",
		ccmd_set_texture_colorkey);
	add_console_command( "spawn", ccmd_spawn);
	add_console_command( "thing_list_check", ccmd_thing_list_check);
	add_console_command( "thing_manager", ccmd_thing_manager);
	add_console_command( "unbind", ccmd_unbind);
	add_console_command( "undef", ccmd_undef);
	add_console_command( "use_console_server", ccmd_use_console_server);
	add_console_command( "use_shared_vertices", ccmd_use_shared_vertices);
	add_console_command( "warn_only", ccmd_warn_only);
	add_console_command( "wireframe", ccmd_wireframe);

	#ifdef RPH_STATIC_BUILD
	add_console_command( "startup_static_thing_module",
		ccmd_startup_static_thing_module);

	add_definition( "rph_static_build", std::string());
	#endif

	// Message/batch system
	m_skip_batch_lines = false;
	m_engine_started = false;
	rcout( "Loading bootup file \"" + p_boot + "\"...\n");
	const int err = process_batch_file( p_boot, true);
	// FIXME: Don't use magic error numbers.
	switch( err) {
	case 0: rcout( "bootup file loaded\n"); break;
	case -1: rcout( "no bootup file present.\n"); break;
	default:
		std::cerr << "Error on line " << err <<
			": Failed to load boot file \"" << p_boot << "\".\n";
		m_shutdown = true;
		return;
	}

	if( m_use_console_server) {
		if( m_console->start_server_thread())
			std::cerr << "Could not start server thread\n";
	}

	// Finialize timing engine
	m_game_clock_tick = 1.0 / double( m_game_clock_hz);
	m_game_clock = 0;
	m_game_timer = new rgTimer( m_game_clock_tick);

	// The "root-master" game engine controls the low-level machince
	// interface. The "root-master" game engine is the first created
	// game engine and MUST be the last game engine destroyed.
	if( !m_root_master) {
		m_root_master = this;
		const std::string emsg( krn_initialize_system());
		if( emsg.length() > 0) {
			std::cerr << ("Failed to initialize system!\n" + emsg +
				"\n");
			m_shutdown = true;
			return;
		}

		if( krn_create_screen( *this)) {
			std::cerr << "Failed to create screen!\n";
			m_shutdown = true;
			return;
		}
	}
}

GameEngine::~GameEngine()
{
	clear();

	if( !m_root_master) {
		std::cerr <<
			"GameEngine::~GameEngine(): The root-master GameEngine"
			" was destroyed while this slave GameEngine( " << this
			<< ") still exists.\n";
		return;
	}

	if( this != m_root_master) return;
	if( mdata) delete mdata;
	SDL_Quit();
	m_root_master = 0;
}

void GameEngine::clear()
{
	if( m_game_timer) { delete m_game_timer; m_game_timer = 0;}
	if( m_fps_timer) { delete m_fps_timer; m_fps_timer = 0;}
	if( m_thing_manager) { delete m_thing_manager; m_thing_manager = 0;}
	if( gfx_engine) { delete gfx_engine; gfx_engine = 0;}
	if( phy_engine) { delete phy_engine; phy_engine = 0;}
	if( snd_engine) { delete snd_engine; snd_engine = 0;}

	#ifdef ENGINE_FILE_SHARED_WITH_ROOT_MASTER
	if( (this == m_root_master) && file_engine) delete file_engine;
	#else
	if( file_engine) delete file_engine;
	#endif
	file_engine = 0;

	if( m_console) { delete m_console; m_console = 0;}
}

int GameEngine::startup()
{
	gfx_engine->startup();
	create_thing_manager();
	m_engine_started = true;
	rcout( "loading startup file...\n");
	const int err = process_batch_file( m_startup_file, true);
	// FIXME: Don't use magic error numbers.
	int result = 0;
	switch( err) {
	case 0: rcout( "startup file loaded\n"); break;
	case -1: rcout( "startup: no startup file present."); break;
	default:
		rcout( "startup: "
			"Error on line " + rg_to_string( err) +
			": while loading startup file \"" + m_startup_file +
			"\".\n");
		result = -2;
		break;
	}
	
	// Game timing starts now.
	m_game_timer->reset();
	m_fps_timer->reset( *m_game_timer);
	return result;
}

int GameEngine::screen_width() const
{
	if( mdata && mdata->m_screen) return mdata->m_screen->w;
	return 0;
}

int GameEngine::screen_height() const
{
	if( mdata && mdata->m_screen) return mdata->m_screen->h;
	return 0;
}

// Return: "true" is returned if given (key, function) pair was added,
//	"false" is returned otherwise.
bool GameEngine::add_console_command( const std::string &key,
	int (*fn)( const std::vector <std::string> &, GameEngine &))
{
#define TMP_PROTO int (*)( const std::vector <std::string> &, \
	GameEngine &)
	typedef std::pair <const std::string, TMP_PROTO> SCPair;
	return (m_commands.insert( SCPair( key, fn))).second;
#undef TMP_PROTO
}

gmInputAction *GameEngine::binding( const std::string &key)
{
	std::map <const std::string, gmInputAction> :: iterator iter =
		m_key_binding.find( key);
	if( iter == m_key_binding.end()) return 0;
	return &(iter->second);
}

int GameEngine::add_binding( const std::string &key,
	const std::vector <std::string> &args,
	const std::string &up_down,
	bool to_front)
{
	gmInputAction *ia = binding( key);
	if( !ia) {
		typedef std::pair <const std::string, gmInputAction> SAPair;
		m_key_binding.insert( SAPair( key, gmInputAction()));
		ia = binding( key);
		if( !ia) return -1;
	}
	std::string tmp;
	for( int i = 0; i < int( args.size()); i++) tmp += args[i] + " ";
	std::vector <std::string> *p = &(ia->m_down);
	if( "up" == up_down) p = &(ia->m_up);
	if( to_front) p->insert( p->begin(), tmp);
	else p->push_back( tmp);
	return 0;
}

int GameEngine::unbind( const std::string &key)
{
	std::map <const std::string, gmInputAction> :: iterator iter =
		m_key_binding.find( key);
	if( iter == m_key_binding.end()) return -1;
	m_key_binding.erase( iter);
	return 0;
}

std::string GameEngine::dump_bound_keys()
{
	std::map <const std::string, gmInputAction> :: const_iterator iter =
		m_key_binding.begin();
	std::string tmp( "Key Bindings:\n");
	for( ; iter != m_key_binding.end(); ++iter)
		tmp += dump_binding( iter->first);
	tmp += "Total bound keys: " + gm_to_string( int(
		m_key_binding.size())) + "\n";
	return tmp;
}

std::string GameEngine::dump_binding( const std::string &key)
{
	gmInputAction *ia = binding( key);
	if( !ia) return std::string( "No bindings for key \"" + key + "\"\n");
	std::string tmp( "Key \"" + key + "\":\n");

	{	tmp += "\t[up] bindings:\n";
		for( int i = 0; i < int( ia->m_up.size()); i++)
			tmp += "\t\t" + ia->m_up[i] + "\n";
	}

	{	tmp += "\t[down] bindings:\n";
		for( int i = 0; i < int( ia->m_down.size()); i++)
			tmp += "\t\t" + ia->m_down[i] + "\n";
	}
	return tmp;
}

int GameEngine::exec_key( const std::string &key, const std::string &up_down)
{
	gmInputAction *ia = binding( key);
	if( !ia) return -1;
	if( "up" == up_down) {
		for( int i = 0; i < int( ia->m_up.size()); i++)
			process_message( ia->m_up[i]);

	} else {
		for( int i = 0; i < int( ia->m_down.size()); i++)
			process_message( ia->m_down[i]);
	}
	return 0;
}

int GameEngine::process()
{
	check_for_console_messages();

	rgLONG64 clk = game_clocks_pending();
	if( clk > 0) {
		int clocks_processed = 0;
		while( clk--) {
			m_thing_manager->process();

			m_game_clock++;
	//		if( !(game_clock() % m_game_clock_hz)) {
	//			std::cerr << "game second passed. "
	//				<< game_clock() << "\n";
	//		}
	
			// FIXME: Check console messages frequently enough to
			//	remain responsive. However, more events must be
			//	checked to keep the game
			//	responsive, implement a better system.
			if( (clocks_processed++) > 100) {
				check_for_console_messages();
				clocks_processed = 0;
			}
		}
		gfx_engine->process();
		m_fps_timer->inc_frame();
		if( m_fps_timer->count_reset) {
			std::cout << "fps = " << m_fps_timer->fps << "\n";
			m_fps_timer->count_reset = 0;
		}
	}

	if( m_shutdown) return -1;
	return 0;
}

rgLONG64 GameEngine::game_clocks_pending()
{
	// FIXME: Game clock scaling does not work properly if a scaling value
	//	greater than 1 is given. Also, small scaling values pause the
	//	game (since clocks pending is integral). Therefore think of
	//	a better implementation.
	if( m_game_timer->time_unit_passed()) {
		const double gc =
			(m_game_timer->elapsed_mili_sec() *
			double( m_game_clock_hz * m_game_clock_scalar))
			/ 1000.0;

		// XXX: Don't need to check for partial game clocks since one
		//	or two missed game clocks won't make much of a
		//	difference.
		// Get complete elapsed game clocks.
		//const rgLONG64 igc = rgLONG64( gc);

		// Keep track of partial elapsed game clocks.
		//m_partial_game_clock += gc - double( igc);
		//const rgLONG64 c = rgLONG64( m_partial_game_clock);
		//m_partial_game_clock -= double( c);

		// Return complete elpased game clocks plus the sum of
		// partial game clocks which make whole/complete game clocks.
		//return (igc + c);

		return rgLONG64( gc);

	} else return 0;
}

int GameEngine::load_startup_file( const char *fname)
{
	return 0;
}

int GameEngine::create_thing_manager()
{
	if( m_thing_manager) return 0;
	if( !(m_thing_manager = new ThingManager( this))) return -1;
	return 0;
}

void GameEngine::check_for_console_messages()
{
	if( gbl_cmesg.console_lock) return;

	std::vector <std::string> msg;
	std::string tmp;
	if( m_pending_console_cmd.length() > 0)
		tmp = m_pending_console_cmd;
	for( int i = 0; i < gbl_cmesg.length; i++){
		if( '\0' == gbl_cmesg.msg[i]) {
			if( tmp.length() > 0) {
				msg.push_back( tmp);
				tmp.clear();
			}
		} else tmp += gbl_cmesg.msg[i];
	}
	if( tmp.length() > 0) m_pending_console_cmd = tmp;
	else m_pending_console_cmd.clear();

	while( !msg.empty()) {
		//std::cerr << "ConMsg: \"" << msg[0] << "\"\n";
		process_message( msg[0]);
		msg.erase( msg.begin());
	}

	gbl_cmesg.console_lock = true;
	gbl_cmesg.game_lock = false;
}

GameConsole::GameConsole( GameEngine *p_ge)
{
	ge = p_ge;
	m_sock = new rgSocket();
	m_first_port = 9000;
	m_last_port = 16000;
	m_server_started = false;
}

GameConsole::~GameConsole()
{
	if( m_server_started) {
/*
		std::cout << "Shutting down console server\n";
		std::cout.flush();

		// FIXME: Must shutdown console server thread, otherwise the
		// program will segfault when "this"
		// (GameConsole::~GameConsole) returns because its memory will
		// still be accessed by the console server thread.
		// pthread_cancel and pthread_join are not working.
		std::cerr << "Waiting for server thread to terminate...";
		int err = pthread_cancel( m_server_thread);
		if( err) {
			std::cerr << "Error " << err <<
				" while canceling server thread.\n";
		}

		if( (err = pthread_join( m_server_thread, 0))) {
			std::cerr << "Error " << err <<
				" while joining server thread.\n";

		} else std::cerr << "done\n";
*/
	}

	// FIXME: Remove the below "exit".
	exit(0);
	//if( m_sock) delete m_sock;
}

int GameConsole::start_server_thread()
{
	int result = pthread_create( &m_server_thread, NULL,
		console_server_thread, this);
	if( result) return -1;
	m_server_started = true;
	return 0;
}

void ConsoleMessage::print_message( std::ostream &o)
{
	std::string tmp;
	for( int i = 0; i < length; i++) {
		if(	((msg[i] >= 'A') && (msg[i] <= 'z')) ||
			((msg[i] >= ' ') && (msg[i] <= '@')) ||
			(msg[i] == '\t') ||
			(msg[i] == '\n')) tmp += msg[i];
		else tmp += '.';
	}
	o << tmp;
}

int hlp_get_message( GameConsole *gc)
{
	while( gbl_cmesg.game_lock) sleep( 1);
	gbl_cmesg.length = gc->m_sock->receive( gbl_cmesg.msg,
		MAX_CONSOLE_MESSAGE_LENGTH - 1);
	if( -1 == gbl_cmesg.length) {
		std::cerr << "Error \"" << strerror( errno) <<
			"\" while receiving console message\n";
		sleep( 1);	// XXX: Required to prevent the console
				// messaging thread from hogging the cpu.
		return errno;
	}

	gbl_cmesg.game_lock = true;
	gbl_cmesg.console_lock = false;
	return 0;
}

void *console_server_thread( void *ptr)
{
	GameConsole *gc = (GameConsole *)ptr;
	int start_port = gc->m_first_port;
	while(	gc->m_sock->open_listen_port( start_port) &&
		(start_port < gc->m_last_port)) start_port++;
	if( start_port >= gc->m_last_port) {
		std::cerr << "Could not find a free port number for server\n";
		return 0;
	}
	std::cout << "Started console server on port " << start_port << "\n";
	std::cout.flush();

	if( -1 == gc->m_sock->accept()) {
		std::cerr << "Error: \"" << strerror( errno) <<
			"\" while accepting console server connections\n";
		return 0;
	}
	std::cout << "Remote console connected\n";
	std::cout << "Console located at " << gc->m_sock->remote_ip() << "\n";
	std::cout.flush();

	while( !(hlp_get_message( gc)));
	std::cerr << "Remote console terminated\n";
	return 0;
}

