/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "engine_game.h"
#include "engine_physics.h"
#include "engine_graphics.h"
#include "things/things.h"
#include <rg3DRom.h>
#include <rg3DRom_physics.h>
#include <rg3DRom_csg.h>

std::string PhyCollisionFragment::dump() const
{
	return std::string(
		"Fragment: prisim index = " + gm_to_string( pid) +
			", prisim = " + gm_to_string( prisim) + " {\n"
		"\tradius = " + gm_to_string( radius) + "\n"
		"\tsurface normal = " + gm_to_string( surface_normal) + "\n"
		"\tsurface point = " + gm_to_string( surface_point) + "\n"
		"\tcollision point = " + gm_to_string( collision_point) + "\n"
		"\tintrusion vector = " + gm_to_string( intrusion_vector) +
		"\n}"
		);
}

std::string PhyDynResult::dump() const
{
	std::string tmp;
	for( int i = 0; i < int( m_point_hit.size()); i++)
		tmp += (m_point_hit[i]).dump() + "\n";
	tmp += "Collision points dumped: " +
		gm_to_string( int( m_point_hit.size())) + "\n";
	return tmp;
}

rgVector PhyDynResult::collision_center( const PhyCollisionContext &cc) const
{
	if( m_point_hit.empty()) return rgVector();
	rgVector com;
	for( int i = 0; i < int( m_point_hit.size()); i++)
		com += cc.m_bpoint[ (m_point_hit[i]).pid];
	return com * (1.0f / float( m_point_hit.size()));
}

void PhySector::draw( const int level) const
{
	bool has_sub_sectors = false;
	if( level < 0) return;
	{
		for( int i = 0; i < PHY_MAX_SUB_SECTORS; i++) {
			if( m_sec[i]) {
				(m_sec[i])->draw( level - 1);
				has_sub_sectors = true;
			}
		}
	}


	glBegin( GL_QUADS);
	glColor3f( 0, 0, 1);
	{
		for( int i = 0; i < int( m_prisim.size()); i++)
			(m_prisim[i])->render();
	}
	glEnd();

	if( has_sub_sectors) {
		glBegin( GL_LINES);
		glColor3f( 1, 0, 0);
		glVertex( (v_a() + v_b()) * 0.5f);
		glVertex( (v_c() + v_d()) * 0.5f);

		glVertex( (v_a() + v_d()) * 0.5f);
		glVertex( (v_b() + v_c()) * 0.5f);
		glEnd();
	}
}

bool PhySector::create_child_sector( int n)
{
	if( (n < 0) || (n >= PHY_MAX_SUB_SECTORS)) return false;
	if( m_sec[n]) return false;
	rgVector v[4];
	switch( n) {
	case 0:	v[0] = v_a();
		v[1] = (v_a() + v_b()) * 0.5f;
		v[2] = center();
		v[3] = (v_a() + v_d()) * 0.5f;
		break;

	case 1:	v[0] = (v_a() + v_b()) * 0.5f;
		v[1] = v_b();
		v[2] = (v_b() + v_c()) * 0.5f;
		v[3] = center();
		break;

	case 2:	v[0] = center();
		v[1] = (v_b() + v_c()) * 0.5f;
		v[2] = v_c();
		v[3] = (v_c() + v_d()) * 0.5f;
		break;

	case 3:	v[0] = (v_a() + v_d()) * 0.5f;
		v[1] = center();
		v[2] = (v_c() + v_d()) * 0.5f;
		v[3] = v_d();
		break;

	default: return false;
	}
	if( !(m_sec[n] = new PhySector( v[0], v[1], v[2], v[3]))) return false;
	return true;
}

bool PhySector::point_collides( const PhyBoundingPoint p,
	PhyCollisionFragment &f) const
{
	const PhySector *s = containing_quadrant( p);
	if( s) return s->point_collides( p, f);
	for( int i = 0; i < int( m_prisim.size()); i++) {
		if( (m_prisim[i])->collision(
			p.m_origin,
			p,
			p.radius,
			f.surface_normal,
			f.surface_point,
			f.collision_point,
			f.intrusion_vector)) {

			f.src_point = p.m_origin;
			f.des_point = p;
			f.radius = p.radius;
			f.prisim = m_prisim[i];
			return true;
		}
	}
	return false;
}

/*
PhyCollisionPrisim *PhySector::point_collides( const PhyBoundingPoint p) const
{
	const PhySector *s = containing_quadrant( p);
	if( s) return s->point_collides( p);

	// Point is in this quadrant, therefore check all prisims in this
	// quadrant to see if the point collides with them.
	// XXX: NOTE, there is no concept of "nearest" collision.
//	for( int i = 0; i < int( m_prisim.size()); i++)
//		if( (m_prisim[i])->collision( p, p.radius)) return m_prisim[i];
	for( int i = 0; i < int( m_prisim.size()); i++) {
		if( (m_prisim[i])->collision( p.m_origin, p, p.radius))
			return m_prisim[i];
	}
	return 0;
}
*/

const PhySector *PhySector::containing_quadrant( const rgVector p) const
{
	const rgVector cp = center();
	if( p.x >= cp.x) { // Possibly in quadrant "A" or "D".
		if( p.z <= cp.z) {
			// Possibly in quadrant "A", but assume quad "A".
			return const_sec_a();

		} else {
			// Possibly in quadrant "D", but assume quad "D".
			return const_sec_d();
		}

	} else { // Possibly in quadrant "B" or "C".
		if( p.z <= cp.z) {
			// Possibly in quadrant "B", but assume quad "B'.
			return const_sec_b();

		} else {
			// Possibly in quadrant "C", but assume quad "C".
			return const_sec_c();
		}
	}
}

void PhySectorMap::clear()
{
	if( m_root) { delete m_root; m_root = 0;}
	m_prisim.clear();
}

void PhySectorMap::draw( const int level, const int p_line_width) const
{
	if( !m_root) return;
	// Set wireframe rendering
	GLint front_back[2];
	const GLint &tmp_front = front_back[0];
	const GLint &tmp_back = front_back[1];
	glGetIntegerv( GL_POLYGON_MODE, front_back);
	glPolygonMode( GL_FRONT_AND_BACK, GL_LINE);

	// No depth test
	GLboolean depth_test;
	glGetBooleanv( GL_DEPTH_TEST, &depth_test);
	glDisable( GL_DEPTH_TEST);

	// No lighting
	GLboolean lighting;
	glGetBooleanv( GL_LIGHTING, &lighting);
	glDisable( GL_LIGHTING);

	// No blend
	GLboolean blend;
	glGetBooleanv( GL_BLEND, &blend);
	glDisable( GL_BLEND);

	// No texturing
	GLboolean texture;
	glGetBooleanv( GL_TEXTURE_2D, &texture);
	glDisable( GL_TEXTURE_2D);

	// Line width
	GLint line_width;
	glGetIntegerv( GL_LINE_WIDTH, &line_width);
	glLineWidth( p_line_width);

	m_root->draw( level);

	// Outside border of quad tree (sector map)
	glBegin( GL_LINE_LOOP);
	glColor3f( 1, 0, 0);
	glVertex( m_root->v_a());
	glVertex( m_root->v_b());
	glVertex( m_root->v_c());
	glVertex( m_root->v_d());
	glEnd();

	// Restore original OpenGL state
	if( depth_test) glEnable( GL_DEPTH_TEST);
	if( lighting) glEnable( GL_LIGHTING);
	if( blend) glEnable( GL_BLEND);
	if( texture) glEnable( GL_TEXTURE_2D);
	glLineWidth( line_width);
	glPolygonMode( GL_FRONT, tmp_front);
	glPolygonMode( GL_BACK, tmp_back);
}

bool PhySectorMap::point_collides( const PhyBoundingPoint p,
	PhyCollisionFragment &f) const
{
	if( !m_root) return false;
	return m_root->point_collides( p, f);
}


//PhyCollisionPrisim *PhySectorMap::point_collides( const PhyBoundingPoint p)
//	const
//{
//	if( !m_root) return 0;
//	return m_root->point_collides( p);
//}

PhysicsEngine::PhysicsEngine( GameEngine *ge)
{
	game_engine = ge;	// XXX: Must be done first.

	m_gravity = rgVector( 0, -9.81f, 0) * (1.0f / gravity_frame_scalar());

}

PhysicsEngine::~PhysicsEngine()
{
	delete_all_collision_lists();
}

PhyDynResult PhysicsEngine::check_model( gmPhyModel *pm, sk3DRomSkelFrame &src,
	sk3DRomSkelFrame &des)
{
	PhyDynResult result;
	const int bcount = (pm->skel())->bone_count();

	// Check which bones collided
	bool col_bone[ bcount];
	if( !check_for_collision( bcount, pm->skel(), col_bone)) return result;

	// There was a collision
	sk3DRomSkelFrame tmp = des;

	float weight = 0.5f;
	float w = 0.5f;
	int resolution = 8;
	while( resolution--) {
		for( int i = 0; i < bcount; i++) {
			if( col_bone[i]) {
				rgCoordSystem::interpolate_csys( src.b[i].rcs,
					des.b[i].rcs, tmp.b[i].rcs, weight);
			}
		}
		w *= 0.5f;
		if( check_for_collision( bcount, pm->skel(), col_bone))
			weight += w;
		else weight -= w;
	}

	// Update the final skeleton state. Bones that never collided will be
	// set to "des". Bones that collided but had the collision resolved
	// will be set to "tmp". Bones that collided and never had their
	// collisions resolved will be set to "src" (unmoved).
	for( int i = 0; i < bcount; i++) {
		if( col_bone[i]) des.b[i].rcs = src.b[i].rcs;
		else des.b[i].rcs = tmp.b[i].rcs;
	}
	return result;
}

bool PhysicsEngine::check_for_collision( const int bcount,
	const sk3DRomPhySkel *s, bool *bcol)
{
	bool collision = false;
	for( int i = 0; i < bcount; i++) {
		if( !(bcol[i]) && hit( s->bone( i))) {
			// Bone collided, mark it for retransformation. Also,
			// mark every bone in the master list for
			// retransformation as well.
			const std::vector <short> &mlist =
				*(s->get_master_list( i));
			for( int j = 0; j < int( mlist.size()); j++)
				bcol[ mlist[j]] = true;
			bcol[i] = true;
			collision = true;
			
		} else bcol[i] = false;
	}
	return collision;
}

int PhysicsEngine::hit( const sk3DRomPhyBone &b) const
{
	return 0;
}

bool PhySectorMap::add_prisims( gmModelThing *m, const std::string &grp)
{
	if( m_root) return false;	// Must delete current sec map first
	if( !m) return false;
	ModelHandle *mh = m->model_handle();
	const int gid = mh->group_id( grp);
	if( !(m->model())) return false;
	sk3DRomModel *mdl = m->model();
	if( !mdl->m_csg) return false;
	for( int i = 0; i < mdl->m_csg->brush_count(); i++) {
		rgCSGBrush *b = mdl->m_csg->brush( i);
		if(	(RG3DROM_CSG_CUBE == b->type()) &&
			(gid == b->group_id())) {
			std::vector <rgVector> verts;
			b->gen_verts( verts);
			PhyCollisionPrisim tmp;
			rgMatrix mat;
			if( m->draw_context())
				mat = (m->draw_context())->mat();
			m_prisim.push_back( tmp.create( verts, mat));
		}
	}
	return true;
}

std::string PhySector::dump( const int p_depth)
{
	std::string tab;
	{ for( int i = 0; i < p_depth; i++) tab += "\t";}
	std::string tmp( tab + "Sub Sector: " + gm_to_string( this) +
		", depth = " + gm_to_string( p_depth) + "\n");
	tmp += tab + "Dimensions:\n" +
		tab + "\t" + gm_to_string( v_a()) + "\n" +
		tab + "\t" + gm_to_string( v_b()) + "\n" +
		tab + "\t" + gm_to_string( v_c()) + "\n" +
		tab + "\t" + gm_to_string( v_d()) + "\n" +
		tab + "\tcenter = " + gm_to_string( center()) + "\n" +
		tab + "\twidth = " + gm_to_string( m_width) + "\n" +
		tab + "\tdepth = " + gm_to_string( m_depth) + "\n";
	tmp += tab + "Total prisims = " + gm_to_string(
		int( m_prisim.size())) + "\n";
	tmp += tab + "sector a = " + gm_to_string( sec_a()) + "\n";
	if( sec_a()) tmp += (sec_a())->dump( p_depth + 1);

	tmp += tab + "sector b = " + gm_to_string( sec_b()) + "\n";
	if( sec_b()) tmp += (sec_b())->dump( p_depth + 1);

	tmp += tab + "sector c = " + gm_to_string( sec_c()) + "\n";
	if( sec_c()) tmp += (sec_c())->dump( p_depth + 1);

	tmp += tab + "sector d = " + gm_to_string( sec_d()) + "\n";
	if( sec_d()) tmp += (sec_d())->dump( p_depth + 1);

	return tmp;
}

std::string PhySectorMap::dump()
{
	std::string tmp( "Sector Map Dump:\n"
		"\tTotal prisims: " + gm_to_string( int( m_prisim.size())) +
			"\n"
		);
	if( !m_root) tmp += "\troot sector = none (not built)\n";
	else tmp += "\troot sector = " + gm_to_string( m_root) + "\n";
	
	tmp += m_root->dump( 1);

	return tmp;
}

bool PhySectorMap::build_map( rgVector a, rgVector b, rgVector c, rgVector d)
{
	if( m_root) delete m_root;
	m_root = new PhySector( a, b, c, d);
	if( !m_root) return false;

	std::vector <PhyCollisionPrisim *> tmp;
	for( int i = 0; i < int( m_prisim.size()); i++)
		tmp.push_back( &(m_prisim[i]));
	m_root->sub_divide( tmp, 8);
	return true;
}

void PhySector::sub_divide( const std::vector <PhyCollisionPrisim *> &p_prisim,
	const int level)
{
	add_contained_prisims( p_prisim);
	if( level <= 0) return;
	if( m_prisim.size() <= 4) return;
	for( int i = 0; i < PHY_MAX_SUB_SECTORS; i++) {
		create_child_sector( i);
		if( m_sec[i]) (m_sec[i])->sub_divide( m_prisim, level - 1);
	}
	m_prisim.clear();
}

void PhySector::add_contained_prisims( const std::vector <PhyCollisionPrisim *>
	&p_prisim)
{
	PhyCollisionPrisim my_sector;
	my_sector.create( m_width, 32768, m_depth);
	// Center prisim over sector
	my_sector.translate( v_b() - my_sector.m_v[4]);
//	std::cerr << "my sector = " << my_sector << "\n";

	for( int i = 0; i < int( p_prisim.size()); i++) {
//		std::cerr << "sec " << i << " = " <<
//			(*(p_prisim[i])) << "\n";
		if( my_sector.collision( *(p_prisim[i])))
			m_prisim.push_back( p_prisim[i]);
	}
}

bool PhysicsEngine::environment_collision( PhyDynResult &result,
	const PhyCollisionContext &cc) const
{
	result.m_point_hit.clear();
	const int count = int( cc.m_bpoint.size());
	for( int i = 0; i < count; i++) {
		PhyCollisionFragment f;
		if( m_sector_map.point_collides( cc.m_bpoint[i], f)) {
			f.pid = i;
			result.m_point_hit.push_back( f); 
		}
	}
	if( result.m_point_hit.size() > 0) return true;
	return false;
}

float PhysicsEngine::quadratic_frame_scalar() const
{
	const float fps = float( game_engine->game_clock_hz());
	if( fps != 1.0f) return (fps * ((fps * 0.5f) - 0.5f));
	else return 1.0f;
}

/*
PhysicsEngine::PhysicsEngine( GameEngine *ge)
{
	m_sector_width = SECTOR_TABLE_WIDTH;
	m_sector_height = SECTOR_TABLE_HEIGHT;

	m_gravity = rgVector( 0, -9.81f, 0);

	game_engine = ge;
	init_sector_table();
}

PhysicsEngine::~PhysicsEngine()
{
	if( chunk_table) delete chunk_table;
}

void PhysicsEngine::init_sector_table()
{
	for( int i = 0; i < SECTOR_TABLE_RESOLUTION; i++)
		m_sector_table[i] = -1;
}

void PhysicsEngine::set_chunk_table( int count)
{
	if( count <= 0) return;
	if( chunk_table) { delete chunk_table; chunk_table = 0;}
	m_chunk_table_size = 0;
	chunk_table = new PhyChunk[ count];
	if( !chunk_table) return;
	m_chunk_table_size = count;
	init_sector_table();
}

// 1. Check which bone-chunks collied with chunks already in sectors.
// 2. If no bone-chunks collied then return.
// 3. Create a new skeleton frame that tries to de-collide the model.
// 4. If the model still collides, then repeat step 3.
// 5. If the model does no collide then try to collide the model again.
// 6. If the model collides then repeat step 3, but if the maximum number
//    of adjustments have been made, then return the "best" no-colliding
//    result generated. If all of the generated skeletons collide, then
//    return the source skeleton.
//
//	src - Skeleton representing the model in a non-colliding state.
//	des - Skeleton representing the desired skeleton (which may or may
//	      not be in a colliding state).
void PhysicsEngine::check_model( gmPhyModel *pm, sk3DRomSkelFrame &src,
	sk3DRomSkelFrame &des)
{
	// Determine if "pm" is valid and get its physics ticket.
	std::map <gmPhyModel *, const PhyTicket> ::
		const_iterator iter = m_things.find( pm);
	if( iter == m_things.end()) return;
	const PhyTicket &pt = (iter->second);
	const int bcount = (pm->skel())->bone_count();

	// Check which bones collided
	bool col_bone[ bcount];
	if( !check_for_collision( bcount, pt.chunk_base, pm->skel(),
		col_bone)) return;

	// There was a collision
	sk3DRomSkelFrame tmp = des;

	float weight = 0.5f;
	float w = 0.5f;
	int resolution = 8;
	while( resolution--) {
		for( int i = 0; i < bcount; i++) {
			if( col_bone[i]) {
				interpolate_bone_csys( weight, src.b[i].rcs,
					des.b[i].rcs, tmp.b[i].rcs);
			}
		}
		w *= 0.5f;
		if( check_for_collision( bcount, pt.chunk_base, pm->skel(),
			col_bone)) weight += w;
		else weight -= w;
	}

	// Update the final skeleton state. Bones that never collided will be
	// set to "des". Bones that collided but had the collision resolved
	// will be set to "tmp". Bones that collided and never had their
	// collisions resolved will be set to "src" (unmoved).
	for( int i = 0; i < bcount; i++) {
		if( col_bone[i]) des.b[i].rcs = src.b[i].rcs;
		else des.b[i].rcs = tmp.b[i].rcs;
	}
}

bool PhysicsEngine::check_for_collision( const int bcount,
	const int chunk_base, const sk3DRomPhySkel *s, bool *bcol)
{
	bool collision = false;
	for( int i = 0; i < bcount; i++) {
		if( !(bcol[i]) && sec_hit( chunk_base + i, s->bone( i))) {
			// Bone collided, mark it for retransformation. Also,
			// mark every bone in the master list for
			// retransformation as well.
			const std::vector <short> &mlist =
				*(s->get_master_list( i));
			for( int j = 0; j < int( mlist.size()); j++)
				bcol[ mlist[j]] = true;
			bcol[i] = true;
			collision = true;
			
		} else bcol[i] = false;
	}
	return collision;
}

int PhysicsEngine::sec_hit( int id, const sk3DRomPhyBone &b) const
{
	const int secid = get_sector( b);
	if( secid < 0) return 0;			// Not in a sector
	if( m_sector_table[ secid] < 0) return 0;	// In empty sector
	// Must check for collision with all bone-chunks in this sector.
	// As well, must check "increasing-adjacent" sectors.
	int i = m_sector_table[ secid];
	while( i >= 0) {
		if(	(m_chunk_table[i].bone) &&
			b.collision( *(m_chunk_table[i].bone))) return 1;
		i = m_chunk_table[i].next;
	}
	return 0;
}

int PhysicsEngine::get_sector( const sk3DRomPhyBone &b) const
{
	rgPhyBoundSphere sph = b.bone_sphere();
	int x = int( sph.center.x - radius) - x_offset;
	int z = int( sph.center.z - radius) - z_offset;
	const int secid = z * m_sector_width + x;
	if( secid >= SECTOR_TABLE_RESOLUTION) return -1;
	return secid;
}

int PhysicsEngine::sectorize( int id, const sk3DRomPhyBone &b)
{
	return 0;
}
*/

void PhysicsEngine::thing_collision(
	std::vector <PhyCollisionFragment> &p_f,
	gmPhyModel *t,
	const std::string &p_list)
{
	p_f.clear();
	if( !t) return;
	Map_String_TCB *tmp = get_collision_list( p_list);
	if( !tmp) return;
	std::map <const std::string, ThingControlBlock *> :: const_iterator
		iter = tmp->begin();
	PhyCollisionFragment f;
	int count = int( tmp->size());
	while( count--) {
		ThingControlBlock *tcb = iter->second;
		++iter;
		if(	!(tcb->is_free()) &&
			(tcb->m_thing->is_a( "gmPhyModel")) &&
			(tcb->m_thing != t)) {
			if( t->collision( f, (gmPhyModel *)(tcb->m_thing)))
				p_f.push_back( f);
		}
	}
}

std::string PhysicsEngine::create_collision_list( const std::string &p_list)
{
	typedef std::pair <const std::string, Map_String_TCB *> SMPair;
	Map_String_TCB *tmp = new Map_String_TCB;
	if( !tmp) return "could not create Map_String_TCB";
	if( false == (m_col_list.insert( SMPair( p_list, tmp))).second) {
		delete tmp;
		return "duplicate list \"" + p_list + "\"";
	}
	return std::string();
}

void PhysicsEngine::delete_collision_list( const std::string &p_list)
{
	std::map <const std::string, Map_String_TCB *> :: iterator iter =
		m_col_list.find( p_list);
	delete_collision_list( iter);
}

void PhysicsEngine::delete_collision_list(
	std::map <const std::string, Map_String_TCB *> :: iterator &iter)
{
	if( iter == m_col_list.end()) return;
	if( iter->second) delete (iter->second);
	m_col_list.erase( iter);
}

Map_String_TCB *PhysicsEngine::get_collision_list( const std::string &p_list)
{
	std::map <const std::string, Map_String_TCB *> :: const_iterator iter =
		m_col_list.find( p_list);
	if( iter == m_col_list.end()) return 0;
	return (iter->second);
}

std::string PhysicsEngine::add_to_collision_list( const std::string &p_list,
	ThingControlBlock *p_tcb)
{
	if( !p_tcb) return "p_tcb == NULL";
	if( p_tcb->is_free()) return "thing is free";
	if( !(p_tcb->m_thing->is_a( "gmPhyModel")))
		return "not a gmPhyModel";
	Map_String_TCB *tmp = get_collision_list( p_list);
	if( !tmp) return "no list \"" + p_list + "\"";
	typedef std::pair <const std::string, ThingControlBlock *> STPair;
	if( false == (tmp->insert( STPair( p_tcb->m_key, p_tcb))).second)
		return "duplicate thing \"" + p_tcb->m_key + "\"";
	return std::string();
}

void PhysicsEngine::remove_from_collision_list(
	const std::string &p_list, ThingControlBlock *p_tcb)
{
	Map_String_TCB *tmp = get_collision_list( p_list);
	if( !tmp) return;
	std::map <const std::string, ThingControlBlock *> :: iterator iter =
		tmp->find( p_list);
	if( iter == tmp->end()) return;
	tmp->erase( iter);
}

void PhysicsEngine::delete_all_collision_lists()
{
	while( !(m_col_list.empty())) {
		std::map <const std::string, Map_String_TCB *> :: iterator iter
			= m_col_list.begin();
		delete_collision_list( iter);
	}
}
