/*
Copyright (C) 2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <stdlib.h>
#include <iostream>
#include <string>
#include <stdio.h>
#include <sstream>
#include <fstream>
#include <rgPort.h>
#include <rgStandard.h>
#include <SDL/SDL.h>
#include <SDL/SDL_audio.h>
#include <SDL/SDL_mixer.h>
#include "engine_game.h"
#include "engine_sound.h"
#include "engine_file.h"

void SoundEngine_mixer( void *unused, Uint8 *stream, int len);

gmAudioOps *gbl_audio_ops = 0;
std::map <const std::string, gmAudioOps *> gmAudio::m_audio_format;

gmAudio::gmAudio( const std::string &p_name)
{
	m_name = p_name;
	m_aops = 0;

	// Prepare the audio formats.
	if( gbl_audio_ops) return;

	gbl_audio_ops = new gmAudioOps[3];
	if( !gbl_audio_ops) {
		std::cerr << "Could not allocate memory for "
			"\"gbl_audio_ops\"\n";
		return;
	}

	typedef std::pair < const std::string, gmAudioOps *> SAPair;
	m_audio_format.clear();

	gmAudioOps *aop = gbl_audio_ops;

	// Chunk files (for mixing?)
	aop->load = gmAudio_load_chunk;
	aop->close = gmAudio_close_chunk;
	aop->mix_play = gmAudio_mix_play_chunk;
	aop->mix_halt = gmAudio_mix_halt_chunk;
	m_audio_format.insert( SAPair( "chunk", aop++));

	// Sound effect (WAV) files.
	aop->load = gmAudio_load_sound;
	aop->close = gmAudio_close_sound;
	aop->convert_to_format = gmAudio_convert_sound_to_format;
	m_audio_format.insert( SAPair( "sound", aop++));

	// Music (MUS, MID, OGG, ...)
	aop->load = gmAudio_load_music;
	aop->close = gmAudio_close_music;
	m_audio_format.insert( SAPair( "music", aop++));
}

gmAudio::~gmAudio()
{
	gmAudio::close();
}

std::string gmAudio::load( FILE *f, const std::string &p_type,
	SoundEngine *p_se)
{
	if( m_aops) gmAudio::close();
	if( !f) return "null FILE pointer passed";

	m_aops = get_audio_ops( p_type);
	if( !m_aops) return "\"" + p_type + "\" is not a valid audio format\"";

	// FIXME: What should be passed in for "autoclose" (2nd parameter of
	//	function "SDL_RWFromFP( FILE *fp, int autoclose)". Currently,
	//	0 is passed (because ONLY "gmFileHandle" should close the
	//	file, and assuming 0 means "no autoclose").
	if( !(m_aops->load)) return "not implemented/supported";
	return m_aops->load( SDL_RWFromFP( f, 0), this, p_se);
}

gmAudioOps *gmAudio::get_audio_ops( const std::string &key)
{
	Map_String_AudioOps :: const_iterator iter = m_audio_format.find( key);
	if( iter == m_audio_format.end()) return 0;
	return (iter->second);
}

std::string gmAudio::mix_play( int channel, int loop, SoundEngine *p_se)
{
	if( !SoundEngine::audio_opened_and_ready())
		return "SDL_OpenAudio/Mix_OpenAudio not ready";
	const std::string cas( check_aop_state());
	if( cas.length() > 0) return ("mix_play:: " + cas);
	if( m_aops->mix_play)
		return (m_aops->mix_play( this, p_se, channel, loop));
	return (m_name + ", mix_play not implemented/supported");
}

std::string gmAudio::mix_halt( SoundEngine *p_se)
{
	if( !SoundEngine::audio_opened_and_ready())
		return "SDL_OpenAudio/Mix_OpenAudio not ready";
	const std::string cas( check_aop_state());
	if( cas.length() > 0) return ("mix_halt:: " + cas);
	if( m_aops->mix_halt)
		return (m_aops->mix_halt( this, p_se));
	return (m_name + ", mix_halt not implemented/supported");
}

/////////////////////////////////////////////////////////////////////////////
// Sound
/////////////////////////////////////////////////////////////////////////////
std::string gmAudio_load_sound( SDL_RWops *rwops, gmAudio *aud,
	SoundEngine *p_se)
{
	if( !SDL_LoadWAV_RW( rwops,
		0,
		&(aud->m_sound.m_audio),
		&(aud->m_sound.wav_buffer),
		&(aud->m_sound.wav_length))) {
		return (std::string( SDL_GetError()));
	}
	return aud->convert_to_format( p_se);
}

std::string gmAudio_close_sound( gmAudio *aud)
{
	aud->m_sound.close();
	return std::string();
}

std::string gmAudio_convert_sound_to_format( gmAudio *aud, SoundEngine *p_se)
{
	// XXX: The SoundEngine that owns the audio, dictates the audio
	//	output format.
	SoundEngine *se = SoundEngine::audio_opened_and_ready();
	if( !se) {
		return ("Audio not opened, SoundEngine not ready, thus audio "
			"output format is undefined, cannot convert.");
	}

	if( !aud->m_sound.wav_buffer) return "no wave to convert";
	SDL_AudioCVT &cvt = aud->m_sound.m_cvt;
	if( -1 == SDL_BuildAudioCVT(
		&cvt,
		aud->m_sound.m_audio.format,
		aud->m_sound.m_audio.channels,
		aud->m_sound.m_audio.freq,
		se->output_format(),
		se->num_output_channels(),
		se->output_rate()))
		return (SDL_GetError());

	cvt.buf = new Uint8[ aud->m_sound.wav_length * cvt.len_mult];
	if( !(cvt.buf)) return "could not create AudioCVT buffer";
	memcpy( cvt.buf, aud->m_sound.wav_buffer, aud->m_sound.wav_length);
	cvt.len = aud->m_sound.wav_length;
	if( -1 == SDL_ConvertAudio( &cvt))
		return (SDL_GetError());
	// Free old wav data since it is no longer needed.
	aud->m_sound.free_raw_wav();
	return std::string();
}

/////////////////////////////////////////////////////////////////////////////
// Music
/////////////////////////////////////////////////////////////////////////////
std::string gmAudio_load_music( SDL_RWops *rwops, gmAudio *aud,
	SoundEngine *p_se)
{
	return "gmAudio_load_music::not implemented";
	//return std::string();
}

std::string gmAudio_close_music( gmAudio *aud)
{
	return "gmAudio_close_music::not implemented";
	//return std::string();
}

/////////////////////////////////////////////////////////////////////////////
// Chunk
/////////////////////////////////////////////////////////////////////////////
std::string gmAudio_load_chunk( SDL_RWops *rwops, gmAudio *aud,
	SoundEngine *p_se)
{
	const std::string pre( "gmAudio_load_chunk:: ");
	if( !(SoundEngine::audio_opened_and_ready())) {
		return ( pre +
			"SDL_OpenAudio/Mix_OpenAudio "
				"was not called (or is not active).\n"
			"Cannot load chunk.\n"
			"Please tell some SoundEngine to open the audio first,"
			" thanks.");
	}
	if( !(aud->m_chunk.m_chunk = Mix_LoadWAV_RW( rwops, 0)))
		return (pre + SDL_GetError());

	return std::string();
}

std::string gmAudio_close_chunk( gmAudio *aud)
{
	if( !(aud->m_chunk.m_chunk)) return std::string();
	aud->mix_halt( (SoundEngine *)0);
	Mix_FreeChunk( aud->m_chunk.m_chunk);
	aud->m_chunk.m_chunk = 0;
	return std::string();
}

std::string gmAudio_mix_play_chunk( gmAudio *aud, SoundEngine *se,
	int channel, int loop)
{
	if( !(aud->m_chunk.m_chunk)) return std::string();
	const int c = Mix_PlayChannel( channel, aud->m_chunk.m_chunk, loop);
	if( -1 == c) {
		return (aud->m_name + ", gmAudio_mix_play_chunk:: " +
			std::string( SDL_GetError()));
	}
	aud->m_chunk.m_play_channel.push_back( c);
	return std::string();
}

std::string gmAudio_mix_halt_chunk( gmAudio *aud, SoundEngine *se)
{
	if( !(aud->m_chunk.m_chunk)) return std::string();
	// Forcefully halt all channels using this chunk.
	while( !(aud->m_chunk.m_play_channel.empty())) {
		std::vector <int> :: iterator iter =
			aud->m_chunk.m_play_channel.begin();
		SoundEngine::halt_channel( *(iter), aud);
		aud->m_chunk.m_play_channel.erase( iter);
	}
	return std::string();
}

/////////////////////////////////////////////////////////////////////////////
// Sound Engine
/////////////////////////////////////////////////////////////////////////////
SoundEngine *SoundEngine::m_audio_opened_and_ready = 0;
std::string SoundEngine::m_audio_type_opened;
std::map <int, gmAudio *> SoundEngine::m_channel_owner;

SoundEngine::SoundEngine( GameEngine *p_ge)
{
	m_ge = p_ge;

	m_output_format = AUDIO_S16;
	m_output_rate = 22050;
	m_num_output_channels = 2;	// STEREO
	m_output_samples = 512;
	m_mix_buffer_size = 4096;
}

SoundEngine::~SoundEngine()
{
	halt_all_channels();
	remove_all_audio();
	if( m_opened_audio) {
		SDL_CloseAudio();
	}
}

void SoundEngine::halt_all_channels()
{
	// If "this" SoundEngine is the one that owns the audio, then halt
	// all channels for everybody/thing.
	if( this == m_audio_opened_and_ready) {
		// FIXME: Some "gmAudio's" may still think that they are
		//	playing on a channel because they are not notified that
		//	the channel has been halted. This is not a big problem
		//	because the SoundEngine will ignore the gmAudio's
		//	request to close an un-owned channel later. Also, the
		//	gmAudio's themselves don't really care if its channel
		//	state is consistent with the hardware state. So long
		//	as the SoundEngine can keep things in order (which
		//	it can) then there is no major problem. However, if a
		//	better solution can be found, then it should be
		//	implemented.
		Mix_HaltChannel( -1);
		m_channel_owner.clear();
		return;
	}

	// Otherwise, only halt channels owned by "this" SoundEngine's audios.
	Map_String_Audio :: const_iterator iter = m_audio.begin();
	int count = int( m_audio.size());
	while( count--) {
		(iter->second)->mix_halt( this);
		++iter;
	}
}

void SoundEngine::halt_channel( const int channel, gmAudio *owner)
{
	std::map <int, gmAudio *> :: iterator iter =
		m_channel_owner.find( channel);
	if( iter == m_channel_owner.end()) return;
	if( owner && ((iter->second) != owner)) return;
	if( Mix_Playing( channel)) Mix_HaltChannel( channel);
	m_channel_owner.erase( iter);
}

std::string  SoundEngine::mix_play_audio( const std::string &p_aud,
	int channel, int loop)
{
	gmAudio *aud = get_audio( p_aud);
	if( !aud) return "no such audio \"" + p_aud + "\"";
	return aud->mix_play( channel, loop, this);
}

std::string SoundEngine::mix_halt_audio( const std::string &p_aud)
{
	gmAudio *aud = get_audio( p_aud);
	if( !aud) return "no such audio \"" + p_aud + "\"";
	return aud->mix_halt( this);
}

std::string SoundEngine::open_audio( const std::string &which)
{
	if( m_audio_opened_and_ready) {
		if( this != m_audio_opened_and_ready) {
			return ("SoundEngine (" +
				rg_to_string( m_audio_opened_and_ready) +
				") owns the audio. This SoundEngine (" +
				rg_to_string( this) +
				") cannot open the audio nor close it. Please "
				"Send open/close commands to SoundEngine (" +
				rg_to_string( m_audio_opened_and_ready) + ")");
		}
		close_audio();
	}

	SDL_AudioSpec fmt;
	fmt.freq = m_output_rate;
	fmt.format = m_output_format;
	fmt.channels = m_num_output_channels;
	fmt.samples = m_output_samples;
	fmt.callback = SoundEngine_mixer;
	fmt.userdata = NULL;

	if( "SDL" == which) {
		if( SDL_OpenAudio( &fmt, NULL) < 0 ) {
			return ("SDL_OpenAudio failed: " +
				std::string( SDL_GetError()));
		}

	} else if( "Mix" == which) {
		if( Mix_OpenAudio( fmt.freq, fmt.format, fmt.channels,
			m_mix_buffer_size)) {
			return ("Mix_OpenAudio failed: " +
				std::string( SDL_GetError()));
		}
	}
	m_audio_opened_and_ready = this;
	return std::string();
}

std::string SoundEngine::close_audio()
{
	if( !m_audio_opened_and_ready) return std::string();
	if( this != m_audio_opened_and_ready) {
		return ("SoundEngine (" +
			rg_to_string( m_audio_opened_and_ready) +
			") owns the audio. This SoundEngine (" +
			rg_to_string( this) +
			") cannot close the audio. Please "
			"Send open/close commands to SoundEngine (" +
			rg_to_string( m_audio_opened_and_ready) + ")");
	}
	if( "SDL" == m_audio_type_opened) SDL_CloseAudio();
	else if( "Mix" == m_audio_type_opened) {
		// XXX: Must close the audio the same number of times it was
		//	opened.
		int hz;
		Uint16 fmt;
		int chan;
		int count = Mix_QuerySpec( &hz, &fmt, &chan);
		while( count--)
			Mix_CloseAudio();
	}
	else return std::string();

	m_audio_opened_and_ready = 0;
	return std::string();
}

std::string SoundEngine::load_audio(
	const std::string &key,
	const std::string &fname,
	const std::string &p_type)
{
	gmFileHandle f( m_ge->file_engine);
	return load_audio( key, fname, f, p_type);
}

// FIXME: SDL does not support reading files from a "dynamic" FILE
//	stream. That is, a FILE stream that is being created on the
//	fly. This is because files from archives are returned with
//	their data still being extracted (full complete files are not
//	provided). To fix this, audio "swap" files are created so that
//	SDL has access to full complete files.
std::string SoundEngine::load_audio(
	const std::string &key,
	const std::string &fname,
	gmFileHandle &p_fh,
	const std::string &p_type)
{
	const std::string err_prefix( "SoundEngine::load_audio( " +
		rg_to_string( this) + "): ");
	
	if( key.length() <= 0)
		return (err_prefix + "null key name passed.");

	// Load the audio file.
	if( !(p_fh.open_file( fname))) return p_fh.m_emsg;
	gmAudio *aud = new gmAudio( key);
	if( !aud) return (err_prefix +
		"could not create \"gmAudio\" for \"" + fname + "\"");

	{ // Create audio swap file (if needed).
		FILE *tmp;
		if( p_fh.is_real_file()) {
			// XXX: Audio files (such as large music/wave files)
			//	may need to have the sound file persist for the
			//	duration of the game. However, "gmFileHandle"
			//	will delete the swap file as soon as this
			//	function terminates. Therefore, the swap file
			//	is detached from "gmFileHandle" and given to
			//	gmAudio (which will persist it).
			aud->m_swap_file = p_fh.swap_file_name();
			p_fh.set_swap_file_name( std::string());
			tmp = fopen( (aud->m_swap_file).c_str(), "r+");

		} else {
			aud->m_swap_file = "rph_audio";
			std::string emsg(
				m_ge->file_engine->create_temporary_file(
					&tmp, aud->m_swap_file));
			if( emsg.length() > 0) {
				aud->m_swap_file = std::string();
				delete aud;
				return (err_prefix + emsg);
			}
			// Populate the swap file with audio data.
			emsg = rgport_file_copy( p_fh.get_data(), tmp);
			if( emsg.length() > 0) {
				delete aud;
				return (err_prefix + "while copying to swap: "
					+ emsg);
			}
			fseek( tmp, 0, SEEK_SET);
		}

		// Load the audio file into SDL.
		const std::string emsg( aud->load( tmp, p_type, this));
		if( emsg.length() > 0) {
			delete aud;
			return err_prefix + "\"" + fname + "\": " + emsg;
		}
		if( tmp) fclose( tmp);
	}

	typedef std::pair <const std::string, gmAudio *> SAPair;
	if( false == (m_audio.insert( SAPair( key, aud))).second) {
		delete aud;
		return (err_prefix + "duplicate key \"" + key + "\" for \"" +
			fname + "\"");
	}

	return std::string();
}

void SoundEngine::remove_audio( const std::string &key)
{
	Map_String_Audio :: iterator iter = m_audio.find( key);
	if( iter == m_audio.end()) return;
	if( iter->second) delete (iter->second);
	m_audio.erase( iter);
}

void SoundEngine::remove_all_audio()
{
	while( !(m_audio.empty())) {
		Map_String_Audio :: iterator iter = m_audio.begin();
		if( iter->second) delete (iter->second);
		m_audio.erase( iter);
	}
}

gmAudio *SoundEngine::get_audio( const std::string &key)
{
	Map_String_Audio :: const_iterator iter = m_audio.find( key);
	if( iter == m_audio.end()) return 0;
	return (iter->second);
}

void SoundEngine_mixer( void *unused, Uint8 *stream, int len)
{

}
