/*
Copyright (C) 2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef ENGINE_SOUND_H
#define ENGINE_SOUND_H

#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <map>
#include <SDL/SDL.h>
//#include <SDL/SDL_audio.h>
#include <SDL/SDL_mixer.h>

class GameEngine;
class FileEngine;
class SoundEngine;
class gmAudio;
class gmAudioOps;
class gmFileHandle;

typedef std::map <const std::string, gmAudio *> Map_String_Audio;
typedef std::map <const std::string, gmAudioOps *> Map_String_AudioOps;

class gmAudioOps
{
public:
	std::string (*load)( SDL_RWops *, gmAudio *, SoundEngine *);
	std::string (*close)( gmAudio *);
	std::string (*convert_to_format)( gmAudio *, SoundEngine *se);
	std::string (*mix_play)( gmAudio *, SoundEngine *, int, int);
	std::string (*mix_halt)( gmAudio *, SoundEngine *);

	gmAudioOps() {
		gmAudioOps::load = 0;
		gmAudioOps::close = 0;
		convert_to_format = 0;
		mix_play = 0;
		mix_halt = 0;
	}
};

class gmAudioSpec
{
public:
	SDL_AudioSpec m_audio;
	SDL_AudioCVT m_cvt;
	Uint32 wav_length;
	Uint8 *wav_buffer;

	gmAudioSpec() {
		wav_length = 0;
		wav_buffer = 0;
		memset( &m_audio, 0, sizeof( m_cvt));
		memset( &m_cvt, 0, sizeof( m_cvt));
	}

	~gmAudioSpec() { close();}

	void close() {
		free_raw_wav();
		if( m_cvt.buf) {
			delete [] (m_cvt.buf);
			memset( &m_cvt, 0, sizeof( m_cvt));
		}
	}

	void free_raw_wav() {
		wav_length = 0;
		if( wav_buffer) {
			SDL_FreeWAV( wav_buffer);
			wav_buffer = 0;
			memset( &m_audio, 0, sizeof( m_audio));
		}
	}
};

class gmAudioChunk
{
public:
	Mix_Chunk *m_chunk;
	std::vector <int> m_play_channel;

	gmAudioChunk() {
		m_chunk = 0;
	}
};

class gmAudio
{
public:
	std::string m_name;

	gmAudio( const std::string &p_name);
	~gmAudio();

	std::string load( FILE *f, const std::string &p_type,
		SoundEngine *p_se);
	std::string close() {
		if( !m_aops) return std::string();
		if( !(m_aops->close))
			return m_name + ", close:: not implemented/supported";
		const std::string emsg( m_aops->close( this));
		if( m_swap_file.length() > 0) remove( m_swap_file.c_str());
		return emsg;
	}

	const std::string &audio_type() const { return m_audio_type;}


	std::string convert_to_format( SoundEngine *p_se) {
		if( !p_se) return "null SoundEngine passed";
		const std::string cas( check_aop_state());
		if( cas.length() > 0) return ("convert_to_format:: " + cas);
		if( m_aops->convert_to_format)
			return m_aops->convert_to_format( this, p_se);
		return (m_name + ", "
			"convert_to_format not implemented/supported");
	}

	std::string mix_play( int channel, int loop, SoundEngine *p_se);
	std::string mix_halt( SoundEngine *p_se);

private:

	friend class SoundEngine;
	std::string m_swap_file;

	gmAudioOps *m_aops;
	friend std::string gmAudio_load_sound( SDL_RWops *, gmAudio *,
		SoundEngine *);
	friend std::string gmAudio_close_sound( gmAudio *);
	friend std::string gmAudio_convert_sound_to_format( gmAudio *,
		SoundEngine *);

	friend std::string gmAudio_load_music( SDL_RWops *, gmAudio *,
		SoundEngine *);
	friend std::string gmAudio_close_music( gmAudio *);

	friend std::string gmAudio_load_chunk( SDL_RWops *, gmAudio *,
		SoundEngine *);
	friend std::string gmAudio_close_chunk( gmAudio *);
	friend std::string gmAudio_mix_play_chunk( gmAudio *, SoundEngine *,
		int, int);
	friend std::string gmAudio_mix_halt_chunk( gmAudio *, SoundEngine *);

	std::string check_aop_state() const {
		if( !m_aops) return ("gmAudio \"" + m_name + "\" not ready");
		return std::string();
	};

	// FIXME: These should probably but put into a union since they are
	//	all mutually exclusive.
	gmAudioSpec m_sound;
	gmAudioChunk m_chunk;
	//Mix_Music m_music;

	static std::map <const std::string, gmAudioOps *> m_audio_format;

	std::string m_audio_type;

	gmAudioOps *get_audio_ops( const std::string &key);

	gmAudio( const gmAudio &);
	gmAudio operator =( const gmAudio &);
};


class SoundEngine
{
public:
	SoundEngine( GameEngine *p_ge);
	~SoundEngine();

	std::string load_audio(
		const std::string &key,
		const std::string &fname,
		const std::string &p_type);

	void remove_audio( const std::string &key);
	void remove_all_audio();

	gmAudio *get_audio( const std::string &key);

	void set_output_rate( int hz) {
		if( hz <= 0) return;
		m_output_rate = hz;}
	std::string set_output_format( const std::string &fmt) {
		if( "AUDIO_U8" == fmt) m_output_format = AUDIO_U8;
		else if( "AUDIO_S8" == fmt) m_output_format = AUDIO_S8;
		else if( "AUDIO_U16" == fmt) m_output_format = AUDIO_U16;
		else if( "AUDIO_S16" == fmt) m_output_format = AUDIO_S16;
		else if( "AUDIO_U16LSB" == fmt) m_output_format = AUDIO_U16LSB;
		else if( "AUDIO_S16LSB" == fmt) m_output_format = AUDIO_S16LSB;
		else if( "AUDIO_U16MSB" == fmt) m_output_format = AUDIO_U16MSB;
		else if( "AUDIO_S16MSB" == fmt) m_output_format = AUDIO_S16MSB;
		else if( "AUDIO_U16SYS" == fmt) m_output_format = AUDIO_U16SYS;
		else if( "AUDIO_S16SYS" == fmt) m_output_format = AUDIO_S16SYS;
		else return ("invalid audio format \"" + fmt + "\"");
		return std::string();
	}
	void set_num_output_channels( int n) {
		if( n <= 0) return;
		m_num_output_channels = n;
	}
	void set_output_samples( int n) {
		if( n <= 0) return;
		m_output_samples = n;
	}
	void set_mix_buffer_size( int n) {
		if( n <= 0) return;
		m_mix_buffer_size = n;
	}

	Uint16 output_format() const { return m_output_format;}
	int output_rate() const { return m_output_rate;}
	int num_output_channels() const { return m_num_output_channels;}
	int output_samples() const { return m_output_samples;}
	int mix_buffer_size() const { return m_mix_buffer_size;}

	std::string open_audio( const std::string &which);
	std::string close_audio();
	void pause_audio( const bool b) {
		if( b) SDL_PauseAudio( 1);
		else SDL_PauseAudio( 0);
	}

	std::string mix_play_audio( const std::string &aud, int channel,
		int loop);
	std::string mix_halt_audio( const std::string &aud);

	static void halt_channel( const int channel, gmAudio *owner);
	void halt_all_channels();

	static SoundEngine *audio_opened_and_ready() {
		return m_audio_opened_and_ready;}

protected:
	GameEngine *m_ge;

	bool m_opened_audio;

	Uint16 m_output_format;
	int m_output_rate;
	int m_num_output_channels;
	int m_output_samples;
	int m_mix_buffer_size;

	std::fstream m_audio_swap_file;

	Map_String_Audio m_audio;

private:
	static SoundEngine *m_audio_opened_and_ready;
	static std::string m_audio_type_opened;

	static std::map <int, gmAudio *> m_channel_owner;


	std::string load_audio(
		const std::string &key,
		const std::string &fname,
		gmFileHandle &p_fh,
		const std::string &p_type);

	SoundEngine( const SoundEngine &);
	SoundEngine &operator =( const SoundEngine &);
};

#endif
