/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "things.h"
#include "engine_game.h"
#include "engine_physics.h"
#include "engine_graphics.h"
#include <fstream>
#include <rg3DRom.h>
#include <rg3DRom_csg.h>
#include <rg3DRom_physics.h>

std::map <const std::string, THING_COMMAND_FN> gmPhyModel::m_ccmd;

extern "C" gmThing *spw_physical_model()
{
        return new gmPhyModel();
}

extern "C" int del_physical_model( gmThing *t)
{
	delete t;
	return 0;
}

gmPhyModel::gmPhyModel()
{
	m_use_gravity = true;
	m_skel = 0;

	m_cc = new PhyCollisionContext( this);

	m_bounciness = 0.5f;

	m_draw_dyn_bboxes = false;

	// XXX: Setup hack data.
	m_hack_ground = 0.0f;
	m_hack_ground_friction = 0.0025f;

	if( !(m_ccmd.empty())) return;
	add_console_command( "set_collision_radius",
		gmPhyModel_set_collision_radius, m_ccmd);
	add_console_command( "add_collision_list",
		gmPhyModel_add_collision_list, m_ccmd);
        add_console_command( "set_model", gmPhyModel_set_model, m_ccmd);
	add_console_command( "load_identity", gmPhyModel_load_identity,
		m_ccmd);
	add_console_command( "scale", gmPhyModel_scale, m_ccmd);
        add_console_command( "set_center_of_mass",
		gmPhyModel_set_center_of_mass, m_ccmd);
        add_console_command( "add_dynamic_bboxes",
		gmPhyModel_add_dynamic_bboxes, m_ccmd);
	add_console_command( "dump_dynamic_bboxes",
		gmPhyModel_dump_dynamic_bboxes, m_ccmd);
	add_console_command( "draw_dynamic_bboxes",
		gmPhyModel_draw_dynamic_bboxes, m_ccmd);
	add_console_command( "bounciness", gmPhyModel_bounciness, m_ccmd);
	add_console_command( "ground_friction", gmPhyModel_ground_friction,
		m_ccmd);
}

gmPhyModel::~gmPhyModel()
{
	if( m_cc) delete m_cc;
	if( m_skel) delete m_skel;
}

void gmPhyModel::process( ThingManager *tm)
{
	// Get models physical state before transformations.
	// Also, save current model skeleton since applying transformations may
	// put the skeleton in an invalid (colliding) physical state. The
	// skeleton "m_draw_context->m_skout" is saved along with supporting
	// matrices.

	PhysicsEngine *pe = phy_engine();
	get_origin_points( *m_cc);
//	m_draw_context->backup_main_skel();


	// Apply transformations to the model.
	animate();
	if( !m_bind_master) {
		apply_gravity();
		apply_velocity();

	} else m_bind_master->update_binding();

	// Get a physics context (so physical stuff can be applied to the
	// model). Also, get the models physical state after the applied
	// transformations.
	if( model()) (model())->skel.update_fcs( *m_draw_context,
		m_draw_context->m_skout);
	// XXX: Since the "fcs" of the entire skeleton is calculated here,
	//	prevent it from being redundantly calculated again later on.
	m_draw_context->m_lock_fcs = true;

	get_bounding_points( *m_cc);



	// Check if the transformed skeleton is in an invalid phyiscal state.
	// If it is, then perform "anti-intrusion" functions on it if possible.
	// The "thing" gets "attempts" number of tries to produce a physically
	// valid transformation. If all attempts fail, then the original
	// skeleton is restored (since it IS in a valid physical state).

	{ // Basic "thing/thing" collision.
		std::vector <PhyCollisionFragment> f;
		{
			for( int i = 0; i < int( m_collision_list.size()); i++)
				pe->thing_collision( f, this,
					m_collision_list[i]);
		}

		if( !(f.empty())) {
			rgVector v;
			for( int i = 0; i < int( f.size()); i++)
				v -= f[i].intrusion_vector;
			m_draw_context->translate(
				v * (1.0f / float( f.size())));
		}
	}


	//////////////////////////////////////////////////////////////////////
	// XXX: Apply a basic ground for all things to collide.
	//////////////////////////////////////////////////////////////////////
	if( hack_ground_anti_intrusion( m_cc) > 0.0f) {
		// FIXME: Do proper friction
		rgVector &v = m_velocity;
		const float len = RG_MIN( m_hack_ground_friction, v.length());
		if( len < 0.00001f) v = rgVector();
		else {
			v -= (rgMakePlanar( v, rgVector( 0, 1, 0)).normalize())
				* (len / float( (ge())->game_clock_hz()));
		}
		v.y *= -m_bounciness;
	}



	//////////////////////////////////////////////////////////////////////
	// XXX: Colliding with additional environmental geometery.
	//////////////////////////////////////////////////////////////////////
	PhyDynResult result;
	if( pe->environment_collision( result, *m_cc)) {
		rgVector avg_surface_normal;
		rgVector anti_intrude;
		for( int i = 0; i < int( result.m_point_hit.size()); i++) {
			PhyCollisionFragment &f = result.m_point_hit[i];
//			std::cout << i << " = " << f << "\n";
			avg_surface_normal += f.surface_normal;
			anti_intrude += anti_intrusion( f);
			const float k = m_velocity.dot( f.surface_normal);
			if( k < 0) {
				const rgVector &n = f.surface_normal;
				rgVector &v = m_velocity;
				v -= (k * (1.0f + m_bounciness)) * n;

				// FIXME: Do proper friction
				const float len = RG_MIN(
					f.prisim->m_friction,
					v.length());
				if( len < 0.00001f) v = rgVector();
				else {
					v -= ((v - (v.dot( n) * n)).
						normalize()) *
						(len /
					float( (ge())->game_clock_hz()));
				}
			}
		}
		const float k = 1.0f / float( result.m_point_hit.size());
		anti_intrude = anti_intrude * k;
		avg_surface_normal.normalize();
		m_draw_context->translate( anti_intrude);
	}
}

rgVector gmPhyModel::anti_intrusion( PhyCollisionFragment &f)
{
	const rgVector aip( f.collision_point + (f.surface_normal *
		(f.radius + 0.001f)));
	return (f.surface_normal * f.surface_normal.dot( aip - f.des_point));
}

float gmPhyModel::hack_ground_anti_intrusion( PhyCollisionContext *p_cc) const
{
	if( !p_cc) return false;
	float h = 0.0f;
	for( int i = 0; i < int( p_cc->m_bpoint.size()); i++) {
		const float y = (p_cc->m_bpoint[i]).y -
			(p_cc->m_bpoint[i]).radius;
		if( y < m_hack_ground)
			h = RG_MAX( 0.0f, m_hack_ground - y);
	}
	m_draw_context->translate( rgVector( 0, h, 0));
	return h;
}

rgVector gmPhyModel::post_collision_vector( const rgVector v, const rgVector n)
	const
{
	const rgVector r( rgReflectAbout( -v, n));
	const rgVector p( (v + r) * 0.5f);
	return ((r * m_bounciness) + (p * (1.0f - m_bounciness)));
/*
	const rgVector q( ((r * m_bounciness) + (p * (1.0f - m_bounciness))));
	std::cout << "----------------------------\n";
	std::cout << "v = " << v << "\n";
	std::cout << "n = " << n << "\n";
	std::cout << "r = " << r << "\n";
	std::cout << "p = " << p << "\n";
	std::cout << "q = " << q << "\n";
	exit(0);
	return q;
	*/
}

void gmPhyModel::draw( gmRenderArgs *ra)
{
	gmModelThing::draw( ra);
	if( m_draw_dyn_bboxes) draw_dynamic_bboxes();
}

void gmPhyModel::draw_dynamic_bboxes()
{
	sk3DRomModel *m = model();
	if( !m) return;
	if( !(m->m_csg)) return;

	// No blend
	GLboolean blend;
	glGetBooleanv( GL_BLEND, &blend);
	glDisable( GL_BLEND);

	// No texturing
	GLboolean texture;
	glGetBooleanv( GL_TEXTURE_2D, &texture);
	glDisable( GL_TEXTURE_2D);

	glBegin( GL_QUADS);
	glColor3f( 0, 0, 1);
	for( int i = 0; i < int( m_dyn_bbox.size()); i++) {
		std::vector <rgVector> v;
		get_bbox( v, m_dyn_bbox[i]);
		rgCollisionPrisim tmp;
		tmp.m_v[0] = v[0];
		tmp.m_v[1] = v[1];
		tmp.m_v[2] = v[2];
		tmp.m_v[3] = v[3];
		tmp.m_v[4] = v[4];
		tmp.m_v[5] = v[5];
		tmp.m_v[6] = v[6];
		tmp.m_v[7] = v[7];
		tmp.render();
	}
	glEnd();

	if( blend) glEnable( GL_BLEND);
	if( texture) glEnable( GL_TEXTURE_2D);
}

gmTE gmPhyModel::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.find( key);
	if( iter == m_ccmd.end())
		return gmModelThing::call_ccmd( key, args);
	return (iter->second)( args);
}

std::string gmPhyModel::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); ++iter) {
		tmp += p_prefix + gmPhyModel::type() + "::" + (iter->first) +
			"\n";
	}
	tmp += gmModelThing::list_ccmd( p_prefix);
	return tmp;
}

gmTE gmPhyModel_set_model( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) return GM_THING_ERR_PARAM;

	gmPhyModel *pm = (gmPhyModel *)(p.thing);
	if( (gmModelThing_set_model( p)).is_error())
		return GM_THING_ERR_GENERIC;
	if( !(pm->model())) return GM_THING_ERR_GENERIC;
	if( pm->m_skel) { delete pm->m_skel; pm->m_skel = 0;}
	pm->m_skel = new sk3DRomPhySkel( (pm->model())->skel,
		*(pm->m_draw_context));
	return 0;
}

// gmPhyModel_load_identity is required to override gmModelThing_load_identity
// because loading the identity matrix may affect the sizes of the bounding
// boxes.
gmTE gmPhyModel_load_identity( const gmCCArg &p)
{
	if( (gmModelThing_load_identity( p)).is_error())
		return GM_THING_ERR_GENERIC;
	gmPhyModel *pm = (gmPhyModel *)(p.thing);
	//(pm->phy_engine())->gen_bounding_points( *(pm->m_cc));
	pm->gen_bounding_points( *(pm->m_cc));
	return 0;
}

// gmPhyModel_scale is required to override gmModelThing_scale because scaling
// the model affects the sizes of the bounding boxes.
gmTE gmPhyModel_scale( const gmCCArg &p)
{
	if( (gmModelThing_scale( p)).is_error())
		return GM_THING_ERR_GENERIC;
	gmPhyModel *pm = (gmPhyModel *)(p.thing);
	//(pm->phy_engine())->gen_bounding_points( *(pm->m_cc));
	pm->gen_bounding_points( *(pm->m_cc));
	return 0;
}

gmTE gmPhyModel_bounciness( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	gmPhyModel *pm = (gmPhyModel *)(p.thing);
	if( 0 == args.size()) {
		p.ge->rcout( "Bounciness = " + gm_to_string( pm->m_bounciness)
			+ "\n");
	} else pm->m_bounciness = strtod( (args[0]).c_str(), 0);
	return 0;
}

gmTE gmPhyModel_ground_friction( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	gmPhyModel *pm = (gmPhyModel *)(p.thing);
	if( 0 == args.size()) {
		p.ge->rcout( "ground_friction = " +
			gm_to_string( pm->m_hack_ground_friction) + "\n");
	} else pm->m_hack_ground_friction = strtod( (args[0]).c_str(), 0);
	return 0;
}

gmTE gmPhyModel_set_center_of_mass( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) return GM_THING_ERR_PARAM;

	gmPhyModel *pm = (gmPhyModel *)(p.thing);
	if( !pm->m_mh) return GM_THING_ERR_GENERIC;
	const int bid = (pm->m_com.m_bone_id = pm->m_mh->bone_id( args[0]));
	if( bid < 0) {
		p.ge->rcout( "No bone named \"" + args[0] + "\"\n");
		return GM_THING_ERR_GENERIC;
	}
	if( !(pm->model())) return GM_THING_ERR_GENERIC;
	pm->m_com.m_bound_position =
		((pm->model())->skel.b[bid]).get_jcs_pivot();
	return 0;
}

gmTE gmPhyModel_add_dynamic_bboxes( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) {
		p.ge->rcout( "usage: add_dynmaic_bboxes <group key>\n");
		return GM_THING_ERR_PARAM;
	}

	gmPhyModel *pm = (gmPhyModel *)(p.thing);
	if( !(pm->add_dynamic_bboxes( args[0]))) {
		p.ge->rcout( "gmPhyModel_add_dynamic_bboxes:\n"
			"Could not add bboxes with group = \"" + args[0] +
			"\"\n");
		return GM_THING_ERR_GENERIC;
	}
	return 0;
}

gmTE gmPhyModel_set_collision_radius( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) {
		return "usage: set_collision_radius <radius>\n"
			"NOTE: \"radius\" is in meters.\n";
	}
	(((gmPhyModel *)(p.thing))->center_of_mass()).m_collision_radius =
		strtod( (args[0]).c_str(), 0);
	return 0;
}

gmTE gmPhyModel_add_collision_list( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() <= 0) {
		return "usage: add_collision_list <list 1> [list 2] "
			"... [list n]\n";
	}

	gmPhyModel *t = (gmPhyModel *)(p.thing);
	for( int i = 0; i < int( args.size()); i++)
		t->m_collision_list.push_back( args[i]);
	return 0;
}

gmTE gmPhyModel_dump_dynamic_bboxes( const gmCCArg &p)
{
	p.ge->rcout( ((gmPhyModel *)(p.thing))->dump_dyn_bboxes());
	return 0;
}

gmTE gmPhyModel_draw_dynamic_bboxes( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1)
		return "usage: draw_dynmaic_bboxes <0/1/on/off/toggle>\n";
	bool &b = ((gmPhyModel *)(p.thing))->m_draw_dyn_bboxes;
	b = gm_set_state( b, args[0]);
	return 0;
}

// Description: Returns a vector of absolute vertex locations representing the
//	bounding box. The model must be transformed at least once by the
//	"thing" inorder to get defined results for the bounding box.
//
// FIXME: The below comment, which deals with the "fcs" (Full Coordinate
//	System) issue has been solved. sk3DRomModel now accepts draw contexts
//	(sk3DRomDrawContext) which allow a "thing" to keep its own copy of the
//	bone "fcs"'s. This is done transparently when ever a "thing" provides
//	its draw context to the model when transforming it. So get_bbox now
//	just uses the "fcs" stored in the draw context rather than the model.
//
//	The vertices cannot be calculated until the 
//	"sk3DRomMode::skel.b[n].fcs" has been calculated for all bones
//	containing bounding boxes. "fcs" stands for (Full Coordinate System).
//	This means the model has to transformed before bounding boxes can
//	be retrieved. This is not convinient (since it would be better to be
//	able to access bounding boxes at random). Also, this forces "things"
//	to retransform their models at every frame (since "fcs" will be
//	corrupted by other "things" using the same model).
//
//	This problem is solved by forcing the "thing" to save the "fcs" of all
//	required bones and then restoring the "fcs" for those bones before
//	it uses the skeleton again.
int gmPhyModel::get_bbox( std::vector <rgVector> &v, int id)
{
	v.clear();
	sk3DRomModel *m = model();
	if( !m) return -1;
	if( !(m->m_csg)) return -2;
	const rgCSGBrush *b = m->m_csg->brush( id);
	if( !b) return -3;
	// FIXME: "b" should ALWAYS be pointing to a "RG3DROM_CSG_CUBE", so
	//	the "if" check shouldn't be neccasary.
	if( RG3DROM_CSG_CUBE == b->type()) {
		b->map_verts( v, m_draw_context->get_fcs(
			((sk3DRomCSGCube *)b)->m_bone_id).map_csys(
				b->raw_coord_system()));

	} else b->gen_verts( v);
	//const rgMatrix &mat = m_draw_context->mat();
	const rgMatrix mat( m_draw_context->mat());
	for( int i = 0; i < int( v.size()); i++)
		v[i] = mat.pointMul( v[i]);
	return 0;
}

PhyBoundingPoint gmPhyModel::get_bbox_point( const int id)
{
	return get_bbox_point( id, m_draw_context->mat());
}

PhyBoundingPoint gmPhyModel::get_bbox_point( const int id, const rgMatrix &mat)
{
	if( !(model())) return PhyBoundingPoint();
	const rgCSGBrush *b = (model())->m_csg->brush( id);
	// FIXME: "b" should ALWAYS be pointing to a "RG3DROM_CSG_CUBE", so the
	//	"if" check shouldn't be neccasary.
	if( RG3DROM_CSG_CUBE == b->type()) {
		return PhyBoundingPoint(
				mat.pointMul(
				((m_draw_context->get_fcs(
				((sk3DRomCSGCube *)b)->m_bone_id)).map_csys(
					b->raw_coord_system())).p),

				// XXX: Bounding prisim's width is not affected
				//	by bone transformations.
				// FIXME: Calculating the length every time a
				//	point is accessed is not efficient.
				//	Use precalculation or use a more
				//	efficient method.
				(mat * (b->raw_coord_system()).side).length()
			);
	}
	return PhyBoundingPoint();
}

std::string gmPhyModel::dump_dyn_bboxes()
{
	std::string tmp( "Dynamic Bounding Boxes (Cube Brushes): " +
		gm_to_string( int( m_dyn_bbox.size())) + "\n");
	sk3DRomModel *m = model();
	if( !m) return tmp;
	if( !(m->m_csg)) return tmp;
	for( int i = 0; i < int( m_dyn_bbox.size()); i++) {
		std::vector <rgVector> v;
		if( get_bbox( v, m_dyn_bbox[i])) {
			tmp += "\tCould not get brush with id = " +
				gm_to_string( m_dyn_bbox[i]) + "\n";

		} else {
			if( v.empty()) {
				tmp += "\tVertex-less brush, id = " +
					gm_to_string( m_dyn_bbox[i]) + "\n";
			} else {
				tmp += "\tBrush id = " + gm_to_string(
						m_dyn_bbox[i]) + ", " +
					gm_to_string( int( v.size())) +
					" vertices\n";
				while( !(v.empty())) {
					tmp += "\t\t" + gm_to_string( v[0]) +
						"\n";
					v.erase( v.begin());
				}
			}
		}
	}
	return tmp;
}

gmPhyModelCOM &gmPhyModel::center_of_mass()
{
	if( !m_skel) return m_com;
	if( !m_draw_context) return m_com;
	return center_of_mass( m_draw_context->mat());
}

gmPhyModelCOM &gmPhyModel::center_of_mass( const rgMatrix &mat)
{
	const sk3DRomSkel *s = m_skel->skel();
	m_com.m_position = mat.pointMul(
		(s->get_full_bone_rcs( m_com.m_bone_id,
		m_draw_context->m_skout)).map_point( m_com.m_bound_position));
	return m_com;
}

rgCoordSystem gmPhyModel::cs()
{
	if( !m_skel) return gmModelThing::cs();
	if( !m_draw_context) return gmModelThing::cs();
	const rgMatrix mat( m_draw_context->mat());
	m_cs.p = (center_of_mass( mat)).m_position;
	m_cs.side = (rgVector( mat.d[0], mat.d[4], mat.d[8])).normalize();
	m_cs.up = (rgVector( mat.d[1], mat.d[5], mat.d[9])).normalize();
	m_cs.eye = (rgVector( mat.d[2], mat.d[6], mat.d[10])).normalize();
	return m_cs;
}


// Description: Apply gravity to the bone that contains the center-of-mass
//	(COM). That bone, will then take all other bones with it (thus
//	creating the effect that gravity is applied to all bones).
void gmPhyModel::apply_gravity()
{
	if( !m_use_gravity) return;
	m_velocity += (phy_engine())->m_gravity;
}

void gmPhyModel::apply_velocity()
{
	//if( !m_skel) return;
	//const sk3DRomSkel *s = m_skel->skel();
	//s->translate_rcs( m_com.m_bone_id, m_draw_context->m_skout,
	//	m_velocity);
	m_draw_context->translate( m_velocity);
//	m_draw_context->update_matrix();
}

bool gmPhyModel::add_dynamic_bboxes( const std::string &key)
{
	if( !m_mh) return false;
	const int gid = m_mh->group_id( key);
	if( gid < 0) return false;
	sk3DRomModel *m = model();
	if( !m) return false;
	if( !(draw_context())) return false;
	for( int i = 0; i < m->m_csg->brush_count(); i++) {
		rgCSGBrush *b = m->m_csg->brush( i);
		if(	(gid == b->group_id()) &&
			(RG3DROM_CSG_CUBE == b->type())) {
			m_dyn_bbox.push_back( i);
		}
	}
	return true;
}
                                                                                
void gmPhyModel::clear_dynamic_bboxes()
{
	m_dyn_bbox.clear();
}

void gmPhyModel::get_origin_points( PhyCollisionContext &cc) const
{
	if( !(cc.m_model)) return;
	const rgMatrix mat( m_draw_context->mat());
	if( cc.m_bpoint.size() < cc.m_model->m_dyn_bbox.size())
		gen_bounding_points( cc, mat);
	for( int i = 0; i < int( cc.m_model->m_dyn_bbox.size()); i++) {
		(cc.m_bpoint[i]).m_origin = cc.m_model->get_bbox_center(
			cc.m_model->m_dyn_bbox[i], mat);
	}
}

void gmPhyModel::get_bounding_points( PhyCollisionContext &cc) const
{
	if( !(cc.m_model)) return;
	const rgMatrix mat( m_draw_context->mat());
	if( cc.m_bpoint.size() < cc.m_model->m_dyn_bbox.size())
		gen_bounding_points( cc, mat);
	for( int i = 0; i < int( cc.m_model->m_dyn_bbox.size()); i++) {
		cc.m_bpoint[i] = cc.m_model->get_bbox_center(
			cc.m_model->m_dyn_bbox[i], mat);
	}
}
                                                                                
void gmPhyModel::gen_bounding_points( PhyCollisionContext &cc) const
{
	gen_bounding_points( cc, m_draw_context->mat());
}

void gmPhyModel::gen_bounding_points( PhyCollisionContext &cc,
	const rgMatrix &mat) const
{
	cc.m_bpoint.clear();
	for( int i = 0; i < int( cc.m_model->m_dyn_bbox.size()); i++) {
		cc.m_bpoint.push_back( cc.m_model->get_bbox_point(
			cc.m_model->m_dyn_bbox[i], mat));
	}
}

bool gmPhyModel::collision( PhyCollisionFragment &f, gmPhyModel *t)
{
	if( !t) return false;
	gmPhyModelCOM &com_a = center_of_mass();
	gmPhyModelCOM &com_b = t->center_of_mass();

	rgVector w( com_a.m_position - com_b.m_position);
	float d = com_a.m_collision_radius + com_b.m_collision_radius;
	const float k = w.length_squared();
	if( k < (d * d)) {

		const float len = sqrt( k);
		f.pid = -1;
		f.prisim = 0;
		f.surface_normal = w * (1.0f / len);
		f.intrusion_vector = f.surface_normal * (len - d);
		f.surface_point = com_b.m_position - f.intrusion_vector;
		return true;
	}
	return false;
}
