/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "things.h"
#include "engine_game.h"
#include "engine_graphics.h"
#include <fstream>

std::map <const std::string, THING_COMMAND_FN> gmThing::m_ccmd;

gmTE gmThing_set_pos( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 3) return GM_THING_ERR_PARAM;
	p.thing->m_cs.p = gm_str_to_vector( args[0], args[1], args[2]);
	return 0;
}

gmTE gmThing_set_velocity( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	const std::string usage( "usage: set_velocity <mode> <x> <y> <z>\n"
			"NOTE: \"mode\" can be one of \"add\", \"sub\", "
				"\"mul\" or \"set\"\n"
			"Each mode refers to how the given velocity is "
				"combined with the\n"
			"thing's current velocity\n"
			"\"x\", \"y\" and \"z\" are floating point "
				"numbers defining vector (x, y, z).\n"
			"Given velocity is in meters per second.\n"
			"(It will be converted internally into meters per "
				"frame.\n");
	if( args.size() != 4) {
		p.ge->rcout( usage);
		return GM_THING_ERR_PARAM;
	}
	const rgVector v( gm_str_to_vector( args[1], args[2], args[3])
		* (1.0f / double( p.ge->game_clock_hz())));
	if( "add" == args[0]) p.thing->m_velocity += v;
	else if( "sub" == args[0]) p.thing->m_velocity -= v;
	else if( "mul" == args[0]) {
		p.thing->m_velocity.x *= v.x;
		p.thing->m_velocity.y *= v.y;
		p.thing->m_velocity.z *= v.z;

	} else if( "set" == args[0]) p.thing->m_velocity = v;
	else {
		p.ge->rcout( "No such mode \"" + args[0] + "\"\n" + usage);
		return GM_THING_ERR_PARAM;
	}
	return 0;
}

gmTE gmThing_set_gravity( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) return GM_THING_ERR_PARAM;
	p.thing->m_use_gravity = gm_set_state( args[0]);
	return 0;
}

gmTE gmThing_move_to_proc( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( args.size() != 1) return GM_THING_ERR_PARAM;
	p.thing->move_to_proc_list( atoi( (args[0]).c_str()));
	p.ge->rcout( "Thing \"" + (p.thing->tcb())->m_key + "\" moved to "
		"process list " + gm_to_string( atoi( (args[0]).c_str())) +
		" ( " + gm_to_string( (p.thing->tcb())->m_proc) + ")\n");
	return 0;
}

gmTE gmThing_control_camera( const gmCCArg &p)
{
	const std::vector <std::string> &args = *(p.args);
	if( 0 == args.size()) {
		p.thing->uncontrol_camera();
		return 0;
	}

	if( args.size() != 1) {
		p.ge->rcout( "usage: control_camera [key]\n"
			"NOTE: If no camera (key) is given, then any currently"
				" controlled camera\nwill be uncontrolled.\n");
		return GM_THING_ERR_PARAM;
	}
	gmCamera *c = p.ge->gfx_engine->get_camera( args[0]);
	if( !c) {
		p.ge->rcout( "No such camera \"" + args[0] + "\"\n");
		return GM_THING_ERR_GENERIC;
	}
	p.thing->control_camera( c);
	return 0;

}

void gmThing::move_to_proc_list( const int pid)
{
	((tcb())->manager())->move_thing_to_proc_list( tcb(), pid);
}

//int gmThing_set_csys( const gmCCArg &p)
//{
//	return 0;
//}

gmThing::gmThing()
{
	m_use_gravity = false;
	m_tcb = 0;
	m_controlled_camera = 0;

	m_ai = 0;

	// FIXME: Keep instance count of the class so that when all classes
	//      of this type are destroyed, the console command list is
	//      removed (though not doing so should waste few memory
	//      resources).
	// Prepare list of console commands for this class (only done once,
	// future instances will use what a previous instance created).
	if( !(m_ccmd.empty())) return;
	add_console_command( "set_pos", gmThing_set_pos, m_ccmd);
	add_console_command( "set_gravity", gmThing_set_gravity, m_ccmd);
	add_console_command( "move_to_proc", gmThing_move_to_proc, m_ccmd);
	add_console_command( "set_velocity", gmThing_set_velocity, m_ccmd);
	//add_console_command( "set_csys", gmThing_set_csys, m_ccmd);
	add_console_command( "control_camera", gmThing_control_camera, m_ccmd);
}

gmThing::~gmThing()
{
	uncontrol_camera();
}

PhysicsEngine *gmThing::phy_engine()
{
	return (((tcb())->manager())->ge())->phy_engine;
}

GameEngine *gmThing::ge()
{
	return ((tcb())->manager())->ge();
}

bool gmThing::add_console_command( const std::string &key,
	gmThingError (*fn)( const gmCCArg &),
	std::map <const std::string, THING_COMMAND_FN> &p_ccmd)
{
	typedef std::pair <const std::string, THING_COMMAND_FN> SAPair;
	return ((p_ccmd.insert( SAPair( key, fn))).second);
}

gmTE gmThing::call_ccmd( const std::string &key, gmCCArg &args)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.find( key);
	if( iter == m_ccmd.end())
		return (tcb())->m_key + ": no such command \"" + key + "\"";
	return (iter->second)( args);
}

std::string gmThing::list_ccmd( const std::string &p_prefix)
{
	std::map <const std::string, THING_COMMAND_FN> :: const_iterator iter =
		m_ccmd.begin();
	std::string tmp;
	for( ; iter != m_ccmd.end(); ++iter) {
		tmp += p_prefix + gmThing::type() + "::" + (iter->first) +
			"\n";
	}
	return tmp;
}

std::string gmThing::get_key()
{
	return (tcb())->m_key;
}

void gmThing::control_camera( gmCamera *p_cam)
{
	if( p_cam) {
		p_cam->set_owner( tcb());
		m_controlled_camera = p_cam;
	}
}

void gmThing::uncontrol_camera()
{
	// FIXME: "gmCamera::set_owner calls gmThing::uncontrol_camera. Also,
	//	gmThing::uncontrol_camera calls gmCamera::set_owner which
	//	results in an infinite recursion. This is resolved by the
	//	below code. Try to find a better way to do this.
	if( !m_controlled_camera) return;
	gmCamera *tmp = m_controlled_camera;
	m_controlled_camera = 0;
	tmp->set_owner( 0);
}
