/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifndef THINGS_H
#define THINGS_H

#include <rgStandard.h>
#include <rgOpenGL.h>
#include <vector>
#include <list>
#include <string>
#include <map>
#include "engine_common_types.h"

#define GM_MAIN \
	gm_main( const std::string *p_key, GameEngine *p_ge, \
		std::string *p_emsg)

#ifdef RPH_STATIC_BUILD
	#define RPH_THING_MODULE_MAIN_FUNCTION( t) \
		int t :: GM_MAIN

	#define RPH_THING_MODULE \
		static int gm_main( const std::string *p_key, \
			GameEngine *p_ge, std::string *p_emsg);
#else
	#define RPH_THING_MODULE_MAIN_FUNCTION( t) extern "C" int GM_MAIN
	#define RPH_THING_MODULE
#endif

class GameEngine;
class sk3DRomModel;
class sk3DRomDrawContext;
class sk3DRomPhySkel;
class ThingControlBlock;
class ThingManager;
class ModelHandle;
class ModelManager;
class gmThing;
class PhysicsEngine;
class PhyBoundingPoint;
class PhyCollisionContext;
class PhyCollisionFragment;
class gmCamera;
class gmModelThing;

class gmThingError;
typedef gmThingError gmTE;

typedef std::map <const std::string, ThingControlBlock *> Map_String_TCB;

class gmThingError
{
public:
	std::string m_msg;

	gmThingError(){}
	gmThingError( int i);
	gmThingError( const char *s) { m_msg = s;}
	gmThingError( const std::string &s) { m_msg = s;}

	friend std::string operator +(
		const std::string &s,
		const gmThingError &e) {
		return (s + e.m_msg);
	}

	std::string operator +( const std::string &s) const {
		return (m_msg + s);}

	bool is_error() const {
		if( m_msg.length() > 0) return true;
		return false;
	}
};

// gmRenderArgs
// Description: Defines how a "thing" should render itself. This class must
//	be modified to suit the desired application.
class gmRenderArgs
{
public:
	int lod_level;
	bool render_shadow;
	bool render_transparent;
	bool render_reflections;
	bool render_opaque;
	bool render_decals;
	bool lock_vertices;
	bool transform;

	rgVector shadow_color;
	float shadow_alpha;

	gmRenderArgs() { init();}

	gmRenderArgs( int lod, bool p_shadow) {
		init();
		lod_level = lod;
		render_shadow = p_shadow;
	}

private:
	void init() {
		lod_level = 0;
		render_shadow = false;
		render_opaque = true;
		render_transparent = false;
		render_reflections = false;
		render_decals = false;
		shadow_alpha = 0.5f;
		lock_vertices = false;
		transform = true;
	}
};


// gmCCArg
// Description: Console Command Arguments.
class gmCCArg
{
public:
	gmThing *thing;
	GameEngine *ge;
	const std::vector <std::string> *args;

	gmCCArg( gmThing *t, GameEngine *g,
		const std::vector <std::string> *a) {
		thing = t;
		ge = g;
		args = a;
	}

private:
	gmCCArg( const gmCCArg &);
	gmCCArg &operator =( const gmCCArg &);
};


#define THING_COMMAND_FN gmThingError (*)( const gmCCArg &)

// gmThing
// Description: The most basic "thing" that can exist in the game world. All
//	things derive from "gmThing".
#define GM_THING_ERR_GENERIC	-1
#define GM_THING_ERR_NO_CCMD	-2
#define GM_THING_ERR_PARAM	-3
class gmThing
{
public:
	bool m_use_gravity;
	rgVector m_velocity;


	gmThing();
	virtual ~gmThing();
	virtual rgVector center_point() { return m_cs.p;}
	virtual rgCoordSystem cs() { return m_cs;}
	virtual void process( ThingManager *tm){}
	virtual void draw( gmRenderArgs *){}

	ThingControlBlock *tcb() { return m_tcb;}

	virtual void apply_velocity(){}
	virtual void apply_gravity(){}


	rgVector direction_to( rgVector p) { return (p - (cs()).p);}
	rgVector direction_to( gmThing *t) {
		if( !t) return rgVector();
		return direction_to( (t->cs()).p);
	}

	rgVector direction_to( rgVector p, const rgVector n) {
		return (rgMakePlanar( p, n) - rgMakePlanar( (cs()).p, n));}
	rgVector direction_to( gmThing *t, const rgVector n) {
		if( !t) return rgVector();
		return direction_to( (t->cs()).p, n);
	}

	bool add_console_command( const std::string &key, THING_COMMAND_FN,
		std::map <const std::string, THING_COMMAND_FN> &p_ccmd);

	virtual std::string type() const { return std::string( "gmThing");}
	virtual bool is_a( const std::string &s) const {
		if( s != gmThing::type()) return false;
		return true;
	}

	virtual gmThingError call_ccmd( const std::string &key, gmCCArg &args);
	virtual std::string list_ccmd( const std::string &p_prefix);

	void move_to_proc_list( const int pid);

	PhysicsEngine *phy_engine();
	GameEngine *ge();

	std::string get_key();

	void control_camera( gmCamera *p_cam);
	void uncontrol_camera();
	gmCamera *controlled_camera() { return m_controlled_camera;}

	ThingControlBlock *get_ai() { return m_ai;}
	void set_ai( ThingControlBlock *t) { m_ai = t;}

protected:
	rgCoordSystem m_cs;
	rgVector m_torque;



	// Map of supported console commands for the thing.
	static std::map <const std::string, THING_COMMAND_FN> m_ccmd;
	friend gmTE gmThing_set_pos( const gmCCArg &p);
	friend gmTE gmThing_set_gravity( const gmCCArg &p);
	friend gmTE gmThing_set_csys( const gmCCArg &p);
	friend gmTE gmThing_set_velocity( const gmCCArg &p);
	friend gmTE gmThing_control_camera( const gmCCArg &p);

private:
	gmCamera *m_controlled_camera;

	// AI that is controlling this "thing".
	ThingControlBlock *m_ai;

	friend class ThingManager;
	ThingControlBlock *m_tcb;

	gmThing( const gmThing &);
	gmThing &operator =( const gmThing &);
};



class gmModelBinder
{
public:
	ThingControlBlock *m_master;
	ThingControlBlock *m_slave;

	std::string m_bind_type;
	rgMatrix m_mat;

	rgMatrix m_pos_mat;

	int m_bid;
	rgCoordSystem m_cs;	// Master coordinate system

	gmModelBinder(
		const std::string &p_bone,
		const rgVector &p_offset,
		ThingControlBlock *p_master,
		ThingControlBlock *p_slave);
	gmModelBinder(
		const int p_bid,
		const rgVector &p_offset,
		ThingControlBlock *p_master,
		ThingControlBlock *p_slave);

	~gmModelBinder();

	bool has_binding() const { if( !m_slave) return false; return true;}
	void update_binding();

private:
	bool can_bind();

	void bind_slave(
		const int p_bid,
		const rgVector &p_offset,
		gmModelThing *t_master,
		gmModelThing *t_slave);

	void update_positioning_matrix( gmModelThing *p_master);

	rgCoordSystem get_mcs( gmModelThing *t) const;

	gmModelBinder( const gmModelBinder &);
	gmModelBinder &operator =( const gmModelBinder &);
};


// gmModelThing
// Description: Provides a "thing" which can also have a 3D model. All "things"
//	which need a basic 3D model or features provided by having a 3D model
//	may inherit from this class.
class gmModelThing : public gmThing
{
public:
	gmModelThing();
	virtual ~gmModelThing();

	rgVector center_point();
	rgCoordSystem cs();
	void process( ThingManager *tm);

	bool reference_model( const std::string &key, ModelManager &m);
	void unreference_model();
	sk3DRomModel *model();
	ModelHandle *model_handle() { return m_mh;}

	void draw( gmRenderArgs *);

	std::string type() const { return std::string( "gmModelThing");}
	bool is_a( const std::string &s) const {
		if( s != gmModelThing::type()) return gmThing::is_a( s);
		return true;
	}

	gmThingError call_ccmd( const std::string &key, gmCCArg &args);
	std::string list_ccmd( const std::string &p_prefix);

	void transform();
	void transform( const std::string &key) { transform( frame_id( key));}
	void transform( int fid);

	int frame_id( const std::string &key) const;

	sk3DRomDrawContext *draw_context() { return m_draw_context;}

	void animate();
	void animate( const int p_elapsed);
	int animate( const std::string &key, int fid, const int p_time,
		const int p_elapsed) {
		return animate( key, fid, p_time, p_elapsed, 1.0f);
	}
	int animate( const std::string &key, int fid, const int p_time,
		const int p_elapsed, const float pace);

	void enable_animation( const bool b);

	GLuint create_display_list( const std::string &key);
	GLuint get_display_list( const std::string &key) const;
	int delete_display_list( const std::string &key);
	int clear_display_lists();
	int gen_display_list( const std::string &key);
	int draw_display_list( const std::string &key) const;

	void render_face_group( const std::string &grp_name,
		const bool use_normals, const bool use_textures); 

	bool seq_stopped();
	bool cur_seq_is( const std::string &s);

	std::vector <int> *add_bbox_list( const std::string &key);
	bool remove_bbox_list( const std::string &key);
	std::vector <int> *get_bbox_list( const std::string &key);

	bool add_bboxes( const std::string &gkey,
		const std::string &list);

	void bind_model_to_box( const std::string &p_box, gmThing *t);
	void bind_model_thing( const std::string &p_bone, gmThing *t);
	void bind_model_thing( const std::string &p_bone,
		const rgVector &p_offset, gmThing *t);
	void bind_model_thing(
		const std::string &p_bone,
		const int p_bid,
		const rgVector &p_offset,
		ThingControlBlock *p_tcb);
	void release_binding( gmThing *t);
	void release_binding( ThingControlBlock *p_tcb);
	void release_all_bindings();

	gmModelBinder *get_binding( const std::string &key);
	gmModelBinder *get_binding( ThingControlBlock *p_tcb);


	bool add_action_anim( const std::string &act, const std::string &anim){
		if( act.length() <= 0) return false;
		typedef std::pair <const std::string, const std::string>
			SSPair;
		return ((m_action_anim.insert( SSPair( act, anim))).second);
	}
	void remove_action_anim( const std::string &act) {
		Map_String_String :: iterator iter = m_action_anim.find( act);
		if( iter == m_action_anim.end()) return;
		m_action_anim.erase( iter);
	}
	void clear_action_anim() { m_action_anim.clear();}

	std::string get_action_anim( const std::string &act) {
		Map_String_String :: const_iterator iter =
			m_action_anim.find( act);
		if( iter == m_action_anim.end()) return std::string();
		return (iter->second);
	}


protected:
	ModelHandle *m_mh;
	sk3DRomDrawContext *m_draw_context;

	bool m_enable_animation;
	rgLONG64 m_ani_clock_prev;
	rgFloat m_ani_pace;

	std::map <const std::string, GLuint> m_display_list;
	std::string m_use_display_list;

	std::map <const std::string, std::vector <int> *> m_bbox;

	friend class gmModelBinder;
	gmModelBinder *m_bind_master;
	std::map <ThingControlBlock *, gmModelBinder *> m_binder;

	std::map <const std::string, const std::string> m_action_anim;

	int get_bbox_bone_id( const int id);
	rgVector get_bbox_center( const int id);
	rgVector get_bbox_center( const int id, const rgMatrix &mat);

	static std::map <const std::string, THING_COMMAND_FN> m_ccmd;
	friend gmTE gmModelThing_set_model( const gmCCArg &p);
	friend gmTE gmModelThing_transform( const gmCCArg &p);
	friend gmTE gmModelThing_dump_matrix( const gmCCArg &p);
	friend gmTE gmModelThing_load_identity( const gmCCArg &p);
	friend gmTE gmModelThing_translate( const gmCCArg &p);
	friend gmTE gmModelThing_rotate( const gmCCArg &p);
	friend gmTE gmModelThing_scale( const gmCCArg &p);
	friend gmTE gmModelThing_animate( const gmCCArg &p);
	friend gmTE gmModelThing_set_dc_root_master( const gmCCArg &p);
	friend gmTE gmModelThing_clear_display_lists( const gmCCArg &p);
	friend gmTE gmModelThing_create_display_list( const gmCCArg &p);
	friend gmTE gmModelThing_gen_display_list( const gmCCArg &p);
	friend gmTE gmModelThing_use_display_list( const gmCCArg &p);
	friend gmTE gmModelThing_use_vertex_array( const gmCCArg &p);
	friend gmTE gmModelThing_add_bboxes( const gmCCArg &p);
	friend gmTE gmModelThing_add_bbox_list( const gmCCArg &p);
	friend gmTE gmModelThing_remove_bbox_list( const gmCCArg &p);
	friend gmTE gmModelThing_add_action_anim( const gmCCArg &p);
	friend gmTE gmModelThing_remove_action_anim( const gmCCArg &p);

private:
	gmModelThing( const gmModelThing &);
	gmModelThing &operator =( const gmModelThing &);
};


class gmPhyModelCOM
{
public:
	rgVector m_position;		// Real actual position

	// PSUEDO: m_position = bone_transformation( m_bound_position)
	rgVector m_bound_position;

	float m_bounding_radius;
	float m_collision_radius;

	float m_mass;
	int m_bone_id;
	//rgVector velocity;
	//rgVector torque;

	gmPhyModelCOM() {
		m_mass = 1;
		m_bone_id = -1;
		m_bounding_radius = 1.0f;
		m_collision_radius = 1.0f;
	}
};

// gmPhyModel
// Description: Similar to "gmModelThing" except that now physical properties
//	can be attributed to the model. For example, the "thing" can perform
//	collision detection with its model.
class gmPhyModel : public gmModelThing
{
public:
	gmPhyModel();
	virtual ~gmPhyModel();

	void process( ThingManager *tm);

	void draw( gmRenderArgs *);
	void draw_dynamic_bboxes();

	// Description: Applies gravity to the "things" center-of-mass (COM).
	//	COM is a point bound to a skeleton's root bone. A skeleton
	//	may have multiple root bones, however ONLY ONE bone can be
	//	used to define the center of mass. Inorder to allow gravity to
	//	affect all bones in the skeleton, all other bones must be
	//	either "bound" or "mounted" to the root bone containing the
	//	center of mass. m_draw_context->m_skout is the skeleton frame
	//	affected by gravity.
	void apply_gravity();

	// Description: Applies the "things" velocity to the center-of-mass
	//	(COM).
	void apply_velocity();

	// Description: Returns information about the models center-of-mass
	//	(COM). This includes things such as the current location of
	//	the COM relative to the world, and the mass of COM.
	gmPhyModelCOM &center_of_mass();

	// The center-of-mass (COM) defines the "things" position.
	rgVector center_point() { return (center_of_mass()).m_position;}
	rgCoordSystem cs();

	std::string type() const { return std::string( "gmPhyModel");}
	bool is_a( const std::string &s) const {
		if( s != gmPhyModel::type()) return gmModelThing::is_a( s);
		return true;
	}

	gmThingError call_ccmd( const std::string &key, gmCCArg &args);
	std::string list_ccmd( const std::string &p_prefix);

	sk3DRomPhySkel *skel() { return m_skel;}

	// Description: Adds dynamic bounding boxes belonging to the given
	//	group number.
	bool add_dynamic_bboxes( const std::string &key);

	void clear_dynamic_bboxes();

	std::string dump_dyn_bboxes();

	int get_bbox( std::vector <rgVector> &v, int id);

	PhyBoundingPoint get_bbox_point( const int id);
	PhyBoundingPoint get_bbox_point( const int id, const rgMatrix &mat);

	rgVector post_collision_vector( const rgVector v, const rgVector n)
		const;

	rgVector anti_intrusion( PhyCollisionFragment &f);

//	rgVector reflection( const rgVector v, const rgCollisionFragment &f)
//		const;

//	rgVector projection( const rgVector v, const rgCollisionFragment &f)
//		const;


	void get_origin_points( PhyCollisionContext &cc) const;
	void get_bounding_points( PhyCollisionContext &cc) const;

	// Description: Generates bounding points for the given physical model
	//	based on the models current skeleton, specified by the models
	//	draw context (m_draw_context).
	void gen_bounding_points( PhyCollisionContext &cc) const;
	void gen_bounding_points( PhyCollisionContext &cc, const rgMatrix &mat)
		const;


	// XXX: Hack Functions.
	//
	//	The following functions provide features that are not generally
	//	applicable to different game types. These functions are
	//	generally used for optimizations or simplification of complex
	//	physical simulations.
	//
	//	In addition, these functions will usually vary from game to
	//	game (or may not exist at all).
	float hack_ground_anti_intrusion( PhyCollisionContext *m_cc) const;

	virtual bool collision( PhyCollisionFragment &f, gmPhyModel *t);

protected:
	sk3DRomPhySkel *m_skel;
	gmPhyModelCOM m_com;

	PhyCollisionContext *m_cc;

	std::vector <std::string> m_collision_list;

	float m_bounciness;

	bool m_draw_dyn_bboxes;

	float m_hack_ground;
	float m_hack_ground_friction;

	// m_dyn_bbox contains a list of bounding boxes that define what parts
	// of the model can collide with the enviornment or other models.
	// These bounding boxes are dynamic (meaning that skeletal
	// transformations affect the bounding boxes).
	// m_dyn_bbox (specifically) contains a list of rg3DRom brushes from
	// the model. When a bounding box is required, the index number is
	// looked up and the brush is converted into a bounding box.
	std::vector <int> m_dyn_bbox;

	static std::map <const std::string, THING_COMMAND_FN> m_ccmd;
	friend gmTE gmPhyModel_set_model( const gmCCArg &p);
	friend gmTE gmPhyModel_load_identity( const gmCCArg &p);
	friend gmTE gmPhyModel_scale( const gmCCArg &p);
	friend gmTE gmPhyModel_set_center_of_mass( const gmCCArg &p);
	friend gmTE gmPhyModel_add_dynamic_bboxes( const gmCCArg &p);
	friend gmTE gmPhyModel_dump_dynamic_bboxes( const gmCCArg &p);
	friend gmTE gmPhyModel_draw_dynamic_bboxes( const gmCCArg &p);
	friend gmTE gmPhyModel_bounciness( const gmCCArg &p);
	friend gmTE gmPhyModel_ground_friction( const gmCCArg &p);
	friend gmTE gmPhyModel_set_collision_radius( const gmCCArg &p);
	friend gmTE gmPhyModel_add_collision_list( const gmCCArg &p);

private:

	gmPhyModelCOM &center_of_mass( const rgMatrix &mat);

	gmPhyModel( const gmPhyModel &);
	gmPhyModel &operator =( const gmPhyModel &);
};


class ThingControlBlock
{
public:
	gmThing *m_thing;			// The thing
	std::string m_key;

	// Process list thing is in
	std::list <ThingControlBlock *> *m_proc;
	std::list <ThingControlBlock *>::iterator m_iter;

	int m_prev;
	int m_next;

	ThingControlBlock() {
		m_thing = 0;
		m_proc = 0;
		m_prev = -1;
		m_next = -1;
		m_free = true;
		m_manager = 0;
	}

	bool is_free() const { return m_free;}

	ThingManager *manager() { return m_manager;}

private:
	friend class ThingManager;
	ThingManager *m_manager;
	bool m_free;
};

#define TM_MAX_PROC_LIST	17
#define TM_MAX_THINGS		16000
#define TM_ERR_GENERIC		-1
#define TM_ERR_SAME_KEY		-3
#define TM_ERR_NO_KEY		-4
#define TM_ERR_NO_THING		-5
#define TM_ERR_OPEN_FILE	-6
#define TM_ERR_WRITE_FILE	-7
class ThingManager
{
public:

	ThingManager( GameEngine *ge);
	~ThingManager();

	void clear();
	int create_thing( const std::string &key, gmThing *thg) {
		return create_thing( key, thg, -1);}
	int create_thing( const std::string &key, gmThing *thg, int p_pid);

	int remove_thing( const std::string &key);

	void process();
	GameEngine *ge() { return m_game_engine;}

	int count() const { return m_thing_count;}
	void dbg_stats( std::string &des) const;

	ThingControlBlock *get( int id) const;
	ThingControlBlock *get( const std::string &key) const;
	std::string dump_thing( const int id) const;
	std::string dump_thing( const std::string &key) const;
	int dump_all_things( std::string *s, const char *fname) const;

	int spawn( const std::string &key, const std::string &type);
	std::string list_spawnable() const;
	int add_to_spawn_list( const std::string &key, gmThing *(*fn)());

	int add_to_delete_list( const std::string &key, int (*fn)( gmThing *));

	void *get_module( const std::string &key) const;
	int load_module( const std::string &key, const std::string &fname,
		std::string &emsg);

	void move_thing_to_proc_list( ThingControlBlock *tcb, int pid);

	int index( ThingControlBlock *tcb) const {
		if( !tcb) return -1;
		return int( tcb - m_things);
	}

	int thing_list_check() const;

	#ifdef RPH_STATIC_BUILD
	int startup_static_thing_module( const std::string &key,
		std::string &emsg);
	#endif

protected:
	void process_list( std::list <ThingControlBlock *> &p_proc);

	GameEngine *m_game_engine;

	int m_max_things;
	int m_thing_count;
	ThingControlBlock *m_things;
	std::list <ThingControlBlock *> m_proc[ TM_MAX_PROC_LIST];
	std::map <const std::string, int> m_thing_map;

	std::map <const std::string, gmThing *(*)()> m_spawn;
	std::map <const std::string, int (*)( gmThing *)> m_delete;

	std::map <const std::string, void *> m_module;

	int m_thing_head;
	int m_thing_free;

	void init_things();
	void move_thing_to_proc( ThingControlBlock &tcb,
		std::list <ThingControlBlock *> *p_proc);
	void remove_thing_from_proc( ThingControlBlock &tcb);
	void init_spawn_list();

	void html_dump_all_things( std::ostream &des) const;
	std::string html_dump_thing( const int id) const;

private:
	ThingManager( const ThingManager &);
	ThingManager &operator =( const ThingManager &);
};

#endif
